@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1995-2008 University of Karlsruhe.  All right reserved.
+ * Copyright (C) 1995-2011 University of Karlsruhe.  All right reserved.
  *
  * This file is part of libFirm.
  *
@@ -22,7 +22,6 @@
  * @brief     Access function for backedges -- private header.
  * @author    Goetz Lindenmaier
  * @date      7.2002
- * @version   $Id: irbackedge_t.h 17275 2008-01-10 02:55:39Z beck $
  */
 #ifndef FIRM_ANA_IRBACKEDGE_T_H
 #define FIRM_ANA_IRBACKEDGE_T_H
@@ -33,12 +32,7 @@
  * @param obst   the obstack to allocate the array on
  * @param size   the size of the backedge array
  */
-unsigned *new_backedge_arr(struct obstack *obst, unsigned size);
-
-/**
- * Allocate new backedge info for a given node.
- */
-void new_backedge_info(ir_node *n);
+bitset_t *new_backedge_arr(struct obstack *obst, size_t size);
 
 /**
  * Adapts backedges array to new size.
