@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1995-2008 University of Karlsruhe.  All right reserved.
+ * Copyright (C) 1995-2011 University of Karlsruhe.  All right reserved.
  *
  * This file is part of libFirm.
  *
@@ -22,7 +22,6 @@
  * @brief    Memory disambiguator
  * @author   Michael Beck
  * @date     27.12.2006
- * @version  $Id: irmemory.c 25805 2009-04-06 23:00:08Z beck $
  */
 #include "config.h"
 
@@ -44,6 +43,7 @@
 #include "debug.h"
 #include "error.h"
 #include "typerep.h"
+#include "irpass.h"
 
 /** The debug handle. */
 DEBUG_ONLY(static firm_dbg_module_t *dbg = NULL;)
@@ -55,35 +55,41 @@ static DISAMBIGUATOR_FUNC language_disambuigator = NULL;
 /** The global memory disambiguator options. */
 static unsigned global_mem_disamgig_opt = aa_opt_no_opt;
 
-/* Returns a human readable name for an alias relation. */
-const char *get_ir_alias_relation_name(ir_alias_relation rel) {
+const char *get_ir_alias_relation_name(ir_alias_relation rel)
+{
 #define X(a) case a: return #a
 	switch (rel) {
 	X(ir_no_alias);
 	X(ir_may_alias);
 	X(ir_sure_alias);
-	default: assert(0); return "UNKNOWN";
+	default:
+		panic("UNKNOWN alias relation");
 	}
 #undef X
 }
 
-/* Get the memory disambiguator options for a graph. */
-unsigned get_irg_memory_disambiguator_options(const ir_graph *irg) {
+unsigned get_irg_memory_disambiguator_options(const ir_graph *irg)
+{
 	unsigned opt = irg->mem_disambig_opt;
 	if (opt & aa_opt_inherited)
 		return global_mem_disamgig_opt;
 	return opt;
-}  /* get_irg_memory_disambiguator_options */
+}
 
-/*  Set the memory disambiguator options for a graph. */
-void set_irg_memory_disambiguator_options(ir_graph *irg, unsigned options) {
+void set_irg_memory_disambiguator_options(ir_graph *irg, unsigned options)
+{
 	irg->mem_disambig_opt = options & ~aa_opt_inherited;
-}  /* set_irg_memory_disambiguator_options */
+}
 
-/* Set the global disambiguator options for all graphs not having local options. */
-void set_irp_memory_disambiguator_options(unsigned options) {
+void set_irp_memory_disambiguator_options(unsigned options)
+{
 	global_mem_disamgig_opt = options;
-}  /* set_irp_memory_disambiguator_options */
+}
+
+ir_storage_class_class_t get_base_sc(ir_storage_class_class_t x)
+{
+	return x & ~ir_sc_modifiers;
+}
 
 /**
  * Find the base address and entity of an Sel node.
@@ -93,7 +99,8 @@ void set_irp_memory_disambiguator_options(unsigned options) {
  *
  * @return the base address.
  */
-static ir_node *find_base_adr(const ir_node *sel, ir_entity **pEnt) {
+static ir_node *find_base_adr(const ir_node *sel, ir_entity **pEnt)
+{
 	ir_node *ptr = get_Sel_ptr(sel);
 
 	while (is_Sel(ptr)) {
@@ -102,7 +109,7 @@ static ir_node *find_base_adr(const ir_node *sel, ir_entity **pEnt) {
 	}
 	*pEnt = get_Sel_entity(sel);
 	return ptr;
-}  /* find_base_adr */
+}
 
 /**
  * Check if a given Const node is greater or equal a given size.
@@ -112,15 +119,16 @@ static ir_node *find_base_adr(const ir_node *sel, ir_entity **pEnt) {
  *
  * @return ir_no_alias if the Const is greater, ir_may_alias else
  */
-static ir_alias_relation check_const(const ir_node *cns, int size) {
-	tarval *tv = get_Const_tarval(cns);
-	tarval *tv_size;
+static ir_alias_relation check_const(const ir_node *cns, int size)
+{
+	ir_tarval *tv = get_Const_tarval(cns);
+	ir_tarval *tv_size;
 
 	if (size == 0)
 		return tarval_is_null(tv) ? ir_may_alias : ir_no_alias;
 	tv_size = new_tarval_from_long(size, get_tarval_mode(tv));
-	return tarval_cmp(tv_size, tv) & (pn_Cmp_Eq|pn_Cmp_Lt) ? ir_no_alias : ir_may_alias;
-}  /* check_const */
+	return tarval_cmp(tv_size, tv) & (ir_relation_less_equal) ? ir_no_alias : ir_may_alias;
+}
 
 /**
  * Treat idx1 and idx2 as integer indexes and check if they differ always more than size.
@@ -133,14 +141,15 @@ static ir_alias_relation check_const(const ir_node *cns, int size) {
  *         ir_no_alias iff they ALWAYS differ more than size
  *         ir_may_alias else
  */
-static ir_alias_relation different_index(const ir_node *idx1, const ir_node *idx2, int size) {
+static ir_alias_relation different_index(const ir_node *idx1, const ir_node *idx2, int size)
+{
 	if (idx1 == idx2)
 		return ir_sure_alias;
 	if (is_Const(idx1) && is_Const(idx2)) {
 		/* both are const, we can compare them */
-		tarval *tv1 = get_Const_tarval(idx1);
-		tarval *tv2 = get_Const_tarval(idx2);
-		tarval *tv, *tv_size;
+		ir_tarval *tv1 = get_Const_tarval(idx1);
+		ir_tarval *tv2 = get_Const_tarval(idx2);
+		ir_tarval *tv, *tv_size;
 		ir_mode *m1, *m2;
 
 		if (size == 0)
@@ -173,20 +182,20 @@ static ir_alias_relation different_index(const ir_node *idx1, const ir_node *idx
 			if (get_mode_sign(m1) != get_mode_sign(m2)) {
 				/* find the signed */
 				if (mode_is_signed(m2)) {
-					tarval *t = tv1;
+					ir_tarval *t = tv1;
 					ir_mode *tm = m1;
 					tv1 = tv2; m1 = m2;
 					tv2 = t;   m2 = tm;
 				}
 
 				/* m1 is now the signed one */
-				if (tarval_cmp(tv1, get_tarval_null(m1)) & (pn_Cmp_Eq|pn_Cmp_Gt)) {
+				if (!tarval_is_negative(tv1)) {
 					/* tv1 is signed, but >= 0, simply cast into unsigned */
 					tv1 = tarval_convert_to(tv1, m2);
 				} else {
 					tv_size = new_tarval_from_long(size, m2);
 
-					if (tarval_cmp(tv2, tv_size) & (pn_Cmp_Eq|pn_Cmp_Gt)) {
+					if (tarval_cmp(tv2, tv_size) & (ir_relation_greater_equal)) {
 						/* tv1 is negative and tv2 >= tv_size, so the difference is bigger than size */
 						return ir_no_alias;
 					}
@@ -200,19 +209,19 @@ static ir_alias_relation different_index(const ir_node *idx1, const ir_node *idx
 					tv1 = tarval_convert_to(tv1, m2);
 
 					/* now we can compare without overflow */
-					return tarval_cmp(tv1, tv2) & (pn_Cmp_Eq|pn_Cmp_Gt) ? ir_no_alias : ir_may_alias;
+					return tarval_cmp(tv1, tv2) & (ir_relation_greater_equal) ? ir_no_alias : ir_may_alias;
 				}
 			}
 		}
-		if (tarval_cmp(tv1, tv2) == pn_Cmp_Gt) {
-			tarval *t = tv1;
+		if (tarval_cmp(tv1, tv2) == ir_relation_greater) {
+			ir_tarval *t = tv1;
 			tv1 = tv2;
 			tv2 = t;
 		}
 		/* tv1 is now the "smaller" one */
 		tv      = tarval_sub(tv2, tv1, NULL);
 		tv_size = new_tarval_from_long(size, get_tarval_mode(tv));
-		return tarval_cmp(tv_size, tv) & (pn_Cmp_Eq|pn_Cmp_Lt) ? ir_no_alias : ir_may_alias;
+		return tarval_cmp(tv_size, tv) & (ir_relation_less_equal) ? ir_no_alias : ir_may_alias;
 	}
 
 	/* Note: we rely here on the fact that normalization puts constants on the RIGHT side */
@@ -283,16 +292,17 @@ static ir_alias_relation different_index(const ir_node *idx1, const ir_node *idx
 
 	}
 	return ir_may_alias;
-}  /* different_index */
+}
 
 /**
- * Two Sel addresses have the same base address, check if there offsets are
+ * Two Sel addresses have the same base address, check if their offsets are
  * different.
  *
  * @param adr1  The first address.
  * @param adr2  The second address.
  */
-static ir_alias_relation different_sel_offsets(const ir_node *sel1, const ir_node *sel2) {
+static ir_alias_relation different_sel_offsets(const ir_node *sel1, const ir_node *sel2)
+{
 	/* seems to be broken */
 	(void) sel1;
 	(void) sel2;
@@ -338,7 +348,7 @@ static ir_alias_relation different_sel_offsets(const ir_node *sel1, const ir_nod
 	(void) different_index;
 #endif
 	return ir_may_alias;
-}  /* different_sel_offsets */
+}
 
 /**
  * Determine the alias relation by checking if adr1 and adr2 are pointer
@@ -351,13 +361,13 @@ static ir_alias_relation different_types(const ir_node *adr1, const ir_node *adr
 {
 	ir_entity *ent1 = NULL, *ent2 = NULL;
 
-	if (is_Global(adr1))
-		ent1 = get_Global_entity(adr1);
+	if (is_SymConst_addr_ent(adr1))
+		ent1 = get_SymConst_entity(adr1);
 	else if (is_Sel(adr1))
 		ent1 = get_Sel_entity(adr1);
 
-	if (is_Global(adr2))
-		ent2 = get_Global_entity(adr2);
+	if (is_SymConst_addr_ent(adr2))
+		ent2 = get_SymConst_entity(adr2);
 	else if (is_Sel(adr2))
 		ent2 = get_Sel_entity(adr2);
 
@@ -388,14 +398,15 @@ static ir_alias_relation different_types(const ir_node *adr1, const ir_node *adr
 		}
 	}
 	return ir_may_alias;
-}  /* different_types */
+}
 
 /**
  * Returns non-zero if a node is a result on a malloc-like routine.
  *
  * @param node  the Proj node to test
  */
-static int is_malloc_Result(const ir_node *node) {
+static int is_malloc_Result(const ir_node *node)
+{
 	node = get_Proj_pred(node);
 	if (! is_Proj(node))
 		return 0;
@@ -403,39 +414,31 @@ static int is_malloc_Result(const ir_node *node) {
 	if (! is_Call(node))
 		return 0;
 	node = get_Call_ptr(node);
-	if (is_Global(node)) {
-		ir_entity *ent = get_Global_entity(node);
+	if (is_SymConst_addr_ent(node)) {
+		ir_entity *ent = get_SymConst_entity(node);
 
 		if (get_entity_additional_properties(ent) & mtp_property_malloc)
 			return 1;
 		return 0;
 	}
 	return 0;
-}  /* is_malloc_Result */
+}
 
-/**
- * Classify a base pointer.
- *
- * @param irg  the graph of the pointer
- * @param irn  the node representing the base address
- * @param ent  the base entity of the base address iff any
- */
-ir_storage_class_class_t classify_pointer(const ir_graph *irg, const ir_node *irn, const ir_entity *ent)
+ir_storage_class_class_t classify_pointer(const ir_node *irn,
+                                          const ir_entity *ent)
 {
+	ir_graph *irg = get_irn_irg(irn);
 	ir_storage_class_class_t res = ir_sc_pointer;
-	if (is_Global(irn)) {
-		ir_entity *entity = get_Global_entity(irn);
-		res = ir_sc_globalvar;
+	if (is_SymConst_addr_ent(irn)) {
+		ir_entity *entity = get_SymConst_entity(irn);
+		ir_type   *owner  = get_entity_owner(entity);
+		res = owner == get_tls_type() ? ir_sc_tls : ir_sc_globalvar;
 		if (! (get_entity_usage(entity) & ir_usage_address_taken))
 			res |= ir_sc_modifier_nottaken;
 	} else if (irn == get_irg_frame(irg)) {
 		res = ir_sc_localvar;
 		if (ent != NULL && !(get_entity_usage(ent) & ir_usage_address_taken))
 			res |= ir_sc_modifier_nottaken;
-	} else if (irn == get_irg_tls(irg)) {
-		res = ir_sc_tls;
-		if (ent != NULL && !(get_entity_usage(ent) & ir_usage_address_taken))
-			res |= ir_sc_modifier_nottaken;
 	} else if (is_Proj(irn) && is_malloc_Result(irn)) {
 		return ir_sc_malloced;
 	} else if (is_Const(irn)) {
@@ -450,11 +453,12 @@ ir_storage_class_class_t classify_pointer(const ir_graph *irg, const ir_node *ir
 /**
  * If adr represents a Bitfield Sel, skip it
  */
-static const ir_node *skip_Bitfield_Sels(const ir_node *adr) {
+static const ir_node *skip_Bitfield_Sels(const ir_node *adr)
+{
 	if (is_Sel(adr)) {
 		ir_entity *ent     = get_Sel_entity(adr);
 		ir_type   *bf_type = get_entity_type(ent);
-		
+
 		/* is it a bitfield type? */
 		if (is_Primitive_type(bf_type) && get_primitive_base_type(bf_type) != NULL)
 			adr = get_Sel_ptr(adr);
@@ -465,7 +469,6 @@ static const ir_node *skip_Bitfield_Sels(const ir_node *adr) {
 /**
  * Determine the alias relation between two addresses.
  *
- * @param irg    the graph of both memory operations
  * @param addr1  pointer address of the first memory operation
  * @param mode1  the mode of the accessed data through addr1
  * @param addr2  pointer address of the second memory operation
@@ -474,7 +477,6 @@ static const ir_node *skip_Bitfield_Sels(const ir_node *adr) {
  * @return found memory relation
  */
 static ir_alias_relation _get_alias_relation(
-	const ir_graph *irg,
 	const ir_node *adr1, const ir_mode *mode1,
 	const ir_node *adr2, const ir_mode *mode2)
 {
@@ -486,6 +488,7 @@ static ir_alias_relation _get_alias_relation(
 	const ir_node         *base2;
 	const ir_node         *orig_adr1 = adr1;
 	const ir_node         *orig_adr2 = adr2;
+	ir_graph              *irg;
 	unsigned              mode_size;
 	ir_storage_class_class_t class1, class2, mod1, mod2;
 	int                   have_const_offsets;
@@ -496,6 +499,7 @@ static ir_alias_relation _get_alias_relation(
 	if (adr1 == adr2)
 		return ir_sure_alias;
 
+	irg = get_irn_irg(adr1);
 	options = get_irg_memory_disambiguator_options(irg);
 
 	/* The Armageddon switch */
@@ -510,7 +514,7 @@ static ir_alias_relation _get_alias_relation(
 	while (is_Add(adr1)) {
 		ir_node *add_right = get_Add_right(adr1);
 		if (is_Const(add_right) && !mode_is_reference(get_irn_mode(add_right))) {
-			tarval *tv  = get_Const_tarval(add_right);
+			ir_tarval *tv  = get_Const_tarval(add_right);
 			offset1    += get_tarval_long(tv);
 			adr1        = get_Add_left(adr1);
 		} else if (mode_is_reference(get_irn_mode(add_right))) {
@@ -524,7 +528,7 @@ static ir_alias_relation _get_alias_relation(
 	while (is_Add(adr2)) {
 		ir_node *add_right = get_Add_right(adr2);
 		if (is_Const(add_right) && !mode_is_reference(get_irn_mode(add_right))) {
-			tarval *tv  = get_Const_tarval(add_right);
+			ir_tarval *tv  = get_Const_tarval(add_right);
 			offset2    += get_tarval_long(tv);
 			adr2        = get_Add_left(adr2);
 		} else if (mode_is_reference(get_irn_mode(add_right))) {
@@ -551,10 +555,10 @@ static ir_alias_relation _get_alias_relation(
 			return ir_sure_alias;
 	}
 
-	/* 
+	/*
 	 * Bitfields can be constructed as Sels from its base address.
 	 * As they have different entities, the disambiguator would find that they are
-	 * alias free. While this is true for it's values, it is false for the addresses
+	 * alias free. While this is true for its values, it is false for the addresses
 	 * (strictly speaking, the Sel's are NOT the addresses of the bitfields).
 	 * So, skip those bitfield selecting Sel's.
 	 */
@@ -581,11 +585,21 @@ static ir_alias_relation _get_alias_relation(
 			return different_sel_offsets(adr1, adr2);
 	}
 
-	mod1 = classify_pointer(irg, base1, ent1);
-	mod2 = classify_pointer(irg, base2, ent2);
+	mod1 = classify_pointer(base1, ent1);
+	mod2 = classify_pointer(base2, ent2);
 
-	class1 = GET_BASE_SC(mod1);
-	class2 = GET_BASE_SC(mod2);
+	class1 = get_base_sc(mod1);
+	class2 = get_base_sc(mod2);
+
+	/* struct-access cannot alias with variables */
+	if (ent1 == NULL && ent2 != NULL && is_compound_type(get_entity_owner(ent2))
+		&& (class1 == ir_sc_globalvar || class1 == ir_sc_localvar || class1 == ir_sc_tls || class1 == ir_sc_globaladdr)) {
+		return ir_no_alias;
+	}
+	if (ent2 == NULL && ent1 != NULL && is_compound_type(get_entity_owner(ent1))
+		&& (class2 == ir_sc_globalvar || class2 == ir_sc_localvar || class2 == ir_sc_tls || class2 == ir_sc_globaladdr)) {
+		return ir_no_alias;
+	}
 
 	if (class1 == ir_sc_pointer || class2 == ir_sc_pointer) {
 		/* swap pointer class to class1 */
@@ -593,8 +607,8 @@ static ir_alias_relation _get_alias_relation(
 			ir_storage_class_class_t temp = mod1;
 			mod1 = mod2;
 			mod2 = temp;
-			class1 = GET_BASE_SC(mod1);
-			class2 = GET_BASE_SC(mod2);
+			class1 = get_base_sc(mod1);
+			class2 = get_base_sc(mod2);
 		}
 		/* a pointer and an object whose address was never taken */
 		if (mod2 & ir_sc_modifier_nottaken) {
@@ -624,10 +638,10 @@ static ir_alias_relation _get_alias_relation(
 			/* for some reason CSE didn't happen yet for the 2 SymConsts... */
 			return ir_may_alias;
 		} else if (class1 == ir_sc_globaladdr) {
-			tarval *tv  = get_Const_tarval(base1);
-			offset1    += get_tarval_long(tv);
-			tv          = get_Const_tarval(base2);
-			offset2    += get_tarval_long(tv);
+			ir_tarval *tv = get_Const_tarval(base1);
+			offset1      += get_tarval_long(tv);
+			tv            = get_Const_tarval(base2);
+			offset2      += get_tarval_long(tv);
 
 			if ((unsigned long)labs(offset2 - offset1) >= mode_size)
 				return ir_no_alias;
@@ -668,71 +682,62 @@ leave_type_based_alias:;
 
 	/* do we have a language specific memory disambiguator? */
 	if (language_disambuigator != NULL) {
-		ir_alias_relation rel = language_disambuigator(irg, orig_adr1, mode1, orig_adr2, mode2);
+		ir_alias_relation rel = language_disambuigator(orig_adr1, mode1, orig_adr2, mode2);
 		if (rel != ir_may_alias)
 			return rel;
 	}
 
 	/* access points-to information here */
 	return ir_may_alias;
-}  /* _get_alias_relation */
+}
 
-/*
- * Determine the alias relation between two addresses.
- */
 ir_alias_relation get_alias_relation(
-	const ir_graph *irg,
 	const ir_node *adr1, const ir_mode *mode1,
 	const ir_node *adr2, const ir_mode *mode2)
 {
-	ir_alias_relation rel = _get_alias_relation(irg, adr1, mode1, adr2, mode2);
+	ir_alias_relation rel = _get_alias_relation(adr1, mode1, adr2, mode2);
 	DB((dbg, LEVEL_1, "alias(%+F, %+F) = %s\n", adr1, adr2, get_ir_alias_relation_name(rel)));
 	return rel;
-}  /* get_alias_relation */
+}
 
-/* Set a source language specific memory disambiguator function. */
-void set_language_memory_disambiguator(DISAMBIGUATOR_FUNC func) {
+void set_language_memory_disambiguator(DISAMBIGUATOR_FUNC func)
+{
 	language_disambuigator = func;
-}  /* set_language_memory_disambiguator */
+}
 
 /** The result cache for the memory disambiguator. */
 static set *result_cache = NULL;
 
 /** An entry in the relation cache. */
 typedef struct mem_disambig_entry {
-	const ir_node	  *adr1;    /**< The first address. */
-	const ir_mode	  *mode1;   /**< The first address mode. */
-	const ir_node	  *adr2;    /**< The second address. */
-	const ir_mode	  *mode2;   /**< The second address mode. */
+	const ir_node     *adr1;    /**< The first address. */
+	const ir_mode     *mode1;   /**< The first address mode. */
+	const ir_node     *adr2;    /**< The second address. */
+	const ir_mode     *mode2;   /**< The second address mode. */
 	ir_alias_relation result;   /**< The alias relation result. */
 } mem_disambig_entry;
 
-#define HASH_ENTRY(adr1, adr2)	(HASH_PTR(adr1) ^ HASH_PTR(adr2))
+#define HASH_ENTRY(adr1, adr2)  (hash_ptr(adr1) ^ hash_ptr(adr2))
 
 /**
  * Compare two relation cache entries.
  */
-static int cmp_mem_disambig_entry(const void *elt, const void *key, size_t size) {
-	const mem_disambig_entry *p1 = elt;
-	const mem_disambig_entry *p2 = key;
+static int cmp_mem_disambig_entry(const void *elt, const void *key, size_t size)
+{
+	const mem_disambig_entry *p1 = (const mem_disambig_entry*) elt;
+	const mem_disambig_entry *p2 = (const mem_disambig_entry*) key;
 	(void) size;
 
-	return p1->adr1 == p2->adr1 && p1->adr2 == p2->adr2 && 
+	return p1->adr1 == p2->adr1 && p1->adr2 == p2->adr2 &&
 	       p1->mode1 == p2->mode1 && p1->mode2 == p2->mode2;
-}  /* cmp_mem_disambig_entry */
+}
 
-/**
- * Initialize the relation cache.
- */
-void mem_disambig_init(void) {
+void mem_disambig_init(void)
+{
 	result_cache = new_set(cmp_mem_disambig_entry, 8);
-}  /* mem_disambig_init */
+}
 
-/*
- * Determine the alias relation between two addresses.
- */
 ir_alias_relation get_alias_relation_ex(
-	const ir_graph *irg,
 	const ir_node *adr1, const ir_mode *mode1,
 	const ir_node *adr2, const ir_mode *mode2)
 {
@@ -753,23 +758,23 @@ ir_alias_relation get_alias_relation_ex(
 	key.adr2  = adr2;
 	key.mode1 = mode1;
 	key.mode2 = mode2;
-	entry = set_find(result_cache, &key, sizeof(key), HASH_ENTRY(adr1, adr2));
+	entry = set_find(mem_disambig_entry, result_cache, &key, sizeof(key), HASH_ENTRY(adr1, adr2));
 	if (entry != NULL)
 		return entry->result;
 
-	key.result = get_alias_relation(irg, adr1, mode1, adr2, mode2);
+	key.result = get_alias_relation(adr1, mode1, adr2, mode2);
 
-	set_insert(result_cache, &key, sizeof(key), HASH_ENTRY(adr1, adr2));
+	(void)set_insert(mem_disambig_entry, result_cache, &key, sizeof(key), HASH_ENTRY(adr1, adr2));
 	return key.result;
-}  /* get_alias_relation_ex */
+}
 
-/* Free the relation cache. */
-void mem_disambig_term(void) {
+void mem_disambig_term(void)
+{
 	if (result_cache != NULL) {
 		del_set(result_cache);
 		result_cache = NULL;
 	}
-}  /* mem_disambig_term */
+}
 
 /**
  * Check the mode of a Load/Store with the mode of the entity
@@ -790,32 +795,33 @@ void mem_disambig_term(void) {
  *
  * @return non-zero if the Load/Store is a hidden cast, zero else
  */
-static int is_hidden_cast(const ir_mode *mode, const ir_mode *ent_mode) {
+static int is_hidden_cast(const ir_mode *mode, const ir_mode *ent_mode)
+{
 	if (ent_mode == NULL)
 		return false;
 
 	if (ent_mode != mode) {
 		if (ent_mode == NULL ||
 			get_mode_size_bits(ent_mode) != get_mode_size_bits(mode) ||
-			get_mode_sort(ent_mode) != get_mode_sort(mode) ||
 			get_mode_arithmetic(ent_mode) != irma_twos_complement ||
 			get_mode_arithmetic(mode) != irma_twos_complement)
 			return true;
 	}
 	return false;
-}  /* is_hidden_cast */
+}
 
 /**
  * Determine the usage state of a node (or its successor Sels).
  *
  * @param irn  the node
  */
-static ir_entity_usage determine_entity_usage(const ir_node *irn, ir_entity *entity) {
+static ir_entity_usage determine_entity_usage(const ir_node *irn, ir_entity *entity)
+{
 	int       i;
 	ir_mode   *emode, *mode;
 	ir_node   *value;
 	ir_type   *tp;
-	ir_entity_usage res = 0;
+	unsigned   res = 0;
 
 	for (i = get_irn_n_outs(irn) - 1; i >= 0; --i) {
 		ir_node *succ = get_irn_out(irn, i);
@@ -872,14 +878,14 @@ static ir_entity_usage determine_entity_usage(const ir_node *irn, ir_entity *ent
 			res |= determine_entity_usage(succ, entity);
 			break;
 		case iro_Sel: {
-			ir_entity *entity = get_Sel_entity(succ);
+			ir_entity *sel_entity = get_Sel_entity(succ);
 			/* this analysis can't handle unions correctly */
-			if (is_Union_type(get_entity_owner(entity))) {
+			if (is_Union_type(get_entity_owner(sel_entity))) {
 				res |= ir_usage_unknown;
 				break;
 			}
 			/* Check the successor of irn. */
-			res |= determine_entity_usage(succ, entity);
+			res |= determine_entity_usage(succ, sel_entity);
 			break;
 		}
 
@@ -912,7 +918,7 @@ static ir_entity_usage determine_entity_usage(const ir_node *irn, ir_entity *ent
 					/* we found one input */
 					for (k = get_irn_n_outs(succ) - 1; k >= 0; --k) {
 						ir_node *proj = get_irn_out(succ, k);
-						
+
 						if (is_Proj(proj) && get_Proj_proj(proj) == input_nr) {
 							res |= determine_entity_usage(proj, entity);
 							break;
@@ -931,37 +937,40 @@ static ir_entity_usage determine_entity_usage(const ir_node *irn, ir_entity *ent
 		}
 	}
 
-	return res;
+	return (ir_entity_usage) res;
 }
 
 /**
  * Update the usage flags of all frame entities.
  */
-static void analyse_irg_entity_usage(ir_graph *irg) {
+static void analyse_irg_entity_usage(ir_graph *irg)
+{
 	ir_type *ft = get_irg_frame_type(irg);
 	ir_node *irg_frame;
-	int i, j, k, static_link_arg;
+	size_t i, n;
+	int j, k, static_link_arg;
+
+	assure_irg_properties(irg, IR_GRAPH_PROPERTY_CONSISTENT_OUTS);
 
 	/* set initial state to not_taken, as this is the "smallest" state */
-	for (i = get_class_n_members(ft) - 1; i >= 0; --i) {
+	for (i = 0, n = get_class_n_members(ft); i < n; ++i) {
 		ir_entity *ent = get_class_member(ft, i);
 
 		/* methods can only be analyzed globally */
 		if (! is_method_entity(ent)) {
-			ir_entity_usage  flags =
-				get_entity_stickyness(ent) == stickyness_sticky ? ir_usage_unknown : 0;
+			ir_entity_usage flags = ir_usage_none;
+			if (get_entity_linkage(ent) & IR_LINKAGE_HIDDEN_USER)
+				flags = ir_usage_unknown;
 			set_entity_usage(ent, flags);
 		}
 	}
 
-	assure_irg_outs(irg);
-
 	irg_frame = get_irg_frame(irg);
 
-	for (i = get_irn_n_outs(irg_frame) - 1; i >= 0; --i) {
-		ir_node        *succ = get_irn_out(irg_frame, i);
+	for (j = get_irn_n_outs(irg_frame) - 1; j >= 0; --j) {
+		ir_node        *succ = get_irn_out(irg_frame, j);
 		ir_entity      *entity;
-		ir_entity_usage flags;
+		unsigned        flags;
 
 		if (!is_Sel(succ))
 			continue;
@@ -969,22 +978,23 @@ static void analyse_irg_entity_usage(ir_graph *irg) {
 		entity = get_Sel_entity(succ);
 		flags  = get_entity_usage(entity);
 		flags |= determine_entity_usage(succ, entity);
-		set_entity_usage(entity, flags);
+		set_entity_usage(entity, (ir_entity_usage) flags);
 	}
 
 	/* check inner functions accessing outer frame */
 	static_link_arg = 0;
-	for (i = get_class_n_members(ft) - 1; i >= 0; --i) {
+	for (i = 0, n = get_class_n_members(ft); i < n; ++i) {
 		ir_entity *ent = get_class_member(ft, i);
 		ir_graph  *inner_irg;
 		ir_node   *args;
 
 		if (! is_method_entity(ent))
 			continue;
-		if (get_entity_peculiarity(ent) == peculiarity_description)
-			continue;
 
 		inner_irg = get_entity_irg(ent);
+		if (inner_irg == NULL)
+			continue;
+
 		assure_irg_outs(inner_irg);
 		args = get_irg_args(inner_irg);
 		for (j = get_irn_n_outs(args) - 1; j >= 0; --j) {
@@ -999,11 +1009,11 @@ static void analyse_irg_entity_usage(ir_graph *irg) {
 
 						if (get_entity_owner(entity) == ft) {
 							/* found an access to the outer frame */
-							ir_entity_usage flags;
+							unsigned flags;
 
 							flags  = get_entity_usage(entity);
 							flags |= determine_entity_usage(succ, entity);
-							set_entity_usage(entity, flags);
+							set_entity_usage(entity, (ir_entity_usage) flags);
 						}
 					}
 				}
@@ -1011,21 +1021,13 @@ static void analyse_irg_entity_usage(ir_graph *irg) {
 		}
 	}
 
-
 	/* now computed */
-	irg->entity_usage_state = ir_entity_usage_computed;
+	add_irg_properties(irg, IR_GRAPH_PROPERTY_CONSISTENT_ENTITY_USAGE);
 }
 
-ir_entity_usage_computed_state get_irg_entity_usage_state(const ir_graph *irg) {
-	return irg->entity_usage_state;
-}
-
-void set_irg_entity_usage_state(ir_graph *irg, ir_entity_usage_computed_state state) {
-	irg->entity_usage_state = state;
-}
-
-void assure_irg_entity_usage_computed(ir_graph *irg) {
-	if (irg->entity_usage_state != ir_entity_usage_not_computed)
+void assure_irg_entity_usage_computed(ir_graph *irg)
+{
+	if (irg_has_properties(irg, IR_GRAPH_PROPERTY_CONSISTENT_ENTITY_USAGE))
 		return;
 
 	analyse_irg_entity_usage(irg);
@@ -1035,21 +1037,19 @@ void assure_irg_entity_usage_computed(ir_graph *irg) {
 /**
  * Initialize the entity_usage flag for a global type like type.
  */
-static void init_entity_usage(ir_type *tp) {
-	int i;
+static void init_entity_usage(ir_type *tp)
+{
+	size_t i, n;
 
 	/* We have to be conservative: All external visible entities are unknown */
-	for (i = get_compound_n_members(tp) - 1; i >= 0; --i) {
-		ir_entity       *ent  = get_compound_member(tp, i);
-		ir_entity_usage flags = ir_usage_none;
-		ir_visibility   vis   = get_entity_visibility(ent);
+	for (i = 0, n = get_compound_n_members(tp); i < n; ++i) {
+		ir_entity       *ent   = get_compound_member(tp, i);
+		unsigned         flags = ir_usage_none;
 
-		if (vis == visibility_external_visible   ||
-		    vis == visibility_external_allocated ||
-		    get_entity_stickyness(ent) == stickyness_sticky) {
+		if (entity_is_externally_visible(ent)) {
 			flags |= ir_usage_unknown;
 		}
-		set_entity_usage(ent, flags);
+		set_entity_usage(ent, (ir_entity_usage) flags);
 	}
 }
 
@@ -1060,15 +1060,15 @@ static void init_entity_usage(ir_type *tp) {
  */
 static void check_initializer_nodes(ir_initializer_t *initializer)
 {
-	unsigned i;
-	ir_node  *n;
+	size_t  i;
+	ir_node *n;
 
 	switch (initializer->kind) {
 	case IR_INITIALIZER_CONST:
 		/* let's check if it's an address */
 		n = initializer->consti.value;
-		if (is_Global(n)) {
-			ir_entity *ent = get_Global_entity(n);
+		if (is_SymConst_addr_ent(n)) {
+			ir_entity *ent = get_SymConst_entity(n);
 			set_entity_usage(ent, ir_usage_unknown);
 		}
 		return;
@@ -1084,7 +1084,7 @@ static void check_initializer_nodes(ir_initializer_t *initializer)
 		return;
 	}
 	panic("invalid initializer found");
-}  /* check_initializer_nodes */
+}
 
 /**
  * Mark all entities used in the initializer for the given entity as unknown
@@ -1092,40 +1092,19 @@ static void check_initializer_nodes(ir_initializer_t *initializer)
  *
  * @param ent  the entity
  */
-static void check_initializer(ir_entity *ent) {
-	ir_node *n;
-	int i;
-
-	/* do not check uninitialized values */
-	if (get_entity_variability(ent) == variability_uninitialized)
-		return;
-
+static void check_initializer(ir_entity *ent)
+{
 	/* Beware: Methods are always initialized with "themself". This does not
-	   count as a taken address. */
+	 * count as a taken address.
+	 * TODO: this initialisation with "themself" is wrong and should be removed
+	 */
 	if (is_Method_type(get_entity_type(ent)))
 		return;
 
-	if (ent->has_initializer) {
-		check_initializer_nodes(ent->attr.initializer);
-	} else if (is_atomic_entity(ent)) {
-		/* let's check if it's an address */
-		n = get_atomic_ent_value(ent);
-		if (is_Global(n)) {
-			ir_entity *ent = get_Global_entity(n);
-			set_entity_usage(ent, ir_usage_unknown);
-		}
-	} else {
-		for (i = get_compound_ent_n_values(ent) - 1; i >= 0; --i) {
-			n = get_compound_ent_value(ent, i);
-
-			/* let's check if it's an address */
-			if (is_Global(n)) {
-				ir_entity *ent = get_Global_entity(n);
-				set_entity_usage(ent, ir_usage_unknown);
-			}
-		}
+	if (ent->initializer != NULL) {
+		check_initializer_nodes(ent->initializer);
 	}
-}  /* check_initializer */
+}
 
 
 /**
@@ -1133,15 +1112,16 @@ static void check_initializer(ir_entity *ent) {
  *
  * @param tp  a compound type
  */
-static void check_initializers(ir_type *tp) {
-	int i;
+static void check_initializers(ir_type *tp)
+{
+	size_t i, n;
 
-	for (i = get_compound_n_members(tp) - 1; i >= 0; --i) {
+	for (i = 0, n = get_compound_n_members(tp); i < n; ++i) {
 		ir_entity *ent = get_compound_member(tp, i);
 
 		check_initializer(ent);
 	}
-}  /* check_initializers */
+}
 
 #ifdef DEBUG_libfirm
 /**
@@ -1149,9 +1129,10 @@ static void check_initializers(ir_type *tp) {
  *
  * @param tp  a compound type
  */
-static void print_entity_usage_flags(ir_type *tp) {
-	int i;
-	for (i = get_compound_n_members(tp) - 1; i >= 0; --i) {
+static void print_entity_usage_flags(const ir_type *tp)
+{
+	size_t i, n;
+	for (i = 0, n = get_compound_n_members(tp); i < n; ++i) {
 		ir_entity *ent = get_compound_member(tp, i);
 		ir_entity_usage flags = get_entity_usage(ent);
 
@@ -1174,53 +1155,51 @@ static void print_entity_usage_flags(ir_type *tp) {
 /**
  * Post-walker: check for global entity address
  */
-static void check_global_address(ir_node *irn, void *env) {
-	ir_node *tls = env;
+static void check_global_address(ir_node *irn, void *data)
+{
 	ir_entity *ent;
-	ir_entity_usage flags;
+	unsigned flags;
+	(void) data;
 
-	if (is_Global(irn)) {
+	if (is_SymConst_addr_ent(irn)) {
 		/* A global. */
-		ent = get_Global_entity(irn);
-	} else if (is_Sel(irn) && get_Sel_ptr(irn) == tls) {
-		/* A TLS variable. */
-		ent = get_Sel_entity(irn);
+		ent = get_SymConst_entity(irn);
 	} else
 		return;
 
 	flags = get_entity_usage(ent);
 	flags |= determine_entity_usage(irn, ent);
-	set_entity_usage(ent, flags);
-}  /* check_global_address */
+	set_entity_usage(ent, (ir_entity_usage) flags);
+}
 
 /**
  * Update the entity usage flags of all global entities.
  */
-static void analyse_irp_globals_entity_usage(void) {
-	int i;
+static void analyse_irp_globals_entity_usage(void)
+{
+	size_t i, n;
 	ir_segment_t s;
 
-	for (s = IR_SEGMENT_FIRST; s < IR_SEGMENT_COUNT; ++s) {
+	for (s = IR_SEGMENT_FIRST; s <= IR_SEGMENT_LAST; ++s) {
 		ir_type *type = get_segment_type(s);
 		init_entity_usage(type);
 	}
 
-	for (s = IR_SEGMENT_FIRST; s < IR_SEGMENT_COUNT; ++s) {
+	for (s = IR_SEGMENT_FIRST; s <= IR_SEGMENT_LAST; ++s) {
 		ir_type *type = get_segment_type(s);
 		check_initializers(type);
 	}
 
-	for (i = get_irp_n_irgs() - 1; i >= 0; --i) {
+	for (i = 0, n = get_irp_n_irgs(); i < n; ++i) {
 		ir_graph *irg = get_irp_irg(i);
 
 		assure_irg_outs(irg);
-		irg_walk_graph(irg, NULL, check_global_address, get_irg_tls(irg));
+		irg_walk_graph(irg, NULL, check_global_address, NULL);
 	}
 
 #ifdef DEBUG_libfirm
 	if (firm_dbg_get_mask(dbg) & LEVEL_1) {
-		ir_segment_t s;
-		for (s = IR_SEGMENT_FIRST; s < IR_SEGMENT_COUNT; ++s) {
+		for (s = IR_SEGMENT_FIRST; s <= IR_SEGMENT_LAST; ++s) {
 			print_entity_usage_flags(get_segment_type(s));
 		}
 	}
@@ -1230,25 +1209,26 @@ static void analyse_irp_globals_entity_usage(void) {
 	irp->globals_entity_usage_state = ir_entity_usage_computed;
 }
 
-/* Returns the current address taken state of the globals. */
-ir_entity_usage_computed_state get_irp_globals_entity_usage_state(void) {
+ir_entity_usage_computed_state get_irp_globals_entity_usage_state(void)
+{
 	return irp->globals_entity_usage_state;
 }
 
-/* Sets the current address taken state of the graph. */
-void set_irp_globals_entity_usage_state(ir_entity_usage_computed_state state) {
+void set_irp_globals_entity_usage_state(ir_entity_usage_computed_state state)
+{
 	irp->globals_entity_usage_state = state;
 }
 
-/* Assure that the address taken flag is computed for the globals. */
-void assure_irp_globals_entity_usage_computed(void) {
+void assure_irp_globals_entity_usage_computed(void)
+{
 	if (irp->globals_entity_usage_state != ir_entity_usage_not_computed)
 		return;
 
 	analyse_irp_globals_entity_usage();
 }
 
-void firm_init_memory_disambiguator(void) {
+void firm_init_memory_disambiguator(void)
+{
 	FIRM_DBG_REGISTER(dbg, "firm.ana.irmemory");
 	FIRM_DBG_REGISTER(dbgcall, "firm.opt.cc");
 }
@@ -1262,29 +1242,24 @@ static pmap *mtp_map;
  *
  * @param tp  the type to clone
  */
-static ir_type *clone_type_and_cache(ir_type *tp) {
-	static ident *prefix = NULL;
-	ir_type *res;
-	pmap_entry *e = pmap_find(mtp_map, tp);
+static ir_type *clone_type_and_cache(ir_type *tp)
+{
+	ir_type *res = pmap_get(ir_type, mtp_map, tp);
 
-	if (e)
-		return e->value;
-
-	if (prefix == NULL)
-		prefix = new_id_from_chars("C", 1);
-
-	res = clone_type_method(tp, prefix);
-	pmap_insert(mtp_map, tp, res);
-	DB((dbgcall, LEVEL_2, "cloned type %+F into %+F\n", tp, res));
+	if (res == NULL) {
+		res = clone_type_method(tp);
+		pmap_insert(mtp_map, tp, res);
+	}
 
 	return res;
-}  /* clone_type_and_cache */
+}
 
 /**
- * Walker: clone all call types of Calls to methods having the 
+ * Walker: clone all call types of Calls to methods having the
  * mtp_property_private property set.
  */
-static void update_calls_to_private(ir_node *call, void *env) {
+static void update_calls_to_private(ir_node *call, void *env)
+{
 	(void) env;
 	if (is_Call(call)) {
 		ir_node *ptr = get_Call_ptr(call);
@@ -1296,18 +1271,18 @@ static void update_calls_to_private(ir_node *call, void *env) {
 			if (get_entity_additional_properties(ent) & mtp_property_private) {
 				if ((get_method_additional_properties(ctp) & mtp_property_private) == 0) {
 					ctp = clone_type_and_cache(ctp);
-					set_method_additional_property(ctp, mtp_property_private);
+					add_method_additional_properties(ctp, mtp_property_private);
 					set_Call_type(call, ctp);
-					DB((dbgcall, LEVEL_1, "changed call to private method %+F\n", ent));
+					DB((dbgcall, LEVEL_1, "changed call to private method %+F using cloned type %+F\n", ent, ctp));
 				}
 			}
 		}
 	}
-}  /* update_calls_to_private */
+}
 
-/* Mark all private methods, i.e. those of which all call sites are known. */
-void mark_private_methods(void) {
-	int i;
+void mark_private_methods(void)
+{
+	size_t i, n;
 	int changed = 0;
 
 	assure_irp_globals_entity_usage_computed();
@@ -1315,25 +1290,22 @@ void mark_private_methods(void) {
 	mtp_map = pmap_create();
 
 	/* first step: change the calling conventions of the local non-escaped entities */
-	for (i = get_irp_n_irgs() - 1; i >= 0; --i) {
+	for (i = 0, n = get_irp_n_irgs(); i < n; ++i) {
 		ir_graph        *irg   = get_irp_irg(i);
 		ir_entity       *ent   = get_irg_entity(irg);
 		ir_entity_usage  flags = get_entity_usage(ent);
 
-		/* If an entity is sticky, it might be called from external
-		   places (like inline assembler), so do NOT mark it as private. */
-		if (get_entity_visibility(ent) == visibility_local &&
-		    !(flags & ir_usage_address_taken) &&
-		    get_entity_stickyness(ent) != stickyness_sticky) {
+		if (!(flags & ir_usage_address_taken) && !entity_is_externally_visible(ent)) {
 			ir_type *mtp = get_entity_type(ent);
 
-			set_entity_additional_property(ent, mtp_property_private);
+			add_entity_additional_properties(ent, mtp_property_private);
 			DB((dbgcall, LEVEL_1, "found private method %+F\n", ent));
 			if ((get_method_additional_properties(mtp) & mtp_property_private) == 0) {
 				/* need a new type */
 				mtp = clone_type_and_cache(mtp);
+				add_method_additional_properties(mtp, mtp_property_private);
 				set_entity_type(ent, mtp);
-				set_method_additional_property(mtp, mtp_property_private);
+				DB((dbgcall, LEVEL_2, "changed entity type of %+F to %+F\n", ent, mtp));
 				changed = 1;
 			}
 		}
@@ -1343,4 +1315,9 @@ void mark_private_methods(void) {
 		all_irg_walk(NULL, update_calls_to_private, NULL);
 
 	pmap_destroy(mtp_map);
-}  /* mark_private_methods */
+}
+
+ir_prog_pass_t *mark_private_methods_pass(const char *name)
+{
+	return def_prog_pass(name ? name : "mark_private_methods", mark_private_methods);
+}
