@@ -20,7 +20,6 @@
 /**
  * @file
  * @brief    declarations for emit functions
- * @version  $Id: TEMPLATE_emitter.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_BE_TEMPLATE_TEMPLATE_EMITTER_H
 #define FIRM_BE_TEMPLATE_TEMPLATE_EMITTER_H
@@ -29,19 +28,28 @@
 #include "irnode.h"
 #include "debug.h"
 
-#include "../bearch_t.h"
-#include "../beemitter.h"
+#include "bearch.h"
+#include "beemitter.h"
 
 #include "bearch_TEMPLATE_t.h"
 
-void TEMPLATE_emit_source_register(const ir_node *node, int pos);
-void TEMPLATE_emit_dest_register(const ir_node *node, int pos);
-void TEMPLATE_emit_immediate(const ir_node *node);
+/**
+ * emit assembler instructions with format string. Automatically indents
+ * instructions and adds debug comments at the end (in verbose-asm mode).
+ * Format specifiers:
+ *
+ * fmt  parameter               output
+ * ---- ----------------------  ---------------------------------------------
+ * %%                           %
+ * %r   const arch_register_t*  register
+ * %Sx  <node>                  source register x
+ * %Dx  <node>                  destination register x
+ * %O   <node>                  shifter operand
+ * %I   <node>                  immediate
+ * %L   <node>                  target label
+ */
+void TEMPLATE_emitf(const ir_node *node, const char *format, ...);
 
-int get_TEMPLATE_reg_nr(ir_node *irn, int posi, int in_out);
-const char *get_TEMPLATE_in_reg_name(ir_node *irn, int pos);
-
-
-void TEMPLATE_gen_routine(const TEMPLATE_code_gen_t *cg, ir_graph *irg);
+void TEMPLATE_emit_routine(ir_graph *irg);
 
 #endif
