@@ -20,11 +20,12 @@
 /**
  * @file
  * @brief   declaration for the transform function (code selection)
- * @version $Id: TEMPLATE_transform.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_BE_TEMPLATE_TEMPLATE_TRANSFORM_H
 #define FIRM_BE_TEMPLATE_TEMPLATE_TRANSFORM_H
 
-void TEMPLATE_transform_node(ir_node *node, void *env);
+void TEMPLATE_init_transform(void);
+
+void TEMPLATE_transform_graph(ir_graph *irg);
 
 #endif
