@@ -21,7 +21,6 @@
  * @file
  * @brief   declarations for transform functions (code selection)
  * @author  Oliver Richter, Tobias Gneist
- * @version $Id: arm_transform.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_BE_ARM_ARM_TRANSFORM_H
 #define FIRM_BE_ARM_ARM_TRANSFORM_H
@@ -29,7 +28,7 @@
 /**
  * Transform a Firm graph into an ARM graph.
  */
-void arm_transform_graph(arm_code_gen_t *cg);
+void arm_transform_graph(ir_graph *irg);
 
 void arm_init_transform(void);
 
