@@ -22,7 +22,6 @@
  * @brief       Block-scheduling strategies.
  * @author      Matthias Braun, Christoph Mallon
  * @date        27.09.2006
- * @version     $Id: beblocksched.c 22981 2008-10-18 09:59:01Z mallon $
  *
  * The goals of the greedy (and ILP) algorithm here works by assuming that
  * we want to change as many jumps to fallthroughs as possible (executed jumps
@@ -52,6 +51,7 @@
 #include "execfreq.h"
 #include "irdump_t.h"
 #include "irtools.h"
+#include "util.h"
 #include "debug.h"
 #include "beirgmod.h"
 #include "bemodule.h"
@@ -61,26 +61,21 @@
 #include "lc_opts.h"
 #include "lc_opts_enum.h"
 
-#ifdef WITH_ILP
-#include <lpp/lpp.h>
-#include <lpp/lpp_net.h>
-#endif /* WITH_ILP */
+#include "lpp.h"
+#include "lpp_net.h"
 
 DEBUG_ONLY(static firm_dbg_module_t *dbg = NULL;)
 
-typedef enum _blocksched_algos_t {
-	BLOCKSCHED_NAIV, BLOCKSCHED_EXTBB, BLOCKSCHED_GREEDY, BLOCKSCHED_ILP
+typedef enum blocksched_algos_t {
+	BLOCKSCHED_NAIV, BLOCKSCHED_GREEDY, BLOCKSCHED_ILP
 } blocksched_algos_t;
 
 static int algo = BLOCKSCHED_GREEDY;
 
 static const lc_opt_enum_int_items_t blockschedalgo_items[] = {
-	{ "naiv",	BLOCKSCHED_NAIV },
-	{ "extbb",	BLOCKSCHED_EXTBB },
+	{ "naiv",   BLOCKSCHED_NAIV },
 	{ "greedy", BLOCKSCHED_GREEDY },
-#ifdef WITH_ILP
 	{ "ilp",    BLOCKSCHED_ILP },
-#endif /* WITH_ILP */
 	{ NULL,     0 }
 };
 
@@ -89,7 +84,7 @@ static lc_opt_enum_int_var_t algo_var = {
 };
 
 static const lc_opt_table_entry_t be_blocksched_options[] = {
-	LC_OPT_ENT_ENUM_INT ("algo", "the block scheduling algorithm", &algo_var),
+	LC_OPT_ENT_ENUM_INT ("blockscheduler", "the block scheduling algorithm", &algo_var),
 	LC_OPT_LAST
 };
 
@@ -102,28 +97,39 @@ static const lc_opt_table_entry_t be_blocksched_options[] = {
  *                            |___/
  */
 
-typedef struct _blocksched_entry_t {
-	ir_node *block;
-	struct _blocksched_entry_t *next;
-	struct _blocksched_entry_t *prev;
-} blocksched_entry_t;
+typedef struct blocksched_entry_t blocksched_entry_t;
+struct blocksched_entry_t {
+	ir_node            *block;
+	blocksched_entry_t *next;
+	blocksched_entry_t *prev;
+};
 
-typedef struct _edge_t {
+typedef struct edge_t edge_t;
+struct edge_t {
 	ir_node *block;             /**< source block */
 	int     pos;                /**< number of cfg predecessor (target) */
 	double  execfreq;           /**< the frequency */
-	int     highest_execfreq;   /**< flag that indicates whether this edge is the edge with the highest
-	                                 execfreq pointing away from this block */
-} edge_t;
+	double  outedge_penalty_freq; /**< for edges leaving the loop this is the
+	                                   penality when we make them a
+	                                   fallthrough. */
+	int     highest_execfreq;   /**< flag that indicates whether this edge is
+	                                 the edge with the highest execfreq pointing
+	                                 away from this block */
+};
 
-typedef struct _blocksched_env_t {
+typedef struct blocksched_env_t blocksched_env_t;
+struct blocksched_env_t {
 	ir_graph       *irg;
 	struct obstack *obst;
-	ir_exec_freq   *execfreqs;
 	edge_t         *edges;
 	pdeq           *worklist;
 	int            blockcount;
-} blocksched_env_t;
+};
+
+static blocksched_entry_t* get_blocksched_entry(const ir_node *block)
+{
+	return (blocksched_entry_t*)get_irn_link(block);
+}
 
 /**
  * Collect cfg frequencies of all edges between blocks.
@@ -131,29 +137,43 @@ typedef struct _blocksched_env_t {
  */
 static void collect_egde_frequency(ir_node *block, void *data)
 {
-	blocksched_env_t   *env = data;
+	blocksched_env_t   *env = (blocksched_env_t*)data;
 	int                arity;
 	edge_t             edge;
 	blocksched_entry_t *entry;
+	ir_loop            *loop;
 
-	entry        = obstack_alloc(env->obst, sizeof(entry[0]));
+	memset(&edge, 0, sizeof(edge));
+
+	entry = OALLOCZ(env->obst, blocksched_entry_t);
 	entry->block = block;
-	entry->next  = NULL;
-	entry->prev  = NULL;
 	set_irn_link(block, entry);
 
+	loop = get_irn_loop(block);
+
 	arity = get_Block_n_cfgpreds(block);
 
 	if (arity == 0) {
+		/* must be the start block (or end-block for endless loops),
+		 * everything else is dead code and should be removed by now */
 		assert(block == get_irg_start_block(env->irg)
 				|| block == get_irg_end_block(env->irg));
-		/* must be the start block (or end-block for endless loops), nothing to
-		 * do here */
+		/* nothing to do here */
 		return;
 	} else if (arity == 1) {
+		ir_node *pred_block = get_Block_cfgpred_block(block, 0);
+		ir_loop *pred_loop  = get_irn_loop(pred_block);
+		float    freq       = (float)get_block_execfreq(block);
+
+		/* is it an edge leaving a loop */
+		if (get_loop_depth(pred_loop) > get_loop_depth(loop)) {
+			float pred_freq = (float)get_block_execfreq(pred_block);
+			edge.outedge_penalty_freq = -(pred_freq - freq);
+		}
+
 		edge.block            = block;
 		edge.pos              = 0;
-		edge.execfreq         = get_block_execfreq(env->execfreqs, block);
+		edge.execfreq         = freq;
 		edge.highest_execfreq = 1;
 		ARR_APP1(edge_t, env->edges, edge);
 	} else {
@@ -166,7 +186,7 @@ static void collect_egde_frequency(ir_node *block, void *data)
 			double  execfreq;
 			ir_node *pred_block = get_Block_cfgpred_block(block, i);
 
-			execfreq = get_block_execfreq(env->execfreqs, pred_block);
+			execfreq = get_block_execfreq(pred_block);
 
 			edge.pos              = i;
 			edge.execfreq         = execfreq;
@@ -179,27 +199,86 @@ static void collect_egde_frequency(ir_node *block, void *data)
 			}
 		}
 
-		if(highest_edge_num >= 0)
+		if (highest_edge_num >= 0)
 			env->edges[highest_edge_num].highest_execfreq = 1;
 	}
 }
 
+static int cmp_edges_base(const edge_t *e1, const edge_t *e2)
+{
+	long nr1 = get_irn_node_nr(e1->block);
+	long nr2 = get_irn_node_nr(e2->block);
+	if (nr1 < nr2) {
+		return 1;
+	} else if (nr1 > nr2) {
+		return -1;
+	} else {
+		if (e1->pos < e2->pos) {
+			return 1;
+		} else if (e1->pos > e2->pos) {
+			return -1;
+		} else {
+			return 0;
+		}
+	}
+}
+
 static int cmp_edges(const void *d1, const void *d2)
 {
-	const edge_t *e1 = d1;
-	const edge_t *e2 = d2;
+	const edge_t *e1 = (const edge_t*)d1;
+	const edge_t *e2 = (const edge_t*)d2;
+	double        freq1 = e1->execfreq;
+	double        freq2 = e2->execfreq;
+	if (freq1 < freq2) {
+		return 1;
+	} else if (freq1 > freq2) {
+		return -1;
+	} else {
+		return cmp_edges_base(e1, e2);
+	}
+}
 
-	return QSORT_CMP(e2->execfreq, e1->execfreq);
+static int cmp_edges_outedge_penalty(const void *d1, const void *d2)
+{
+	const edge_t *e1   = (const edge_t*)d1;
+	const edge_t *e2   = (const edge_t*)d2;
+	double        pen1 = e1->outedge_penalty_freq;
+	double        pen2 = e2->outedge_penalty_freq;
+	if (pen1 > pen2) {
+		return 1;
+	} else if (pen1 < pen2) {
+		return -1;
+	} else {
+		return cmp_edges_base(e1, e2);
+	}
+}
+
+static void clear_loop_links(ir_loop *loop)
+{
+	int i, n;
+
+	set_loop_link(loop, NULL);
+	n = get_loop_n_elements(loop);
+	for (i = 0; i < n; ++i) {
+		loop_element elem = get_loop_element(loop, i);
+		if (*elem.kind == k_ir_loop) {
+			clear_loop_links(elem.son);
+		}
+	}
 }
 
 static void coalesce_blocks(blocksched_env_t *env)
 {
 	int i;
 	int edge_count = ARR_LEN(env->edges);
+	edge_t *edges = env->edges;
+
+	/* sort interblock edges by execution frequency */
+	qsort(edges, ARR_LEN(edges), sizeof(edges[0]), cmp_edges);
 
 	/* run1: only look at jumps */
 	for (i = 0; i < edge_count; ++i) {
-		const edge_t *edge  = &env->edges[i];
+		const edge_t *edge  = &edges[i];
 		ir_node      *block = edge->block;
 		int           pos   = edge->pos;
 		ir_node      *pred_block;
@@ -214,8 +293,8 @@ static void coalesce_blocks(blocksched_env_t *env)
 			continue;
 
 		pred_block = get_Block_cfgpred_block(block, pos);
-		entry      = get_irn_link(block);
-		pred_entry = get_irn_link(pred_block);
+		entry      = get_blocksched_entry(block);
+		pred_entry = get_blocksched_entry(pred_block);
 
 		if (pred_entry->next != NULL || entry->prev != NULL)
 			continue;
@@ -225,38 +304,85 @@ static void coalesce_blocks(blocksched_env_t *env)
 			continue;
 
 		/* schedule the 2 blocks behind each other */
-		DBG((dbg, LEVEL_1, "Coalesce (Jump) %+F -> %+F (%.3g)\n",
+		DB((dbg, LEVEL_1, "Coalesce (Jump) %+F -> %+F (%.3g)\n",
 		           pred_entry->block, entry->block, edge->execfreq));
 		pred_entry->next = entry;
 		entry->prev      = pred_entry;
 	}
 
-	/* run2: remaining edges */
+	/* run2: pick loop fallthroughs */
+	clear_loop_links(get_irg_loop(env->irg));
+
+	qsort(edges, ARR_LEN(edges), sizeof(edges[0]), cmp_edges_outedge_penalty);
 	for (i = 0; i < edge_count; ++i) {
-		const edge_t *edge  = &env->edges[i];
+		const edge_t *edge  = &edges[i];
+		ir_node      *block = edge->block;
+		int           pos   = edge->pos;
+		ir_node      *pred_block;
+		blocksched_entry_t *entry, *pred_entry;
+		ir_loop      *loop;
+		ir_loop      *outer_loop;
+
+		/* already seen all loop outedges? */
+		if (edge->outedge_penalty_freq == 0)
+			break;
+
+		/* the block might have been removed already... */
+		if (is_Bad(get_Block_cfgpred(block, pos)))
+			continue;
+
+		pred_block = get_Block_cfgpred_block(block, pos);
+		entry      = get_blocksched_entry(block);
+		pred_entry = get_blocksched_entry(pred_block);
+
+		if (pred_entry->next != NULL || entry->prev != NULL)
+			continue;
+
+		/* we want at most 1 outedge fallthrough per loop */
+		loop = get_irn_loop(pred_block);
+		if (get_loop_link(loop) != NULL)
+			continue;
+
+		/* schedule the 2 blocks behind each other */
+		DB((dbg, LEVEL_1, "Coalesce (Loop Outedge) %+F -> %+F (%.3g)\n",
+		           pred_entry->block, entry->block, edge->execfreq));
+		pred_entry->next = entry;
+		entry->prev      = pred_entry;
+
+		/* all loops left have an outedge now */
+		outer_loop = get_irn_loop(block);
+		do {
+			/* we set loop link to loop to mark it */
+			set_loop_link(loop, loop);
+			loop = get_loop_outer_loop(loop);
+		} while (loop != outer_loop);
+	}
+
+	/* sort interblock edges by execution frequency */
+	qsort(edges, ARR_LEN(edges), sizeof(edges[0]), cmp_edges);
+
+	/* run3: remaining edges */
+	for (i = 0; i < edge_count; ++i) {
+		const edge_t *edge  = &edges[i];
 		ir_node      *block = edge->block;
 		int           pos   = edge->pos;
 		ir_node      *pred_block;
 		blocksched_entry_t *entry, *pred_entry;
 
 		/* the block might have been removed already... */
-		if (is_Bad(get_Block_cfgpred(block, 0)))
-			continue;
-
-		/* we can't do fallthroughs in backedges */
-		if (is_backedge(block, pos))
+		if (is_Bad(get_Block_cfgpred(block, pos)))
 			continue;
 
 		pred_block = get_Block_cfgpred_block(block, pos);
-		entry      = get_irn_link(block);
-		pred_entry = get_irn_link(pred_block);
+		entry      = get_blocksched_entry(block);
+		pred_entry = get_blocksched_entry(pred_block);
 
 		/* is 1 of the blocks already attached to another block? */
 		if (pred_entry->next != NULL || entry->prev != NULL)
 			continue;
 
 		/* schedule the 2 blocks behind each other */
-		DBG((dbg, LEVEL_1, "Coalesce (CondJump) %+F -> %+F (%.3g)\n",
+		DB((dbg, LEVEL_1, "Coalesce (CondJump) %+F -> %+F (%.3g)\n",
 		           pred_entry->block, entry->block, edge->execfreq));
 		pred_entry->next = entry;
 		entry->prev      = pred_entry;
@@ -268,15 +394,14 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 	ir_node            *block = entry->block;
 	ir_node            *succ  = NULL;
 	blocksched_entry_t *succ_entry;
-	const ir_edge_t    *edge;
-	double             best_succ_execfreq;
+	double              best_succ_execfreq;
 
 	if (irn_visited_else_mark(block))
 		return;
 
 	env->blockcount++;
 
-	DBG((dbg, LEVEL_1, "Pick succ of %+F\n", block));
+	DB((dbg, LEVEL_1, "Pick succ of %+F\n", block));
 
 	/* put all successors into the worklist */
 	foreach_block_succ(block, edge) {
@@ -287,7 +412,7 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 
 		/* we only need to put the first of a series of already connected
 		 * blocks into the worklist */
-		succ_entry = get_irn_link(succ_block);
+		succ_entry = get_blocksched_entry(succ_block);
 		while (succ_entry->prev != NULL) {
 			/* break cycles... */
 			if (succ_entry->prev->block == succ_block) {
@@ -296,12 +421,12 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 				break;
 			}
 			succ_entry = succ_entry->prev;
-		};
+		}
 
 		if (irn_visited(succ_entry->block))
 			continue;
 
-		DBG((dbg, LEVEL_1, "Put %+F into worklist\n", succ_entry->block));
+		DB((dbg, LEVEL_1, "Put %+F into worklist\n", succ_entry->block));
 		pdeq_putr(env->worklist, succ_entry->block);
 	}
 
@@ -310,7 +435,7 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 		return;
 	}
 
-	DBG((dbg, LEVEL_1, "deciding...\n"));
+	DB((dbg, LEVEL_1, "deciding...\n"));
 	best_succ_execfreq = -1;
 
 	/* no successor yet: pick the successor block with the highest execution
@@ -318,16 +443,15 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 
 	foreach_block_succ(block, edge) {
 		ir_node *succ_block = get_edge_src_irn(edge);
-		double  execfreq;
 
 		if (irn_visited(succ_block))
 			continue;
 
-		succ_entry = get_irn_link(succ_block);
+		succ_entry = get_blocksched_entry(succ_block);
 		if (succ_entry->prev != NULL)
 			continue;
 
-		execfreq = get_block_execfreq(env->execfreqs, succ_block);
+		double execfreq = get_block_execfreq(succ_block);
 		if (execfreq > best_succ_execfreq) {
 			best_succ_execfreq = execfreq;
 			succ = succ_block;
@@ -335,18 +459,18 @@ static void pick_block_successor(blocksched_entry_t *entry, blocksched_env_t *en
 	}
 
 	if (succ == NULL) {
-		DBG((dbg, LEVEL_1, "pick from worklist\n"));
+		DB((dbg, LEVEL_1, "pick from worklist\n"));
 
 		do {
 			if (pdeq_empty(env->worklist)) {
-				DBG((dbg, LEVEL_1, "worklist empty\n"));
+				DB((dbg, LEVEL_1, "worklist empty\n"));
 				return;
 			}
-			succ = pdeq_getl(env->worklist);
+			succ = (ir_node*)pdeq_getl(env->worklist);
 		} while (irn_visited(succ));
 	}
 
-	succ_entry       = get_irn_link(succ);
+	succ_entry       = get_blocksched_entry(succ);
 	entry->next      = succ_entry;
 	succ_entry->prev = entry;
 
@@ -357,7 +481,7 @@ static blocksched_entry_t *finish_block_schedule(blocksched_env_t *env)
 {
 	ir_graph           *irg        = env->irg;
 	ir_node            *startblock = get_irg_start_block(irg);
-	blocksched_entry_t *entry      = get_irn_link(startblock);
+	blocksched_entry_t *entry      = get_blocksched_entry(startblock);
 
 	ir_reserve_resources(irg, IR_RESOURCE_IRN_VISITED);
 	inc_irg_visited(irg);
@@ -381,19 +505,19 @@ static ir_node **create_blocksched_array(blocksched_env_t *env, blocksched_entry
 	(void) env;
 
 	block_list = NEW_ARR_D(ir_node *, obst, count);
-	DBG((dbg, LEVEL_1, "Blockschedule:\n"));
+	DB((dbg, LEVEL_1, "Blockschedule:\n"));
 
 	for (entry = first; entry != NULL; entry = entry->next) {
 		assert(i < count);
 		block_list[i++] = entry->block;
-		DBG((dbg, LEVEL_1, "\t%+F\n", entry->block));
+		DB((dbg, LEVEL_1, "\t%+F\n", entry->block));
 	}
 	assert(i == count);
 
 	return block_list;
 }
 
-static ir_node **create_block_schedule_greedy(ir_graph *irg, ir_exec_freq *execfreqs)
+static ir_node **create_block_schedule_greedy(ir_graph *irg)
 {
 	blocksched_env_t   env;
 	struct obstack     obst;
@@ -404,24 +528,23 @@ static ir_node **create_block_schedule_greedy(ir_graph *irg, ir_exec_freq *execf
 
 	env.irg        = irg;
 	env.obst       = &obst;
-	env.execfreqs  = execfreqs;
 	env.edges      = NEW_ARR_F(edge_t, 0);
 	env.worklist   = NULL;
 	env.blockcount = 0;
 
+	assure_loopinfo(irg);
+
 	// collect edge execution frequencies
 	irg_block_walk_graph(irg, collect_egde_frequency, NULL, &env);
 
-	// sort interblock edges by execution frequency
-	qsort(env.edges, ARR_LEN(env.edges), sizeof(env.edges[0]), cmp_edges);
-
 	(void)be_remove_empty_blocks(irg);
 
 	if (algo != BLOCKSCHED_NAIV)
 		coalesce_blocks(&env);
 
 	start_entry = finish_block_schedule(&env);
-	block_list  = create_blocksched_array(&env, start_entry, env.blockcount, get_irg_obstack(irg));
+	block_list  = create_blocksched_array(&env, start_entry, env.blockcount,
+	                                      be_get_be_obst(irg));
 
 	DEL_ARR_F(env.edges);
 	obstack_free(&obst, NULL);
@@ -438,23 +561,22 @@ static ir_node **create_block_schedule_greedy(ir_graph *irg, ir_exec_freq *execf
  *
  */
 
-#ifdef WITH_ILP
-typedef struct _ilp_edge_t {
+typedef struct ilp_edge_t {
 	ir_node *block;   /**< source block */
 	int     pos;      /**< number of cfg predecessor (target) */
 	int     ilpvar;
 } ilp_edge_t;
 
-typedef struct _blocksched_ilp_env_t {
+typedef struct blocksched_ilp_env_t {
 	blocksched_env_t env;
 	ilp_edge_t       *ilpedges;
 	lpp_t            *lpp;
 } blocksched_ilp_env_t;
 
-typedef struct _blocksched_ilp_entry_t {
+typedef struct blocksched_ilp_entry_t {
 	ir_node *block;
-	struct _blocksched_entry_t *next;
-	struct _blocksched_entry_t *prev;
+	struct blocksched_entry_t *next;
+	struct blocksched_entry_t *prev;
 
 	int out_cst;
 } blocksched_ilp_entry_t;
@@ -477,11 +599,10 @@ static int add_ilp_edge(ir_node *block, int pos, double execfreq, blocksched_ilp
 
 static void collect_egde_frequency_ilp(ir_node *block, void *data)
 {
-	blocksched_ilp_env_t *env        = data;
+	blocksched_ilp_env_t *env        = (blocksched_ilp_env_t*)data;
 	ir_graph             *irg        = env->env.irg;
 	ir_node              *startblock = get_irg_start_block(irg);
 	int                  arity;
-	lpp_cst_t            cst;
 	char                 name[64];
 	int                  out_count;
 	blocksched_ilp_entry_t *entry;
@@ -489,11 +610,11 @@ static void collect_egde_frequency_ilp(ir_node *block, void *data)
 	snprintf(name, sizeof(name), "block_out_constr_%ld", get_irn_node_nr(block));
 	out_count = get_irn_n_edges_kind(block, EDGE_KIND_BLOCK);
 
-	entry          = obstack_alloc(env->env.obst, sizeof(entry[0]));
+	entry          = OALLOC(env->env.obst, blocksched_ilp_entry_t);
 	entry->block   = block;
 	entry->next    = NULL;
 	entry->prev    = NULL;
-	entry->out_cst = lpp_add_cst_uniq(env->lpp, name, lpp_greater, out_count - 1);
+	entry->out_cst = lpp_add_cst_uniq(env->lpp, name, lpp_greater_equal, out_count - 1);
 	set_irn_link(block, entry);
 
 	if (block == startblock)
@@ -501,14 +622,15 @@ static void collect_egde_frequency_ilp(ir_node *block, void *data)
 
 	arity = get_irn_arity(block);
 	if (arity == 1) {
-		double execfreq = get_block_execfreq(env->env.execfreqs, block);
+		double execfreq = get_block_execfreq(block);
 		add_ilp_edge(block, 0, execfreq, env);
 	}
 	else {
 		int i;
+		int cst_idx;
 
 		snprintf(name, sizeof(name), "block_in_constr_%ld", get_irn_node_nr(block));
-		cst = lpp_add_cst_uniq(env->lpp, name, lpp_greater, arity - 1);
+		cst_idx = lpp_add_cst_uniq(env->lpp, name, lpp_greater_equal, arity - 1);
 
 		for (i = 0; i < arity; ++i) {
 			double     execfreq;
@@ -516,22 +638,26 @@ static void collect_egde_frequency_ilp(ir_node *block, void *data)
 			ilp_edge_t *edge;
 			ir_node    *pred_block = get_Block_cfgpred_block(block, i);
 
-			execfreq = get_block_execfreq(env->env.execfreqs, pred_block);
+			execfreq = get_block_execfreq(pred_block);
 			edgenum  = add_ilp_edge(block, i, execfreq, env);
 			edge     = &env->ilpedges[edgenum];
-			lpp_set_factor_fast(env->lpp, cst, edge->ilpvar, 1.0);
+			lpp_set_factor_fast(env->lpp, cst_idx, edge->ilpvar, 1.0);
 		}
 	}
 }
 
+static blocksched_ilp_entry_t *get_blocksched_ilp_entry(const ir_node *block)
+{
+	return (blocksched_ilp_entry_t*)get_irn_link(block);
+}
 
 static void coalesce_blocks_ilp(blocksched_ilp_env_t *env)
 {
-	int  i;
-	int  edge_count = ARR_LEN(env->ilpedges);
+	int           edge_count = ARR_LEN(env->ilpedges);
+	int           i;
 
 	/* complete out constraints */
-	for(i = 0; i < edge_count; ++i) {
+	for (i = 0; i < edge_count; ++i) {
 		const ilp_edge_t *edge  = &env->ilpedges[i];
 		ir_node          *block = edge->block;
 		ir_node          *pred;
@@ -542,27 +668,14 @@ static void coalesce_blocks_ilp(blocksched_ilp_env_t *env)
 			continue;
 
 		pred  = get_Block_cfgpred_block(block, edge->pos);
-		entry = get_irn_link(pred);
+		entry = get_blocksched_ilp_entry(pred);
 
-		DBG((dbg, LEVEL_1, "Adding out cst to %+F from %+F,%d\n",
+		DB((dbg, LEVEL_1, "Adding out cst to %+F from %+F,%d\n",
 				  pred, block, edge->pos));
 		lpp_set_factor_fast(env->lpp, entry->out_cst, edge->ilpvar, 1.0);
 	}
 
-#if 0
-	{
-		FILE *f;
-		char fname[256];
-		lpp_dump(env->lpp, "lpp.out");
-		snprintf(fname, sizeof(fname), "lpp_%s.plain", get_irg_dump_name(env->env.irg));
-		f = fopen(fname, "w");
-		lpp_dump_plain(env->lpp, f);
-		fclose(f);
-	}
-#endif
-
-	//lpp_solve_net(env->lpp, main_env->options->ilp_server, main_env->options->ilp_solver);
-	lpp_solve_net(env->lpp, "i44pc52", "cplex");
+	lpp_solve_net(env->lpp, be_options.ilp_server, be_options.ilp_solver);
 	assert(lpp_is_sol_valid(env->lpp));
 
 	/* Apply results to edges */
@@ -583,8 +696,8 @@ static void coalesce_blocks_ilp(blocksched_ilp_env_t *env)
 			continue;
 
 		pred       = get_Block_cfgpred_block(block, edge->pos);
-		entry      = get_irn_link(block);
-		pred_entry = get_irn_link(pred);
+		entry      = get_blocksched_entry(block);
+		pred_entry = get_blocksched_entry(pred);
 
 		assert(entry->prev == NULL && pred_entry->next == NULL);
 		entry->prev      = pred_entry;
@@ -592,7 +705,7 @@ static void coalesce_blocks_ilp(blocksched_ilp_env_t *env)
 	}
 }
 
-static ir_node **create_block_schedule_ilp(ir_graph *irg, ir_exec_freq *execfreqs)
+static ir_node **create_block_schedule_ilp(ir_graph *irg)
 {
 	blocksched_ilp_env_t env;
 	struct obstack       obst;
@@ -603,12 +716,11 @@ static ir_node **create_block_schedule_ilp(ir_graph *irg, ir_exec_freq *execfreq
 
 	env.env.irg        = irg;
 	env.env.obst       = &obst;
-	env.env.execfreqs  = execfreqs;
 	env.env.worklist   = NULL;
 	env.env.blockcount = 0;
 	env.ilpedges       = NEW_ARR_F(ilp_edge_t, 0);
 
-	env.lpp = new_lpp("blockschedule", lpp_minimize);
+	env.lpp = lpp_new("blockschedule", lpp_minimize);
 	lpp_set_time_limit(env.lpp, 20);
 	lpp_set_log(env.lpp, stdout);
 
@@ -618,114 +730,16 @@ static ir_node **create_block_schedule_ilp(ir_graph *irg, ir_exec_freq *execfreq
 	coalesce_blocks_ilp(&env);
 
 	start_entry = finish_block_schedule(&env.env);
-	block_list  = create_blocksched_array(&env.env, start_entry, env.env.blockcount, get_irg_obstack(irg));
+	block_list  = create_blocksched_array(&env.env, start_entry,
+	                                      env.env.blockcount,
+	                                      be_get_be_obst(irg));
 
 	DEL_ARR_F(env.ilpedges);
-	free_lpp(env.lpp);
+	lpp_free(env.lpp);
 	obstack_free(&obst, NULL);
 
 	return block_list;
 }
-#endif /* WITH_ILP */
-
-/*
- *  _____      _   ____  ____
- * | ____|_  _| |_| __ )| __ )
- * |  _| \ \/ / __|  _ \|  _ \
- * | |___ >  <| |_| |_) | |_) |
- * |_____/_/\_\\__|____/|____/
- *
- */
-
-/** A simple forward single linked list. */
-typedef struct {
-	ir_node  *start;   /**< start of the list */
-	ir_node  *end;     /**< last block in the list */
-	unsigned n_blks;  /**< number of blocks in the list */
-} anchor;
-
-static void add_block(anchor *list, ir_node *block) {
-	if (list->start == NULL) {
-		list->start = block;
-		list->end   = block;
-	} else {
-		set_irn_link(list->end, block);
-		list->end = block;
-	}
-
-	list->n_blks++;
-}
-
-static void create_block_list(ir_node *leader_block, anchor *list) {
-	int             i;
-	const ir_edge_t *edge;
-	ir_node         *block = NULL;
-	ir_extblk       *extbb = get_Block_extbb(leader_block);
-
-	if (extbb_visited(extbb))
-		return;
-	mark_extbb_visited(extbb);
-
-	for (i = 0; i < get_extbb_n_blocks(extbb); ++i) {
-		block = get_extbb_block(extbb, i);
-		add_block(list, block);
-	}
-
-	assert(block != NULL);
-
-	/* pick successor extbbs */
-	foreach_block_succ(block, edge) {
-		ir_node *succ = get_edge_src_irn(edge);
-		create_block_list(succ, list);
-	}
-
-	for (i = 0; i < get_extbb_n_blocks(extbb) - 1; ++i) {
-		block = get_extbb_block(extbb, i);
-
-		foreach_block_succ(block, edge) {
-			ir_node *succ = get_edge_src_irn(edge);
-			create_block_list(succ, list);
-		}
-	}
-}
-
-void compute_extbb_execfreqs(ir_graph *irg, ir_exec_freq *execfreqs);
-
-/*
- * Calculates a block schedule. The schedule is stored as a linked
- * list starting at the start_block of the irg.
- */
-static ir_node **create_extbb_block_schedule(ir_graph *irg, ir_exec_freq *execfreqs)
-{
-	anchor list;
-	ir_node **blk_list, *b, *n;
-	unsigned i;
-
-	/* schedule extended basic blocks */
-	compute_extbb_execfreqs(irg, execfreqs);
-	//compute_extbb(irg);
-
-	list.start  = NULL;
-	list.end    = NULL;
-	list.n_blks = 0;
-
-	ir_reserve_resources(irg, IR_RESOURCE_IRN_VISITED | IR_RESOURCE_IRN_LINK);
-	inc_irg_block_visited(irg);
-
-	create_block_list(get_irg_start_block(irg), &list);
-
-	/** create an array, so we can go forward and backward */
-	blk_list = NEW_ARR_D(ir_node *, irg->obst,list.n_blks);
-
-	for (i = 0, b = list.start; b; b = n, ++i) {
-		n = get_irn_link(b);
-		blk_list[i] = b;
-	}
-
-	ir_free_resources(irg, IR_RESOURCE_IRN_VISITED | IR_RESOURCE_IRN_LINK);
-
-	return blk_list;
-}
 
 /*
  *  __  __       _
@@ -735,32 +749,25 @@ static ir_node **create_extbb_block_schedule(ir_graph *irg, ir_exec_freq *execfr
  * |_|  |_|\__,_|_|_| |_|
  *
  */
+BE_REGISTER_MODULE_CONSTRUCTOR(be_init_blocksched)
 void be_init_blocksched(void)
 {
 	lc_opt_entry_t *be_grp = lc_opt_get_grp(firm_opt_get_root(), "be");
-	lc_opt_entry_t *blocksched_grp = lc_opt_get_grp(be_grp, "blocksched");
 
-	lc_opt_add_table(blocksched_grp, be_blocksched_options);
+	lc_opt_add_table(be_grp, be_blocksched_options);
 
 	FIRM_DBG_REGISTER(dbg, "firm.be.blocksched");
 }
 
-BE_REGISTER_MODULE_CONSTRUCTOR(be_init_blocksched);
-
-ir_node **be_create_block_schedule(ir_graph *irg, ir_exec_freq *execfreqs)
+ir_node **be_create_block_schedule(ir_graph *irg)
 {
-	switch(algo) {
+	switch (algo) {
 	case BLOCKSCHED_GREEDY:
 	case BLOCKSCHED_NAIV:
-		return create_block_schedule_greedy(irg, execfreqs);
-	case BLOCKSCHED_EXTBB:
-		return create_extbb_block_schedule(irg, execfreqs);
-#ifdef WITH_ILP
+		return create_block_schedule_greedy(irg);
 	case BLOCKSCHED_ILP:
-		return create_block_schedule_ilp(irg, execfreqs);
-#endif /* WITH_ILP */
+		return create_block_schedule_ilp(irg);
 	}
 
 	panic("unknown blocksched algo");
-	return NULL;
 }
