@@ -22,13 +22,12 @@
  * @brief       Block schedule calculation.
  * @author      Matthias Braun, Christoph Mallon
  * @date        27.09.2006
- * @version     $Id: beblocksched.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_BE_BEBLOCKSCHED_H
 #define FIRM_BE_BEBLOCKSCHED_H
 
 #include "firm_types.h"
 
-ir_node **be_create_block_schedule(ir_graph *irg, ir_exec_freq *execfreqs);
+ir_node **be_create_block_schedule(ir_graph *irg);
 
-#endif /* FIRM_BE_BEBLOCKSCHED_H */
+#endif
