@@ -22,7 +22,6 @@
  * @brief       First simple copy minimization heuristics.
  * @author      Daniel Grund
  * @date        12.04.2005
- * @version     $Id: becopyheur.c 25803 2009-04-06 22:52:15Z beck $
  *
  * Heuristic for minimizing copies using a queue which holds 'qnodes' not yet
  * examined. A qnode has a 'target color', nodes out of the opt unit and
@@ -42,10 +41,14 @@
 #include "becopyopt_t.h"
 #include "becopystat.h"
 #include "beintlive_t.h"
-#include "beirg_t.h"
+#include "beirg.h"
+#include "bemodule.h"
 
 DEBUG_ONLY(static firm_dbg_module_t *dbg = NULL;)
 
+/** Defines an invalid register index. */
+#define NO_COLOR (-1)
+
 #define SEARCH_FREE_COLORS
 
 #define SLOTS_PINNED_GLOBAL 64
@@ -58,7 +61,7 @@ DEBUG_ONLY(static firm_dbg_module_t *dbg = NULL;)
 /**
  * Modeling additional conflicts between nodes. NOT live range interference
  */
-typedef struct _conflict_t {
+typedef struct conflict_t {
 	const ir_node *n1, *n2;
 } conflict_t;
 
@@ -66,7 +69,7 @@ typedef struct _conflict_t {
  * If an irn is changed, the changes first get stored in a node_stat_t,
  * to allow undo of changes (=drop new data) in case of conflicts.
  */
-typedef struct _node_stat_t {
+typedef struct node_stat_t {
 	ir_node *irn;
 	int     new_color;
 	int     pinned_local :1;
@@ -75,7 +78,7 @@ typedef struct _node_stat_t {
 /**
  * Represents a node in the optimization queue.
  */
-typedef struct _qnode_t {
+typedef struct qnode_t {
 	struct list_head queue;            /**< chaining of unit_t->queue */
 	const unit_t     *ou;              /**< the opt unit this node belongs to */
 	int              color;            /**< target color */
@@ -86,19 +89,22 @@ typedef struct _qnode_t {
 	set              *changed_nodes;   /**< contains node_stat_t's. */
 } qnode_t;
 
-static pset *pinned_global;			/**< optimized nodes should not be altered any more */
+static pset *pinned_global;  /**< optimized nodes should not be altered any more */
 
 static inline int nodes_interfere(const be_chordal_env_t *env, const ir_node *a, const ir_node *b)
 {
 	if (env->ifg)
 		return be_ifg_connected(env->ifg, a, b);
-	else
-		return values_interfere(env->birg, a, b);
+	else {
+		be_lv_t *lv = be_get_irg_liveness(env->irg);
+		return be_values_interfere(lv, a, b);
+	}
 }
 
-static int set_cmp_conflict_t(const void *x, const void *y, size_t size) {
-	const conflict_t *xx = x;
-	const conflict_t *yy = y;
+static int set_cmp_conflict_t(const void *x, const void *y, size_t size)
+{
+	const conflict_t *xx = (const conflict_t*)x;
+	const conflict_t *yy = (const conflict_t*)y;
 	(void) size;
 
 	return xx->n1 != yy->n1 || xx->n2 != yy->n2;
@@ -108,7 +114,8 @@ static int set_cmp_conflict_t(const void *x, const void *y, size_t size) {
  * If a local pinned conflict occurs, a new edge in the conflict graph is added.
  * The next maximum independent set build, will regard it.
  */
-static inline void qnode_add_conflict(const qnode_t *qn, const ir_node *n1, const ir_node *n2) {
+static inline void qnode_add_conflict(const qnode_t *qn, const ir_node *n1, const ir_node *n2)
+{
 	conflict_t c;
 	DBG((dbg, LEVEL_4, "\t      %+F -- %+F\n", n1, n2));
 
@@ -119,13 +126,14 @@ static inline void qnode_add_conflict(const qnode_t *qn, const ir_node *n1, cons
 		c.n1 = n2;
 		c.n2 = n1;
 	}
-	set_insert(qn->conflicts, &c, sizeof(c), HASH_CONFLICT(c));
+	(void)set_insert(conflict_t, qn->conflicts, &c, sizeof(c), HASH_CONFLICT(c));
 }
 
 /**
  * Checks if two nodes are in a conflict.
  */
-static inline int qnode_are_conflicting(const qnode_t *qn, const ir_node *n1, const ir_node *n2) {
+static inline int qnode_are_conflicting(const qnode_t *qn, const ir_node *n1, const ir_node *n2)
+{
 	conflict_t c;
 	/* search for live range interference */
 	if (n1!=n2 && nodes_interfere(qn->ou->co->cenv, n1, n2))
@@ -138,10 +146,11 @@ static inline int qnode_are_conflicting(const qnode_t *qn, const ir_node *n1, co
 		c.n1 = n2;
 		c.n2 = n1;
 	}
-	return set_find(qn->conflicts, &c, sizeof(c), HASH_CONFLICT(c)) != 0;
+	return set_find(conflict_t, qn->conflicts, &c, sizeof(c), HASH_CONFLICT(c)) != 0;
 }
 
-static int set_cmp_node_stat_t(const void *x, const void *y, size_t size) {
+static int set_cmp_node_stat_t(const void *x, const void *y, size_t size)
+{
 	(void) size;
 	return ((const node_stat_t*)x)->irn != ((const node_stat_t*)y)->irn;
 }
@@ -149,28 +158,31 @@ static int set_cmp_node_stat_t(const void *x, const void *y, size_t size) {
 /**
  * Finds a node status entry of a node if existent. Otherwise return NULL
  */
-static inline const node_stat_t *qnode_find_node(const qnode_t *qn, ir_node *irn) {
+static inline const node_stat_t *qnode_find_node(const qnode_t *qn, ir_node *irn)
+{
 	node_stat_t find;
 	find.irn = irn;
-	return set_find(qn->changed_nodes, &find, sizeof(find), hash_irn(irn));
+	return set_find(node_stat_t, qn->changed_nodes, &find, sizeof(find), hash_irn(irn));
 }
 
 /**
  * Finds a node status entry of a node if existent. Otherwise it will return
  * an initialized new entry for this node.
  */
-static inline node_stat_t *qnode_find_or_insert_node(const qnode_t *qn, ir_node *irn) {
+static inline node_stat_t *qnode_find_or_insert_node(const qnode_t *qn, ir_node *irn)
+{
 	node_stat_t find;
 	find.irn = irn;
 	find.new_color = NO_COLOR;
 	find.pinned_local = 0;
-	return set_insert(qn->changed_nodes, &find, sizeof(find), hash_irn(irn));
+	return set_insert(node_stat_t, qn->changed_nodes, &find, sizeof(find), hash_irn(irn));
 }
 
 /**
  * Returns the virtual color of a node if set before, else returns the real color.
  */
-static inline int qnode_get_new_color(const qnode_t *qn, ir_node *irn) {
+static inline int qnode_get_new_color(const qnode_t *qn, ir_node *irn)
+{
 	const node_stat_t *found = qnode_find_node(qn, irn);
 	if (found)
 		return found->new_color;
@@ -181,7 +193,8 @@ static inline int qnode_get_new_color(const qnode_t *qn, ir_node *irn) {
 /**
  * Sets the virtual color of a node.
  */
-static inline void qnode_set_new_color(const qnode_t *qn, ir_node *irn, int color) {
+static inline void qnode_set_new_color(const qnode_t *qn, ir_node *irn, int color)
+{
 	node_stat_t *found = qnode_find_or_insert_node(qn, irn);
 	found->new_color = color;
 	DBG((dbg, LEVEL_3, "\t      col(%+F) := %d\n", irn, color));
@@ -192,7 +205,8 @@ static inline void qnode_set_new_color(const qnode_t *qn, ir_node *irn, int colo
  * to the same optimization unit and has been optimized before the current
  * processed node.
  */
-static inline int qnode_is_pinned_local(const qnode_t *qn, ir_node *irn) {
+static inline int qnode_is_pinned_local(const qnode_t *qn, ir_node *irn)
+{
 	const node_stat_t *found = qnode_find_node(qn, irn);
 	if (found)
 		return found->pinned_local;
@@ -204,7 +218,8 @@ static inline int qnode_is_pinned_local(const qnode_t *qn, ir_node *irn) {
  * Local-pins a node, so optimizations of further nodes of the same opt unit
  * can handle situations in which a color change would undo prior optimizations.
  */
-static inline void qnode_pin_local(const qnode_t *qn, ir_node *irn) {
+static inline void qnode_pin_local(const qnode_t *qn, ir_node *irn)
+{
 	node_stat_t *found = qnode_find_or_insert_node(qn, irn);
 	found->pinned_local = 1;
 	if (found->new_color == NO_COLOR)
@@ -236,15 +251,16 @@ static inline void qnode_pin_local(const qnode_t *qn, ir_node *irn) {
  *         Else the first conflicting ir_node encountered is returned.
  *
  */
-static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const ir_node *trigger) {
+static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const ir_node *trigger)
+{
 	copy_opt_t *co = qn->ou->co;
 	const be_chordal_env_t *chordal_env = co->cenv;
 	const arch_register_class_t *cls = co->cls;
 	int irn_col = qnode_get_new_color(qn, irn);
 	ir_node *sub_res, *curr;
 	be_ifg_t *ifg = chordal_env->ifg;
-	void *iter = be_ifg_neighbours_iter_alloca(ifg);
-
+	neighbours_iter_t iter;
+	const arch_register_req_t *req;
 
 	DBG((dbg, LEVEL_3, "\t    %+F \tcaused col(%+F) \t%2d --> %2d\n", trigger, irn, irn_col, col));
 
@@ -260,22 +276,20 @@ static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const
 		return irn;
 	}
 
+	req = arch_get_irn_register_req(irn);
 #ifdef SEARCH_FREE_COLORS
 	/* If we resolve conflicts (recursive calls) we can use any unused color.
 	 * In case of the first call @p col must be used.
 	 */
 	if (irn != trigger) {
 		bitset_t *free_cols = bitset_alloca(cls->n_regs);
-		const arch_register_req_t *req;
 		ir_node *curr;
 		int free_col;
 
 		/* Get all possible colors */
-		bitset_copy(free_cols, co->cenv->ignore_colors);
-		bitset_flip_all(free_cols);
+		bitset_copy(free_cols, co->cenv->allocatable_regs);
 
 		/* Exclude colors not assignable to the irn */
-		req = arch_get_register_req_out(irn);
 		if (arch_register_req_is(req, limited)) {
 			bitset_t *limited = bitset_alloca(cls->n_regs);
 			rbitset_copy_to_bitset(req->limited, limited);
@@ -286,7 +300,7 @@ static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const
 		bitset_clear(free_cols, irn_col);
 
 		/* Exclude all colors used by adjacent nodes */
-		be_ifg_foreach_neighbour(ifg, iter, irn, curr)
+		be_ifg_foreach_neighbour(ifg, &iter, irn, curr)
 			bitset_clear(free_cols, qnode_get_new_color(qn, curr));
 
 		free_col = bitset_next_set(free_cols, 0);
@@ -299,7 +313,7 @@ static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const
 #endif /* SEARCH_FREE_COLORS */
 
 	/* If target color is not allocatable changing color is impossible */
-	if (!arch_reg_out_is_allocatable(irn, arch_register_for_index(cls, col))) {
+	if (!arch_reg_is_allocatable(req, arch_register_for_index(cls, col))) {
 		DBG((dbg, LEVEL_3, "\t      %+F impossible\n", irn));
 		return CHANGE_IMPOSSIBLE;
 	}
@@ -308,12 +322,12 @@ static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const
 	 * If we arrive here changing color may be possible, but there may be conflicts.
 	 * Try to color all conflicting nodes 'curr' with the color of the irn itself.
 	 */
-	be_ifg_foreach_neighbour(ifg, iter, irn, curr) {
+	be_ifg_foreach_neighbour(ifg, &iter, irn, curr) {
 		DBG((dbg, LEVEL_3, "\t      Confl %+F(%d)\n", curr, qnode_get_new_color(qn, curr)));
 		if (qnode_get_new_color(qn, curr) == col && curr != trigger) {
 			sub_res = qnode_color_irn(qn, curr, irn_col, irn);
 			if (sub_res != CHANGE_SAVE) {
-				be_ifg_neighbours_break(ifg, iter);
+				be_ifg_neighbours_break(&iter);
 				return sub_res;
 			}
 		}
@@ -334,7 +348,8 @@ static ir_node *qnode_color_irn(const qnode_t *qn, ir_node *irn, int col, const
  * @returns 1 iff all members colors could be set
  *          0 else
  */
-static int qnode_try_color(const qnode_t *qn) {
+static int qnode_try_color(const qnode_t *qn)
+{
 	int i;
 	for (i=0; i<qn->mis_size; ++i) {
 		ir_node *test_node, *confl_node;
@@ -379,11 +394,11 @@ static int qnode_try_color(const qnode_t *qn) {
  * Determines a maximum weighted independent set with respect to
  * the interference and conflict edges of all nodes in a qnode.
  */
-static inline void qnode_max_ind_set(qnode_t *qn, const unit_t *ou) {
+static inline void qnode_max_ind_set(qnode_t *qn, const unit_t *ou)
+{
 	ir_node **safe, **unsafe;
 	int i, o, safe_count, safe_costs, unsafe_count, *unsafe_costs;
 	bitset_t *curr, *best;
-	bitset_pos_t pos;
 	int next, curr_weight, best_weight = 0;
 
 	/* assign the nodes into two groups.
@@ -396,9 +411,9 @@ static inline void qnode_max_ind_set(qnode_t *qn, const unit_t *ou) {
 	unsafe       = ALLOCAN(ir_node*, ou->node_count - 1);
 	unsafe_costs = ALLOCAN(int,      ou->node_count - 1);
 	unsafe_count = 0;
-	for(i=1; i<ou->node_count; ++i) {
+	for (i=1; i<ou->node_count; ++i) {
 		int is_safe = 1;
-		for(o=1; o<ou->node_count; ++o) {
+		for (o=1; o<ou->node_count; ++o) {
 			if (qnode_are_conflicting(qn, ou->nodes[i], ou->nodes[o])) {
 				if (i!=o) {
 					unsafe_costs[unsafe_count] = ou->costs[i];
@@ -446,7 +461,7 @@ static inline void qnode_max_ind_set(qnode_t *qn, const unit_t *ou) {
 							goto no_stable_set;
 
 			/* if we arrive here, we have a stable set */
-			/* compute the weigth of the stable set*/
+			/* compute the weight of the stable set*/
 			curr_weight = 0;
 			bitset_foreach(curr, pos)
 				curr_weight += unsafe_costs[pos];
@@ -463,7 +478,7 @@ no_stable_set:
 	}
 
 	/* transfer the best set into the qn */
-	qn->mis_size = 1+safe_count+bitset_popcnt(best);
+	qn->mis_size = 1+safe_count+bitset_popcount(best);
 	qn->mis_costs = safe_costs+best_weight;
 	qn->mis[0] = ou->nodes[0]; /* the root is always in a max stable set */
 	next = 1;
@@ -476,7 +491,8 @@ no_stable_set:
 /**
  * Creates a new qnode
  */
-static inline qnode_t *new_qnode(const unit_t *ou, int color) {
+static inline qnode_t *new_qnode(const unit_t *ou, int color)
+{
 	qnode_t *qn = XMALLOC(qnode_t);
 	qn->ou            = ou;
 	qn->color         = color;
@@ -489,7 +505,8 @@ static inline qnode_t *new_qnode(const unit_t *ou, int color) {
 /**
  * Frees space used by a queue node
  */
-static inline void free_qnode(qnode_t *qn) {
+static inline void free_qnode(qnode_t *qn)
+{
 	del_set(qn->conflicts);
 	del_set(qn->changed_nodes);
 	xfree(qn->mis);
@@ -500,7 +517,8 @@ static inline void free_qnode(qnode_t *qn) {
  * Inserts a qnode in the sorted queue of the optimization unit. Queue is
  * ordered by field 'size' (the size of the mis) in decreasing order.
  */
-static inline void ou_insert_qnode(unit_t *ou, qnode_t *qn) {
+static inline void ou_insert_qnode(unit_t *ou, qnode_t *qn)
+{
 	struct list_head *lh;
 
 	if (qnode_are_conflicting(qn, ou->nodes[0], ou->nodes[0])) {
@@ -529,14 +547,14 @@ static inline void ou_insert_qnode(unit_t *ou, qnode_t *qn) {
  * case for approximately 80% of all phi classes and 100% of register constrained
  * nodes. (All other phi classes are reduced to this case.)
  */
-static void ou_optimize(unit_t *ou) {
-	qnode_t                     *curr = NULL;
-	qnode_t                     *tmp;
-	const arch_register_req_t   *req;
-	bitset_t const*              ignore;
-	bitset_pos_t                 n_regs;
-	bitset_pos_t                 idx;
-	int                          i;
+static void ou_optimize(unit_t *ou)
+{
+	qnode_t                   *curr = NULL;
+	const arch_register_req_t *req;
+	bitset_t const*            allocatable_regs;
+	unsigned                   n_regs;
+	unsigned                   idx;
+	int                        i;
 
 	DBG((dbg, LEVEL_1, "\tOptimizing unit:\n"));
 	for (i=0; i<ou->node_count; ++i)
@@ -545,14 +563,14 @@ static void ou_optimize(unit_t *ou) {
 	/* init queue */
 	INIT_LIST_HEAD(&ou->queue);
 
-	req     = arch_get_register_req_out(ou->nodes[0]);
-	ignore  = ou->co->cenv->ignore_colors;
-	n_regs  = req->cls->n_regs;
+	req              = arch_get_irn_register_req(ou->nodes[0]);
+	allocatable_regs = ou->co->cenv->allocatable_regs;
+	n_regs           = req->cls->n_regs;
 	if (arch_register_req_is(req, limited)) {
-		rawbs_base_t const* limited = req->limited;
+		unsigned const* limited = req->limited;
 
 		for (idx = 0; idx != n_regs; ++idx) {
-			if (bitset_is_set(ignore, idx))
+			if (!bitset_is_set(allocatable_regs, idx))
 				continue;
 			if (!rbitset_is_set(limited, idx))
 				continue;
@@ -561,7 +579,7 @@ static void ou_optimize(unit_t *ou) {
 		}
 	} else {
 		for (idx = 0; idx != n_regs; ++idx) {
-			if (bitset_is_set(ignore, idx))
+			if (!bitset_is_set(allocatable_regs, idx))
 				continue;
 
 			ou_insert_qnode(ou, new_qnode(ou, idx));
@@ -588,7 +606,6 @@ static void ou_optimize(unit_t *ou) {
 
 	/* apply the best found qnode */
 	if (curr->mis_size >= 2) {
-		node_stat_t *ns;
 		int root_col = qnode_get_new_color(curr, ou->nodes[0]);
 		DBG((dbg, LEVEL_1, "\t  Best color: %d  Costs: %d << %d << %d\n", curr->color, ou->min_nodes_costs, ou->all_nodes_costs - curr->mis_costs, ou->all_nodes_costs));
 		/* globally pin root and all args which have the same color */
@@ -601,11 +618,11 @@ static void ou_optimize(unit_t *ou) {
 		}
 
 		/* set color of all changed nodes */
-		for (ns = set_first(curr->changed_nodes); ns; ns = set_next(curr->changed_nodes)) {
+		foreach_set(curr->changed_nodes, node_stat_t, ns) {
 			/* NO_COLOR is possible, if we had an undo */
 			if (ns->new_color != NO_COLOR) {
 				DBG((dbg, LEVEL_1, "\t    color(%+F) := %d\n", ns->irn, ns->new_color));
-				set_irn_col(ou->co, ns->irn, ns->new_color);
+				set_irn_col(ou->co->cls, ns->irn, ns->new_color);
 			}
 		}
 	}
@@ -616,10 +633,12 @@ static void ou_optimize(unit_t *ou) {
 		free_qnode(curr);
 }
 
-int co_solve_heuristic(copy_opt_t *co) {
-	unit_t *curr;
-	FIRM_DBG_REGISTER(dbg, "ir.be.copyoptheur");
-
+/**
+ * Solves the problem using a heuristic approach
+ * Uses the OU data structure
+ */
+int co_solve_heuristic(copy_opt_t *co)
+{
 	ASSERT_OU_AVAIL(co);
 
 	pinned_global = pset_new_ptr(SLOTS_PINNED_GLOBAL);
@@ -630,3 +649,14 @@ int co_solve_heuristic(copy_opt_t *co) {
 	del_pset(pinned_global);
 	return 0;
 }
+
+BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copyheur)
+void be_init_copyheur(void)
+{
+	static co_algo_info copyheur = {
+		co_solve_heuristic, 0
+	};
+
+	be_register_copyopt("heur1", &copyheur);
+	FIRM_DBG_REGISTER(dbg, "ir.be.copyoptheur");
+}
