@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1995-2008 University of Karlsruhe.  All right reserved.
+ * Copyright (C) 1995-2011 University of Karlsruhe.  All right reserved.
  *
  * This file is part of libFirm.
  *
@@ -22,7 +22,6 @@
  * @brief       More experiments on coalescing.
  * @author      Sebastian Hack
  * @date        14.04.2006
- * @version     $Id: becopyheur2.c 23142 2008-10-23 18:02:53Z matze $
  */
 #include "config.h"
 
@@ -40,19 +39,20 @@
 #include "debug.h"
 #include "bitfiddle.h"
 
-#include "irphase_t.h"
 #include "irgraph_t.h"
 #include "irnode_t.h"
 #include "irprintf.h"
+#include "util.h"
 #include "irtools.h"
+#include "irnodemap.h"
 
 #include "bemodule.h"
 #include "beabi.h"
-#include "benode_t.h"
+#include "benode.h"
 #include "becopyopt.h"
 #include "becopyopt_t.h"
 #include "bechordal_t.h"
-#include "beirg_t.h"
+#include "beirg.h"
 
 #define DUMP_BEFORE 1
 #define DUMP_AFTER  2
@@ -84,18 +84,6 @@ static const lc_opt_table_entry_t options[] = {
 	LC_OPT_LAST
 };
 
-void be_init_copyheur2(void)
-{
-	lc_opt_entry_t *be_grp = lc_opt_get_grp(firm_opt_get_root(), "be");
-	lc_opt_entry_t *ra_grp = lc_opt_get_grp(be_grp, "ra");
-	lc_opt_entry_t *chordal_grp = lc_opt_get_grp(ra_grp, "chordal");
-	lc_opt_entry_t *co2_grp = lc_opt_get_grp(chordal_grp, "co2");
-
-	lc_opt_add_table(co2_grp, options);
-}
-
-BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copyheur2);
-
 /*
   ____  _             _
  / ___|| |_ __ _ _ __| |_
@@ -107,13 +95,11 @@ BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copyheur2);
 
 #define INFEASIBLE(cost) ((cost) == INT_MAX)
 
-static be_ifg_dump_dot_cb_t ifg_dot_cb;
-
 typedef unsigned col_t;
 
-typedef struct _co2_irn_t       co2_irn_t;
-typedef struct _co2_cloud_t     co2_cloud_t;
-typedef struct _co2_cloud_irn_t co2_cloud_irn_t;
+typedef struct co2_irn_t       co2_irn_t;
+typedef struct co2_cloud_t     co2_cloud_t;
+typedef struct co2_cloud_irn_t co2_cloud_irn_t;
 
 typedef struct {
 	col_t col;
@@ -121,9 +107,10 @@ typedef struct {
 } col_cost_pair_t;
 
 typedef struct {
-	ir_phase     ph;
+	ir_nodemap     map;
+	struct obstack obst;
 	copy_opt_t *co;
-	bitset_t   *ignore_regs;
+	bitset_t   *allocatable_regs;
 	co2_irn_t  *touched;
 	int         visited;
 	int         n_regs;
@@ -131,13 +118,13 @@ typedef struct {
 	DEBUG_ONLY(firm_dbg_module_t *dbg;)
 } co2_t;
 
-struct _co2_irn_t {
+struct co2_irn_t {
 	const ir_node   *irn;
 	affinity_node_t *aff;
 	co2_irn_t       *touched_next;
 	col_t            tmp_col;
 	col_t            orig_col;
-	int				 last_color_change;
+	int              last_color_change;
 	bitset_t        *adm_cache;
 	unsigned         fixed          : 1;
 	unsigned         tmp_fixed      : 1;
@@ -145,8 +132,8 @@ struct _co2_irn_t {
 	struct list_head changed_list;
 };
 
-struct _co2_cloud_irn_t {
-	struct _co2_irn_t  inh;
+struct co2_cloud_irn_t {
+	struct co2_irn_t   inh;
 	co2_cloud_t       *cloud;
 	int                visited;
 	int                index;
@@ -163,7 +150,7 @@ struct _co2_cloud_irn_t {
 	struct list_head   mst_list;
 };
 
-struct _co2_cloud_t {
+struct co2_cloud_t {
 	co2_t            *env;
 	struct obstack    obst;
 	int               costs;
@@ -173,7 +160,7 @@ struct _co2_cloud_t {
 	int               n_memb;
 	int               n_constr;
 	int               max_degree;
-	int			      ticks;
+	int               ticks;
 	double            freedom;
 	co2_cloud_irn_t  *master;
 	co2_cloud_irn_t  *mst_root;
@@ -189,30 +176,42 @@ typedef struct {
 
 #define FRONT_BASE(ci,col)  ((ci)->fronts + col * (ci)->mst_n_childs)
 
-#define get_co2_irn(co2, irn)         ((co2_irn_t *)       phase_get_or_set_irn_data(&co2->ph, irn))
-#define get_co2_cloud_irn(co2, irn)   ((co2_cloud_irn_t *) phase_get_or_set_irn_data(&co2->ph, irn))
-
-static void *co2_irn_init(ir_phase *ph, const ir_node *irn, void *data)
+static co2_irn_t *get_co2_irn(co2_t *env, const ir_node *node)
 {
-	co2_t *env         = (co2_t *) ph;
-	affinity_node_t *a = get_affinity_info(env->co, irn);
-	size_t size        = a ? sizeof(co2_cloud_irn_t) : sizeof(co2_irn_t);
-	co2_irn_t *ci      = data ? data : phase_alloc(ph, size);
+	co2_irn_t *ci = ir_nodemap_get(co2_irn_t, &env->map, node);
+	if (ci == NULL) {
+		ci = OALLOCZ(&env->obst, co2_irn_t);
 
-	memset(ci, 0, size);
-	INIT_LIST_HEAD(&ci->changed_list);
-	ci->touched_next = env->touched;
-	ci->orig_col     = get_irn_col(irn);
-	env->touched     = ci;
-	ci->irn          = irn;
-	ci->aff          = a;
+		INIT_LIST_HEAD(&ci->changed_list);
+		ci->touched_next = env->touched;
+		ci->orig_col     = get_irn_col(node);
+		env->touched     = ci;
+		ci->irn          = node;
+		ci->aff          = NULL;
 
-	if(a) {
-		co2_cloud_irn_t *cci = (co2_cloud_irn_t *) ci;
-		INIT_LIST_HEAD(&cci->cloud_list);
-		cci->mst_parent   = cci;
+		ir_nodemap_insert(&env->map, node, ci);
 	}
+	return ci;
+}
 
+static co2_cloud_irn_t *get_co2_cloud_irn(co2_t *env, const ir_node *node)
+{
+	co2_cloud_irn_t *ci = ir_nodemap_get(co2_cloud_irn_t, &env->map, node);
+	if (ci == NULL) {
+		ci = OALLOCZ(&env->obst, co2_cloud_irn_t);
+
+		INIT_LIST_HEAD(&ci->inh.changed_list);
+		ci->inh.touched_next = env->touched;
+		ci->inh.orig_col     = get_irn_col(node);
+		env->touched         = &ci->inh;
+		ci->inh.irn          = node;
+		ci->inh.aff          = get_affinity_info(env->co, node);
+
+		INIT_LIST_HEAD(&ci->cloud_list);
+		ci->mst_parent = ci;
+
+		ir_nodemap_insert(&env->map, node, ci);
+	}
 	return ci;
 }
 
@@ -220,8 +219,8 @@ static void *co2_irn_init(ir_phase *ph, const ir_node *irn, void *data)
 
 static int cmp_clouds_gt(const void *a, const void *b)
 {
-	const co2_cloud_t * const *p = a;
-	const co2_cloud_t * const *q = b;
+	const co2_cloud_t * const *p = (const co2_cloud_t*const*)a;
+	const co2_cloud_t * const *q = (const co2_cloud_t*const*)b;
 	double c = CLOUD_WEIGHT(*p);
 	double d = CLOUD_WEIGHT(*q);
 	return QSORT_CMP(d, c);
@@ -233,17 +232,17 @@ static int cmp_clouds_gt(const void *a, const void *b)
  */
 static int col_cost_pair_lt(const void *a, const void *b)
 {
-	const col_cost_pair_t *p = a;
-	const col_cost_pair_t *q = b;
+	const col_cost_pair_t *p = (const col_cost_pair_t*)a;
+	const col_cost_pair_t *q = (const col_cost_pair_t*)b;
 	int c = p->costs;
 	int d = q->costs;
 	return QSORT_CMP(c, d);
 }
 
-int cmp_edges(const void *a, const void *b)
+static int cmp_edges(const void *a, const void *b)
 {
-	const edge_t *p = a;
-	const edge_t *q = b;
+	const edge_t *p = (const edge_t*)a;
+	const edge_t *q = (const edge_t*)b;
 	return QSORT_CMP(q->costs, p->costs);
 }
 
@@ -261,23 +260,22 @@ static inline int color_is_fix(co2_t *env, const ir_node *irn)
 
 static inline bitset_t *get_adm(co2_t *env, co2_irn_t *ci)
 {
-	if(ci->adm_cache == NULL) {
+	if (ci->adm_cache == NULL) {
 		const arch_register_req_t *req;
-		ci->adm_cache = bitset_obstack_alloc(phase_obst(&env->ph), env->n_regs);
-		req = arch_get_register_req_out(ci->irn);
+		ci->adm_cache = bitset_obstack_alloc(&env->obst, env->n_regs);
+		req = arch_get_irn_register_req(ci->irn);
 
-		if(arch_register_req_is(req, limited)) {
+		if (arch_register_req_is(req, limited)) {
 			int i, n;
 
 			n = env->n_regs;
-			for(i = 0; i < n; ++i) {
-				if(rbitset_is_set(req->limited, i))
+			for (i = 0; i < n; ++i) {
+				if (rbitset_is_set(req->limited, i))
 					bitset_set(ci->adm_cache, i);
 			}
 			ci->is_constrained = 1;
 		} else {
-			bitset_copy(ci->adm_cache, env->ignore_regs);
-			bitset_flip_all(ci->adm_cache);
+			bitset_copy(ci->adm_cache, env->allocatable_regs);
 		}
 	}
 
@@ -298,21 +296,21 @@ static inline int is_color_admissible(co2_t *env, co2_irn_t *ci, col_t col)
 
 static inline int is_constrained(co2_t *env, co2_irn_t *ci)
 {
-	if(!ci->adm_cache)
+	if (!ci->adm_cache)
 		get_adm(env, ci);
 	return ci->is_constrained;
 }
 
 static void incur_constraint_costs(co2_t *env, const ir_node *irn, col_cost_pair_t *col_costs, int costs)
 {
-	const arch_register_req_t *req = arch_get_register_req_out(irn);
+	const arch_register_req_t *req = arch_get_irn_register_req(irn);
 
 	if (arch_register_req_is(req, limited)) {
 		unsigned n_regs   = env->co->cls->n_regs;
 		unsigned n_constr = 0;
 		unsigned i;
 
-		n_constr = rbitset_popcnt(req->limited, n_regs);
+		n_constr = rbitset_popcount(req->limited, n_regs);
 		for (i = 0; i < n_regs; ++i) {
 			if (rbitset_is_set(req->limited, i)) {
 				col_costs[i].costs = add_saturated(col_costs[i].costs, costs / n_constr);
@@ -339,25 +337,22 @@ static void determine_color_costs(co2_t *env, co2_irn_t *ci, col_cost_pair_t *co
 	bitset_t *forb     = bitset_alloca(n_regs);
 	affinity_node_t *a = ci->aff;
 
-	bitset_pos_t elm;
 	const ir_node *pos;
-	void *it;
+	neighbours_iter_t it;
 	int i;
 
 	/* Put all forbidden colors into the aux bitset. */
 	admissible_colors(env, ci, forb);
 	bitset_flip_all(forb);
 
-	for(i = 0; i < n_regs; ++i) {
+	for (i = 0; i < n_regs; ++i) {
 		col_costs[i].col   = i;
 		col_costs[i].costs = 0;
 	}
 
-	if(a) {
-		neighb_t *n;
-
+	if (a) {
 		co_gs_foreach_neighb(a, n) {
-			if(color_is_fix(env, n->irn)) {
+			if (color_is_fix(env, n->irn)) {
 				col_t col = get_col(env, n->irn);
 				col_costs[col].costs = add_saturated(col_costs[col].costs, -n->costs * 128);
 			}
@@ -366,10 +361,9 @@ static void determine_color_costs(co2_t *env, co2_irn_t *ci, col_cost_pair_t *co
 		}
 	}
 
-	it = be_ifg_neighbours_iter_alloca(ifg);
-	be_ifg_foreach_neighbour(ifg, it, irn, pos) {
+	be_ifg_foreach_neighbour(ifg, &it, irn, pos) {
 		col_t col = get_col(env, pos);
-		if(color_is_fix(env, pos)) {
+		if (color_is_fix(env, pos)) {
 			col_costs[col].costs  = INT_MAX;
 		}
 		else {
@@ -377,7 +371,7 @@ static void determine_color_costs(co2_t *env, co2_irn_t *ci, col_cost_pair_t *co
 			col_costs[col].costs = add_saturated(col_costs[col].costs, 8 * be_ifg_degree(ifg, pos));
 		}
 	}
-	be_ifg_neighbours_break(ifg, it);
+	be_ifg_neighbours_break(&it);
 
 	/* Set the costs to infinity for each color which is not allowed at this node. */
 	bitset_foreach(forb, elm) {
@@ -391,7 +385,7 @@ static void single_color_cost(co2_t *env, co2_irn_t *ci, col_t col, col_cost_pai
 	int n_regs = env->co->cls->n_regs;
 	int i;
 
-	for(i = 0; i < n_regs; ++i) {
+	for (i = 0; i < n_regs; ++i) {
 		seq[i].col   = i;
 		seq[i].costs = INT_MAX;
 	}
@@ -405,16 +399,12 @@ static void single_color_cost(co2_t *env, co2_irn_t *ci, col_t col, col_cost_pai
 
 static void reject_coloring(struct list_head *h)
 {
-	co2_irn_t *pos;
-
 	list_for_each_entry(co2_irn_t, pos, h, changed_list)
 		pos->tmp_fixed = 0;
 }
 
 static void materialize_coloring(struct list_head *h)
 {
-	co2_irn_t *pos;
-
 	list_for_each_entry(co2_irn_t, pos, h, changed_list) {
 		pos->orig_col  = pos->tmp_col;
 		pos->tmp_fixed = 0;
@@ -432,22 +422,22 @@ static int recolor(co2_t *env, const ir_node *irn, col_cost_pair_t *col_list, st
 
 	int i;
 
-	if(depth >= max_depth)
+	if (depth >= max_depth)
 	  return 0;
 
-	for(i = 0; i < n_regs; ++i) {
+	for (i = 0; i < n_regs; ++i) {
 		col_t tgt_col  = col_list[i].col;
 		unsigned costs = col_list[i].costs;
 		int neigh_ok   = 1;
 
 		struct list_head changed;
 		const ir_node *n;
-		void *it;
+		neighbours_iter_t it;
 
 		DBG((env->dbg, LEVEL_3, "\t\t%2{firm:indent}trying color %d(%d) on %+F\n", depth, tgt_col, costs, irn));
 
 		/* If the costs for that color (and all successive) are infinite, bail out we won't make it anyway. */
-		if(INFEASIBLE(costs)) {
+		if (INFEASIBLE(costs)) {
 			DB((env->dbg, LEVEL_4, "\t\t%2{firm:indent}color %d infeasible\n", depth, tgt_col));
 			ci->tmp_fixed = 0;
 			return 0;
@@ -464,11 +454,10 @@ static int recolor(co2_t *env, const ir_node *irn, col_cost_pair_t *col_list, st
 		INIT_LIST_HEAD(&changed);
 		list_add(&ci->changed_list, &changed);
 
-		it = be_ifg_neighbours_iter_alloca(ifg);
-		be_ifg_foreach_neighbour(ifg, it, irn, n) {
+		be_ifg_foreach_neighbour(ifg, &it, irn, n) {
 
 			/* try to re-color the neighbor if it has the target color. */
-			if(get_col(env, n) == tgt_col) {
+			if (get_col(env, n) == tgt_col) {
 				struct list_head tmp;
 
 				/*
@@ -480,17 +469,17 @@ static int recolor(co2_t *env, const ir_node *irn, col_cost_pair_t *col_list, st
 				INIT_LIST_HEAD(&tmp);
 				neigh_ok = change_color_not(env, n, tgt_col, &tmp, depth + 1);
 				list_splice(&tmp, &changed);
-				if(!neigh_ok)
+				if (!neigh_ok)
 					break;
 			}
 		}
-		be_ifg_neighbours_break(ifg, it);
+		be_ifg_neighbours_break(&it);
 
 		/*
 		We managed to assign the target color to all neighbors, so from the perspective
 		of the current node, every thing was ok and we can return safely.
 		*/
-		if(neigh_ok) {
+		if (neigh_ok) {
 			DBG((env->dbg, LEVEL_3, "\t\t%2{firm:indent}color %d(%d) was ok\n", depth, tgt_col, costs));
 			list_splice(&changed, parent_changed);
 			res = 1;
@@ -517,8 +506,8 @@ static int change_color_not(co2_t *env, const ir_node *irn, col_t not_col, struc
 	DBG((env->dbg, LEVEL_3, "\t\t%2{firm:indent}clearing %+F(%d) of color %d\n", depth, irn, col, not_col));
 
 	/* the node does not have to forbidden color. That's fine, mark it as visited and return. */
-	if(col != not_col) {
-		if(!ci->tmp_fixed) {
+	if (col != not_col) {
+		if (!ci->tmp_fixed) {
 			ci->tmp_col     = col;
 			ci->tmp_fixed   = 1;
 		}
@@ -528,7 +517,7 @@ static int change_color_not(co2_t *env, const ir_node *irn, col_t not_col, struc
 	}
 
 	/* The node has the color it should not have _and_ has not been visited yet. */
-	if(!color_is_fix(env, irn)) {
+	if (!color_is_fix(env, irn)) {
 		int n_regs            = env->co->cls->n_regs;
 		col_cost_pair_t *csts = ALLOCAN(col_cost_pair_t, n_regs);
 
@@ -558,8 +547,8 @@ static int change_color_single(co2_t *env, const ir_node *irn, col_t tgt_col, st
 	DBG((env->dbg, LEVEL_3, "\t\t%2{firm:indent}trying to set %+F(%d) to color %d\n", depth, irn, col, tgt_col));
 
 	/* the node has the wanted color. That's fine, mark it as visited and return. */
-	if(col == tgt_col) {
-		if(!ci->tmp_fixed) {
+	if (col == tgt_col) {
+		if (!ci->tmp_fixed) {
 			ci->tmp_col     = col;
 			ci->tmp_fixed   = 1;
 			list_add(&ci->changed_list, parent_changed);
@@ -569,7 +558,7 @@ static int change_color_single(co2_t *env, const ir_node *irn, col_t tgt_col, st
 		goto end;
 	}
 
-	if(!color_is_fix(env, irn) && is_color_admissible(env, ci, tgt_col)) {
+	if (!color_is_fix(env, irn) && is_color_admissible(env, ci, tgt_col)) {
 		int n_regs           = env->co->cls->n_regs;
 		col_cost_pair_t *seq = ALLOCAN(col_cost_pair_t, n_regs);
 
@@ -598,7 +587,7 @@ static int examine_subtree_coloring(co2_cloud_irn_t *ci, col_t col)
 	int cost   = 0;
 	int i;
 
-	for(i = 0; i < ci->mst_n_childs; ++i) {
+	for (i = 0; i < ci->mst_n_childs; ++i) {
 		co2_cloud_irn_t *chld = ci->mst_childs[i];
 		col_t chld_col        = front[i];
 
@@ -626,9 +615,8 @@ static void node_color_badness(co2_cloud_irn_t *ci, int *badness)
 	be_ifg_t *ifg  = env->co->cenv->ifg;
 	bitset_t *bs   = bitset_alloca(n_regs);
 
-	bitset_pos_t elm;
 	const ir_node *irn;
-	void *it;
+	neighbours_iter_t it;
 
 	admissible_colors(env, &ci->inh, bs);
 	bitset_flip_all(bs);
@@ -636,22 +624,21 @@ static void node_color_badness(co2_cloud_irn_t *ci, int *badness)
 		badness[elm] = ci->costs;
 
 	/* Use constrained/fixed interfering neighbors to influence the color badness */
-	it = be_ifg_neighbours_iter_alloca(ifg);
-	be_ifg_foreach_neighbour(ifg, it, ir->irn, irn) {
+	be_ifg_foreach_neighbour(ifg, &it, ir->irn, irn) {
 		co2_irn_t *ni = get_co2_irn(env, irn);
 
 		admissible_colors(env, ni, bs);
-		if(bitset_popcnt(bs) == 1) {
-			bitset_pos_t c = bitset_next_set(bs, 0);
+		if (bitset_popcount(bs) == 1) {
+			size_t c = bitset_next_set(bs, 0);
 			badness[c] += ci->costs;
 		}
 
-		else if(ni->fixed) {
+		else if (ni->fixed) {
 			col_t c = get_col(env, ni->irn);
 			badness[c] += ci->costs;
 		}
 	}
-	be_ifg_neighbours_break(ifg, it);
+	be_ifg_neighbours_break(&it);
 }
 
 /**
@@ -669,15 +656,15 @@ static void determine_color_badness(co2_cloud_irn_t *ci, int depth)
 	node_color_badness(ci, ci->color_badness);
 
 	/* Collect the color badness for the whole subtree */
-	for(i = 0; i < ci->mst_n_childs; ++i) {
+	for (i = 0; i < ci->mst_n_childs; ++i) {
 		co2_cloud_irn_t *child = ci->mst_childs[i];
 		determine_color_badness(child, depth + 1);
 
-		for(j = 0; j < env->n_regs; ++j)
+		for (j = 0; j < env->n_regs; ++j)
 			ci->color_badness[j] += child->color_badness[j];
 	}
 
-	for(j = 0; j < env->n_regs; ++j)
+	for (j = 0; j < env->n_regs; ++j)
 		DBG((env->dbg, LEVEL_2, "%2{firm:indent}%+F col %d badness %d\n", depth, ci->inh.irn, j, ci->color_badness[j]));
 }
 
@@ -689,7 +676,7 @@ static void unfix_subtree(co2_cloud_irn_t *ci)
 	int i;
 
 	ci->inh.fixed = 0;
-	for(i = 0; i < ci->mst_n_childs; ++i)
+	for (i = 0; i < ci->mst_n_childs; ++i)
 		unfix_subtree(ci->mst_childs[i]);
 }
 
@@ -708,7 +695,7 @@ static int coalesce_top_down(co2_cloud_irn_t *ci, int child_nr, int depth)
 	struct list_head changed;
 	int ok, i, j;
 
-	for(i = 0; i < n_regs; ++i) {
+	for (i = 0; i < n_regs; ++i) {
 		int badness = ci->color_badness[i];
 
 		seq[i].col   = i;
@@ -718,7 +705,7 @@ static int coalesce_top_down(co2_cloud_irn_t *ci, int child_nr, int depth)
 	}
 
 	/* If we are not the root and the parent's color is allowed for this node give it top prio. */
-	if(!is_root && is_color_admissible(env, &ci->inh, parent_col))
+	if (!is_root && is_color_admissible(env, &ci->inh, parent_col))
 		seq[parent_col].costs = min_badness - 1;
 
 	/* Sort the colors. The will be processed in that ordering. */
@@ -726,7 +713,7 @@ static int coalesce_top_down(co2_cloud_irn_t *ci, int child_nr, int depth)
 
 	DBG((env->dbg, LEVEL_2, "\t%2{firm:indent}starting top-down coalesce for %+F\n", depth, ci->inh.irn));
 	INIT_LIST_HEAD(&changed);
-	for(i = 0; i < (best_col < 0 ? n_regs : n_iter); ++i) {
+	for (i = 0; i < (best_col < 0 ? n_regs : n_iter); ++i) {
 		col_t col    = seq[i].col;
 		int add_cost = !is_root && col != parent_col ? ci->mst_costs : 0;
 
@@ -737,7 +724,7 @@ static int coalesce_top_down(co2_cloud_irn_t *ci, int child_nr, int depth)
 		unfix_subtree(ci);
 		INIT_LIST_HEAD(&changed);
 		ok = change_color_single(env, ci->inh.irn, col, &changed, depth);
-		if(ok) {
+		if (ok) {
 			materialize_coloring(&changed);
 			ci->inh.fixed = 1;
 		}
@@ -745,34 +732,34 @@ static int coalesce_top_down(co2_cloud_irn_t *ci, int child_nr, int depth)
 		else
 			continue;
 
-		for(j = 0; j < ci->mst_n_childs; ++j) {
+		for (j = 0; j < ci->mst_n_childs; ++j) {
 			co2_cloud_irn_t *child = ci->mst_childs[j];
 			ok = coalesce_top_down(child, j, depth + 1) >= 0;
-			if(ok)
+			if (ok)
 				child->inh.fixed = 1;
 			else
 				break;
 		}
 
 		/* If the subtree could not be colored, we have to try another color. */
-		if(!ok)
+		if (!ok)
 			continue;
 
 		subtree_costs      = examine_subtree_coloring(ci, col);
 		sum_costs          = subtree_costs + add_cost;
 		DBG((env->dbg, LEVEL_2, "\t%2{firm:indent}-> %+F costing %d + %d is ok.\n", depth, ci->inh.irn, subtree_costs, add_cost));
 
-		if(sum_costs < best_col_costs) {
+		if (sum_costs < best_col_costs) {
 			best_col           = col;
 			best_col_costs     = sum_costs;
 			ci->col_costs[col] = subtree_costs;
 		}
 
-		if(sum_costs == 0)
+		if (sum_costs == 0)
 			break;
 	}
 
-	if(!is_root) {
+	if (!is_root) {
 		int *front = FRONT_BASE(ci->mst_parent, parent_col);
 		front[child_nr] = best_col;
 	}
@@ -785,9 +772,8 @@ static void populate_cloud(co2_t *env, co2_cloud_t *cloud, affinity_node_t *a, i
 	be_ifg_t *ifg       = env->co->cenv->ifg;
 	co2_cloud_irn_t *ci = get_co2_cloud_irn(env, a->irn);
 	int costs           = 0;
-	neighb_t *n;
 
-	if(ci->cloud)
+	if (ci->cloud)
 		return;
 
 	/* mark the node as visited and add it to the cloud. */
@@ -800,7 +786,7 @@ static void populate_cloud(co2_t *env, co2_cloud_t *cloud, affinity_node_t *a, i
 	co_gs_foreach_neighb(a, n) {
 		costs += n->costs;
 		DB((env->dbg, LEVEL_3, "\t\tneigh %+F cost %d\n", n->irn, n->costs));
-		if(be_ifg_connected(ifg, a->irn, n->irn))
+		if (be_ifg_connected(ifg, a->irn, n->irn))
 			cloud->inevit += n->costs;
 	}
 
@@ -808,12 +794,12 @@ static void populate_cloud(co2_t *env, co2_cloud_t *cloud, affinity_node_t *a, i
 	ci->costs          = costs;
 	cloud->costs      += costs;
 	cloud->n_constr   += is_constrained(env, &ci->inh);
-	cloud->freedom    += bitset_popcnt(get_adm(env, &ci->inh));
+	cloud->freedom    += bitset_popcount(get_adm(env, &ci->inh));
 	cloud->max_degree  = MAX(cloud->max_degree, ci->inh.aff->degree);
 	cloud->n_memb++;
 
 	/* If this is the heaviest node in the cloud, set it as the cloud's master. */
-	if(costs >= curr_costs) {
+	if (costs >= curr_costs) {
 		curr_costs    = costs;
 		cloud->master = ci;
 	}
@@ -828,8 +814,7 @@ static void populate_cloud(co2_t *env, co2_cloud_t *cloud, affinity_node_t *a, i
 
 static co2_cloud_t *new_cloud(co2_t *env, affinity_node_t *a)
 {
-	co2_cloud_t *cloud = phase_alloc(&env->ph, sizeof(cloud[0]));
-	co2_cloud_irn_t *ci;
+	co2_cloud_t *cloud = OALLOC(&env->obst, co2_cloud_t);
 	int i;
 
 	DBG((env->dbg, LEVEL_2, "new cloud with %+F\n", a->irn));
@@ -844,7 +829,7 @@ static co2_cloud_t *new_cloud(co2_t *env, affinity_node_t *a)
 	cloud->freedom = (cloud->n_memb * env->n_regs) / cloud->freedom;
 
 	/* Also allocate space for the node sequence and compute that sequence. */
-	cloud->seq    = phase_alloc(&env->ph, cloud->n_memb * sizeof(cloud->seq[0]));
+	cloud->seq = OALLOCN(&env->obst, co2_cloud_irn_t*, cloud->n_memb);
 
 	i = 0;
 	list_for_each_entry(co2_cloud_irn_t, ci, &cloud->members_head, cloud_list) {
@@ -860,24 +845,23 @@ static void apply_coloring(co2_cloud_irn_t *ci, col_t col, int depth)
 {
 	const ir_node *irn = ci->inh.irn;
 	int *front   = FRONT_BASE(ci, col);
-	int i, ok;
+	int i;
 	struct list_head changed;
 
 	INIT_LIST_HEAD(&changed);
 
 	DBG((ci->cloud->env->dbg, LEVEL_2, "%2{firm:indent}setting %+F to %d\n", depth, irn, col));
-	ok = change_color_single(ci->cloud->env, irn, col, &changed, depth);
-	// assert(ok && "Color changing may not fail while committing the coloring");
+	change_color_single(ci->cloud->env, irn, col, &changed, depth);
 	materialize_coloring(&changed);
 
- 	for(i = 0; i < ci->mst_n_childs; ++i) {
+	for (i = 0; i < ci->mst_n_childs; ++i) {
 		apply_coloring(ci->mst_childs[i], front[i], depth + 1);
 	}
 }
 
 static co2_cloud_irn_t *find_mst_root(co2_cloud_irn_t *ci)
 {
-	while(ci != ci->mst_parent)
+	while (ci != ci->mst_parent)
 		ci = ci->mst_parent;
 	return ci;
 }
@@ -897,13 +881,12 @@ static void process_cloud(co2_cloud_t *cloud)
 
 	/* Collect all edges in the cloud on an obstack and sort the increasingly */
 	obstack_init(&cloud->obst);
-	for(i = 0; i < cloud->n_memb; ++i) {
+	for (i = 0; i < cloud->n_memb; ++i) {
 		co2_cloud_irn_t *ci = cloud->seq[i];
-		neighb_t *n;
 
 		co_gs_foreach_neighb(ci->inh.aff, n) {
 			co2_cloud_irn_t *ni = get_co2_cloud_irn(cloud->env, n->irn);
-			if(ci->index < ni->index) {
+			if (ci->index < ni->index) {
 				edge_t e;
 				e.src   = ci;
 				e.tgt   = ni;
@@ -913,18 +896,18 @@ static void process_cloud(co2_cloud_t *cloud)
 			}
 		}
 	}
-	edges = obstack_finish(&cloud->obst);
+	edges = (edge_t*)obstack_finish(&cloud->obst);
 	qsort(edges, n_edges, sizeof(edges[0]), cmp_edges);
 
 	/* Compute the maximum spanning tree using Kruskal/Union-Find */
 	DBG((env->dbg, LEVEL_2, "computing spanning tree of cloud with master %+F\n", cloud->master->inh.irn));
-	for(i = 0; i < n_edges; ++i) {
+	for (i = 0; i < n_edges; ++i) {
 		edge_t *e        = &edges[i];
 		co2_cloud_irn_t *rs = find_mst_root(e->src);
 		co2_cloud_irn_t *rt = find_mst_root(e->tgt);
 
 		/* if the union/find roots are different */
-		if(rs != rt) {
+		if (rs != rt) {
 			int si = e->src->index;
 			int ti = e->tgt->index;
 
@@ -942,15 +925,15 @@ static void process_cloud(co2_cloud_t *cloud)
 	cloud->master->mst_parent = cloud->master;
 	cloud->mst_root = cloud->master;
 	q = new_pdeq1(cloud->master);
-	while(!pdeq_empty(q)) {
-		co2_cloud_irn_t *ci = pdeq_getl(q);
+	while (!pdeq_empty(q)) {
+		co2_cloud_irn_t *ci = (co2_cloud_irn_t*)pdeq_getl(q);
 		int ofs    = ci->index * cloud->n_memb;
 		int end    = ofs + cloud->n_memb;
 		int i;
 
 		ci->mst_n_childs = 0;
-		for(i = ofs; i < end; ++i) {
-			if(mst_edges[i] != 0) {
+		for (i = ofs; i < end; ++i) {
+			if (mst_edges[i] != 0) {
 				int other = i - ofs;
 				co2_cloud_irn_t *child = cloud->seq[i - ofs];
 
@@ -969,35 +952,30 @@ static void process_cloud(co2_cloud_t *cloud)
 		}
 
 		obstack_ptr_grow(&cloud->obst, NULL);
-		ci->mst_childs = obstack_finish(&cloud->obst);
+		ci->mst_childs = (co2_cloud_irn_t**)obstack_finish(&cloud->obst);
 	}
 	del_pdeq(q);
 	free(mst_edges);
 
 
 	DBG((env->dbg, LEVEL_3, "mst:\n"));
-	for(i = 0; i < cloud->n_memb; ++i) {
-		DEBUG_ONLY(co2_cloud_irn_t *ci = cloud->seq[i]);
+	for (i = 0; i < cloud->n_memb; ++i) {
+		DEBUG_ONLY(co2_cloud_irn_t *ci = cloud->seq[i];)
 		DBG((env->dbg, LEVEL_3, "\t%+F -> %+F\n", ci->inh.irn, ci->mst_parent->inh.irn));
 	}
 
-	for(i = 0; i < cloud->n_memb; ++i) {
+	for (i = 0; i < cloud->n_memb; ++i) {
 		co2_cloud_irn_t *ci = cloud->seq[i];
 		int n_childs = ci->mst_n_childs;
 		int j;
 
-		ci->col_costs       = obstack_alloc(&cloud->obst, n_regs * sizeof(ci->col_costs[0]));
-		ci->tmp_coloring    = obstack_alloc(&cloud->obst, n_regs * sizeof(ci->tmp_coloring[0]));
-		ci->fronts          = obstack_alloc(&cloud->obst, n_regs * n_childs * sizeof(ci->fronts[0]));
-		ci->color_badness   = obstack_alloc(&cloud->obst, n_regs * sizeof(ci->fronts[0]));
-		memset(ci->color_badness, 0, n_regs * sizeof(ci->color_badness[0]));
-		memset(ci->col_costs,     0, n_regs * sizeof(ci->col_costs[0]));
-		memset(ci->tmp_coloring,  0, n_regs * sizeof(ci->tmp_coloring[0]));
-		memset(ci->fronts,        0, n_regs * n_childs * sizeof(ci->fronts[0]));
+		ci->col_costs       = OALLOCNZ(&cloud->obst, int,             n_regs);
+		ci->tmp_coloring    = OALLOCNZ(&cloud->obst, col_cost_pair_t, n_regs);
+		ci->fronts          = OALLOCNZ(&cloud->obst, int,             n_regs * n_childs);
+		ci->color_badness   = OALLOCNZ(&cloud->obst, int,             n_regs);
 
-		for(j = 0; j < env->n_regs; j++)
+		for (j = 0; j < env->n_regs; j++)
 			ci->col_costs[j] = INT_MAX;
-
 	}
 
 	determine_color_badness(cloud->mst_root, 0);
@@ -1011,7 +989,7 @@ static void process_cloud(co2_cloud_t *cloud)
 		cloud->mst_root->inh.irn, best_col, examine_subtree_coloring(cloud->mst_root, best_col)));
 
 	/* Fix all nodes in the cloud. */
-	for(i = 0; i < cloud->n_memb; ++i)
+	for (i = 0; i < cloud->n_memb; ++i)
 		cloud->seq[i]->inh.fixed = 1;
 
 	/* Free all space used while optimizing this cloud. */
@@ -1021,9 +999,8 @@ static void process_cloud(co2_cloud_t *cloud)
 static int cloud_costs(co2_cloud_t *cloud)
 {
 	int i, costs = 0;
-	neighb_t *n;
 
-	for(i = 0; i < cloud->n_memb; ++i) {
+	for (i = 0; i < cloud->n_memb; ++i) {
 		co2_irn_t *ci = (co2_irn_t *) cloud->seq[i];
 		col_t col = get_col(cloud->env, ci->irn);
 		co_gs_foreach_neighb(ci->aff, n) {
@@ -1035,10 +1012,18 @@ static int cloud_costs(co2_cloud_t *cloud)
 	return costs / 2;
 }
 
+static void writeback_colors(co2_t *env)
+{
+	co2_irn_t *irn;
+
+	for (irn = env->touched; irn; irn = irn->touched_next) {
+		const arch_register_t *reg = arch_register_for_index(env->co->cls, irn->orig_col);
+		arch_set_irn_register((ir_node*)irn->irn, reg);
+	}
+}
+
 static void process(co2_t *env)
 {
-	affinity_node_t *a;
-	co2_cloud_t *pos;
 	co2_cloud_t **clouds;
 	int n_clouds;
 	int i;
@@ -1050,7 +1035,7 @@ static void process(co2_t *env)
 	co_gs_foreach_aff_node(env->co, a) {
 		co2_cloud_irn_t *ci = get_co2_cloud_irn(env, a->irn);
 
-		if(!ci->cloud) {
+		if (!ci->cloud) {
 			new_cloud(env, a);
 			n_clouds++;
 		}
@@ -1062,7 +1047,7 @@ static void process(co2_t *env)
 		clouds[i++] = pos;
 	qsort(clouds, n_clouds, sizeof(clouds[0]), cmp_clouds_gt);
 
-	for(i = 0; i < n_clouds; ++i) {
+	for (i = 0; i < n_clouds; ++i) {
 		init_costs  += cloud_costs(clouds[i]);
 
 		/* Process the cloud. */
@@ -1070,19 +1055,6 @@ static void process(co2_t *env)
 
 		all_costs   += clouds[i]->costs;
 		final_costs += cloud_costs(clouds[i]);
-
-		/* Dump the IFG if the user demanded it. */
-		if (dump_flags & DUMP_CLOUD) {
-			char buf[256];
-			FILE *f;
-
-			ir_snprintf(buf, sizeof(buf), "ifg_%F_%s_cloud_%d.dot", env->co->irg, env->co->cls->name, i);
-			f = fopen(buf, "wt");
-			if(f != NULL) {
-				be_ifg_dump_dot(env->co->cenv->ifg, env->co->irg, f, &ifg_dot_cb, env);
-				fclose(f);
-			}
-		}
 	}
 
 	DB((env->dbg, LEVEL_1, "all costs: %d, init costs: %d, final costs: %d\n", all_costs, init_costs, final_costs));
@@ -1090,195 +1062,41 @@ static void process(co2_t *env)
 	xfree(clouds);
 }
 
-static void writeback_colors(co2_t *env)
+static int co_solve_heuristic_new(copy_opt_t *co)
 {
-	co2_irn_t *irn;
-
-	for(irn = env->touched; irn; irn = irn->touched_next) {
-		const arch_register_t *reg = arch_register_for_index(env->co->cls, irn->orig_col);
-		arch_set_irn_register((ir_node*)irn->irn, reg);
-	}
-}
-
-
-/*
-  ___ _____ ____   ____   ___ _____   ____                        _
- |_ _|  ___/ ___| |  _ \ / _ \_   _| |  _ \ _   _ _ __ ___  _ __ (_)_ __   __ _
-  | || |_ | |  _  | | | | | | || |   | | | | | | | '_ ` _ \| '_ \| | '_ \ / _` |
-  | ||  _|| |_| | | |_| | |_| || |   | |_| | |_| | | | | | | |_) | | | | | (_| |
- |___|_|   \____| |____/ \___/ |_|   |____/ \__,_|_| |_| |_| .__/|_|_| |_|\__, |
-                                                           |_|            |___/
-*/
-
-static const char *get_dot_color_name(size_t col)
-{
-	static const char *const names[] = {
-		"blue",
-		"red",
-		"green",
-		"yellow",
-		"cyan",
-		"magenta",
-		"orange",
-		"chocolate",
-		"beige",
-		"navy",
-		"darkgreen",
-		"darkred",
-		"lightPink",
-		"chartreuse",
-		"lightskyblue",
-		"linen",
-		"pink",
-		"lightslateblue",
-		"mintcream",
-		"red",
-		"darkolivegreen",
-		"mediumblue",
-		"mistyrose",
-		"salmon",
-		"darkseagreen",
-		"mediumslateblue"
-		"moccasin",
-		"tomato",
-		"forestgreen",
-		"darkturquoise",
-		"palevioletred"
-	};
-
-	return col < (sizeof(names)/sizeof(names[0])) ? names[col] : "white";
-}
-
-static const char *get_dot_shape_name(co2_irn_t *ci)
-{
-	const arch_register_req_t *req = arch_get_register_req_out(ci->irn);
-
-	if(arch_register_req_is(req, limited))
-		return "diamond";
-
-	if(ci->fixed)
-		return "rectangle";
-
-	if(ci->tmp_fixed)
-		return "hexagon";
-
-	return "ellipse";
-}
-
-static void ifg_dump_graph_attr(FILE *f, void *self)
-{
-	(void) self;
-	fprintf(f, "overlay=false");
-}
-
-static int ifg_is_dump_node(void *self, ir_node *irn)
-{
-	const arch_register_req_t *req = arch_get_register_req_out(irn);
-	(void)self;
-	return !(req->type & arch_register_req_type_ignore);
-}
-
-static void ifg_dump_node_attr(FILE *f, void *self, ir_node *irn)
-{
-	co2_t *env    = self;
-	co2_irn_t *ci = get_co2_irn(env, irn);
-	int peri      = 1;
-
-	char buf[128] = "";
-
-	if(ci->aff) {
-		co2_cloud_irn_t *cci = (void *) ci;
-		if (cci->cloud && cci->cloud->mst_root == cci)
-			peri = 2;
-
-		if(cci->cloud && cci->cloud->mst_root)
-			ir_snprintf(buf, sizeof(buf), "%+F", cci->cloud->mst_root->inh.irn);
-	}
-
-	ir_fprintf(f, "label=\"%+F%s\" style=filled peripheries=%d color=%s shape=%s", irn, buf, peri,
-		get_dot_color_name(get_col(env, irn)), get_dot_shape_name(ci));
-}
-
-static void ifg_dump_at_end(FILE *file, void *self)
-{
-	co2_t *env = self;
-	affinity_node_t *a;
-
-	co_gs_foreach_aff_node(env->co, a) {
-		co2_cloud_irn_t *ai = get_co2_cloud_irn(env, a->irn);
-		int idx = get_irn_idx(a->irn);
-		neighb_t *n;
-
-		if(ai->mst_parent != ai)
-			fprintf(file, "\tn%d -- n%d [style=dotted color=blue arrowhead=normal];\n", idx, get_irn_idx(ai->mst_parent->inh.irn));
-
-		co_gs_foreach_neighb(a, n) {
-			int nidx = get_irn_idx(n->irn);
-			co2_cloud_irn_t *ci = get_co2_cloud_irn(env, n->irn);
-
-			if(idx < nidx) {
-				const char *color = get_col(env, a->irn) == get_col(env, n->irn) ? "black" : "red";
-				const char *arr = "arrowhead=dot arrowtail=dot";
-
-				if(ci->mst_parent == ai)
-					arr = "arrowtail=normal";
-				else if(ai->mst_parent == ci)
-					arr = "arrowhead=normal";
-
-				fprintf(file, "\tn%d -- n%d [label=\"%d\" %s style=dashed color=%s weight=0.01];\n", idx, nidx, n->costs, arr, color);
-			}
-		}
-	}
-}
-
-
-static be_ifg_dump_dot_cb_t ifg_dot_cb = {
-	ifg_is_dump_node,
-	ifg_dump_graph_attr,
-	ifg_dump_node_attr,
-	NULL,
-	NULL,
-	ifg_dump_at_end
-};
-
-
-int co_solve_heuristic_new(copy_opt_t *co)
-{
-	char  buf[256];
 	co2_t env;
-	FILE  *f;
 
-	phase_init(&env.ph, "co2", co->cenv->birg->irg, PHASE_DEFAULT_GROWTH, co2_irn_init, NULL);
+	ir_nodemap_init(&env.map, co->irg);
+	obstack_init(&env.obst);
 	env.touched     = NULL;
 	env.visited     = 0;
 	env.co          = co;
 	env.n_regs      = co->cls->n_regs;
-	env.ignore_regs = bitset_alloca(co->cls->n_regs);
-	be_put_ignore_regs(co->cenv->birg, co->cls, env.ignore_regs);
+	env.allocatable_regs = bitset_alloca(co->cls->n_regs);
+	be_put_allocatable_regs(co->cenv->irg, co->cls, env.allocatable_regs);
 	FIRM_DBG_REGISTER(env.dbg, "firm.be.co2");
 	INIT_LIST_HEAD(&env.cloud_head);
 
-	if(dump_flags & DUMP_BEFORE) {
-		ir_snprintf(buf, sizeof(buf), "ifg_%F_%s_before.dot", co->irg, co->cls->name);
-		f = fopen(buf, "wt");
-		if (f != NULL) {
-			be_ifg_dump_dot(co->cenv->ifg, co->irg, f, &ifg_dot_cb, &env);
-			fclose(f);
-		}
-	}
-
 	process(&env);
 
-	if(dump_flags & DUMP_AFTER) {
-		ir_snprintf(buf, sizeof(buf), "ifg_%F_%s_after.dot", co->irg, co->cls->name);
-		f = fopen(buf, "wt");
-		if (f != NULL) {
-			be_ifg_dump_dot(co->cenv->ifg, co->irg, f, &ifg_dot_cb, &env);
-			fclose(f);
-		}
-	}
-
 	writeback_colors(&env);
-	phase_free(&env.ph);
+	obstack_free(&env.obst, NULL);
+	ir_nodemap_destroy(&env.map);
 	return 0;
 }
+
+BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copyheur2)
+void be_init_copyheur2(void)
+{
+	lc_opt_entry_t *be_grp = lc_opt_get_grp(firm_opt_get_root(), "be");
+	lc_opt_entry_t *ra_grp = lc_opt_get_grp(be_grp, "ra");
+	lc_opt_entry_t *chordal_grp = lc_opt_get_grp(ra_grp, "chordal");
+	lc_opt_entry_t *co2_grp = lc_opt_get_grp(chordal_grp, "co2");
+
+	static co_algo_info copyheur = {
+		co_solve_heuristic_new, 0
+	};
+
+	lc_opt_add_table(co2_grp, options);
+	be_register_copyopt("heur2", &copyheur);
+}
