@@ -22,7 +22,6 @@
  * @brief       Copy minimization driver.
  * @author      Daniel Grund
  * @date        12.04.2005
- * @version     $Id: becopyopt.c 25941 2009-05-14 09:17:29Z beck $
  *
  * Main file for the optimization reducing the copies needed for:
  * - Phi coalescing
@@ -31,36 +30,37 @@
  */
 #include "config.h"
 
-#include "execfreq.h"
-#include "xmalloc.h"
 #include "debug.h"
-#include "pmap.h"
-#include "raw_bitset.h"
-#include "irnode.h"
+#include "error.h"
+#include "execfreq_t.h"
+#include "irdump_t.h"
+#include "iredges_t.h"
 #include "irgraph.h"
 #include "irgwalk.h"
-#include "irprog.h"
 #include "irloop_t.h"
-#include "iredges_t.h"
-#include "phiclass.h"
-#include "irbitset.h"
-#include "irphase_t.h"
+#include "irnode.h"
 #include "irprintf_t.h"
+#include "irprog.h"
+#include "irtools.h"
+#include "pmap.h"
+#include "raw_bitset.h"
+#include "util.h"
+#include "xmalloc.h"
 
-#include "bemodule.h"
-#include "bearch_t.h"
-#include "benode_t.h"
-#include "beutil.h"
-#include "beifg_t.h"
-#include "beintlive_t.h"
+#include "bearch.h"
 #include "becopyopt_t.h"
 #include "becopystat.h"
-#include "belive_t.h"
+#include "bedump.h"
+#include "beifg.h"
 #include "beinsn_t.h"
-#include "besched_t.h"
-#include "bestatevent.h"
-#include "beirg_t.h"
-#include "error.h"
+#include "beintlive_t.h"
+#include "beirg.h"
+#include "belive_t.h"
+#include "bemodule.h"
+#include "benode.h"
+#include "besched.h"
+#include "statev_t.h"
+#include "beutil.h"
 
 #include "lc_opts.h"
 #include "lc_opts_enum.h"
@@ -74,11 +74,24 @@
 #define COST_FUNC_LOOP     2
 #define COST_FUNC_ALL_ONE  3
 
+/**
+ * Flags for dumping the IFG.
+ */
+enum {
+	CO_IFG_DUMP_COLORS = 1 << 0, /**< Dump the graph colored. */
+	CO_IFG_DUMP_LABELS = 1 << 1, /**< Dump node/edge labels. */
+	CO_IFG_DUMP_SHAPE  = 1 << 2, /**< Give constrained nodes special shapes. */
+	CO_IFG_DUMP_CONSTR = 1 << 3, /**< Dump the node constraints in the label. */
+};
+
+static int co_get_costs_loop_depth(const ir_node *root, int pos);
+static int co_get_costs_exec_freq(const ir_node *root, int pos);
+static int co_get_costs_all_one(const ir_node *root, int pos);
+
 static unsigned   dump_flags  = 0;
-static unsigned   style_flags = 0;
-static unsigned   do_stats    = 0;
+static unsigned   style_flags = CO_IFG_DUMP_COLORS;
+static int        do_stats    = 0;
 static cost_fct_t cost_func   = co_get_costs_exec_freq;
-static unsigned   algo        = CO_ALGO_HEUR4;
 static int        improve     = 1;
 
 static const lc_opt_enum_mask_items_t dump_items[] = {
@@ -98,18 +111,7 @@ static const lc_opt_enum_mask_items_t style_items[] = {
 	{ NULL,      0 }
 };
 
-static const lc_opt_enum_mask_items_t algo_items[] = {
-	{ "none",   CO_ALGO_NONE  },
-	{ "heur",   CO_ALGO_HEUR  },
-	{ "heur2",  CO_ALGO_HEUR2 },
-	{ "heur3",  CO_ALGO_HEUR3 },
-	{ "heur4",  CO_ALGO_HEUR4 },
-	{ "ilp",    CO_ALGO_ILP   },
-	{ NULL,     0 }
-};
-
 typedef int (*opt_funcptr)(void);
-
 static const lc_opt_enum_func_ptr_items_t cost_func_items[] = {
 	{ "freq",   (opt_funcptr) co_get_costs_exec_freq },
 	{ "loop",   (opt_funcptr) co_get_costs_loop_depth },
@@ -125,30 +127,31 @@ static lc_opt_enum_mask_var_t style_var = {
 	&style_flags, style_items
 };
 
-static lc_opt_enum_mask_var_t algo_var = {
-	&algo, algo_items
-};
-
 static lc_opt_enum_func_ptr_var_t cost_func_var = {
 	(opt_funcptr*) &cost_func, cost_func_items
 };
 
 static const lc_opt_table_entry_t options[] = {
-	LC_OPT_ENT_ENUM_INT      ("algo",    "select copy optimization algo",                           &algo_var),
 	LC_OPT_ENT_ENUM_FUNC_PTR ("cost",    "select a cost function",                                  &cost_func_var),
 	LC_OPT_ENT_ENUM_MASK     ("dump",    "dump ifg before or after copy optimization",              &dump_var),
 	LC_OPT_ENT_ENUM_MASK     ("style",   "dump style for ifg dumping",                              &style_var),
 	LC_OPT_ENT_BOOL          ("stats",   "dump statistics after each optimization",                 &do_stats),
-	LC_OPT_ENT_BOOL          ("improve", "run heur3 before if algo can exploit start solutions",    &improve),
+	LC_OPT_ENT_BOOL          ("improve", "run heur1 before if algo can exploit start solutions",    &improve),
 	LC_OPT_LAST
 };
 
-/* Insert additional options registration functions here. */
-extern void be_co_ilp_register_options(lc_opt_entry_t *grp);
-extern void be_co2_register_options(lc_opt_entry_t *grp);
-extern void be_co3_register_options(lc_opt_entry_t *grp);
+static be_module_list_entry_t *copyopts = NULL;
+static const co_algo_info *selected_copyopt = NULL;
 
-void be_init_copycoal(void)
+void be_register_copyopt(const char *name, co_algo_info *copyopt)
+{
+	if (selected_copyopt == NULL)
+		selected_copyopt = copyopt;
+	be_add_module_to_list(&copyopts, name, copyopt);
+}
+
+BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copyopt)
+void be_init_copyopt(void)
 {
 	lc_opt_entry_t *be_grp = lc_opt_get_grp(firm_opt_get_root(), "be");
 	lc_opt_entry_t *ra_grp = lc_opt_get_grp(be_grp, "ra");
@@ -156,9 +159,25 @@ void be_init_copycoal(void)
 	lc_opt_entry_t *co_grp = lc_opt_get_grp(chordal_grp, "co");
 
 	lc_opt_add_table(co_grp, options);
+	be_add_module_list_opt(co_grp, "algo", "select copy optimization algo",
+		                       &copyopts, (void**) &selected_copyopt);
 }
 
-BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copycoal);
+static int void_algo(copy_opt_t *co)
+{
+	(void) co;
+	return 0;
+}
+
+BE_REGISTER_MODULE_CONSTRUCTOR(be_init_copynone)
+void be_init_copynone(void)
+{
+	static co_algo_info copyheur = {
+		void_algo, 0
+	};
+
+	be_register_copyopt("none", &copyheur);
+}
 
 #undef QUICK_AND_DIRTY_HACK
 
@@ -166,8 +185,10 @@ static int nodes_interfere(const be_chordal_env_t *env, const ir_node *a, const
 {
 	if (env->ifg)
 		return be_ifg_connected(env->ifg, a, b);
-	else
-		return values_interfere(env->birg, a, b);
+	else {
+		be_lv_t *lv = be_get_irg_liveness(env->irg);
+		return be_values_interfere(lv, a, b);
+	}
 }
 
 
@@ -187,7 +208,7 @@ DEBUG_ONLY(static firm_dbg_module_t *dbg = NULL;)
 copy_opt_t *new_copy_opt(be_chordal_env_t *chordal_env, cost_fct_t get_costs)
 {
 	const char *s1, *s2, *s3;
-	int len;
+	size_t len;
 	copy_opt_t *co;
 
 	FIRM_DBG_REGISTER(dbg, "ir.be.copyopt");
@@ -208,7 +229,8 @@ copy_opt_t *new_copy_opt(be_chordal_env_t *chordal_env, cost_fct_t get_costs)
 	return co;
 }
 
-void free_copy_opt(copy_opt_t *co) {
+void free_copy_opt(copy_opt_t *co)
+{
 	xfree(co->name);
 	free(co);
 }
@@ -220,62 +242,70 @@ void free_copy_opt(copy_opt_t *co) {
 static int co_is_optimizable_root(ir_node *irn)
 {
 	const arch_register_req_t *req;
-	const arch_register_t     *reg;
 
 	if (arch_irn_is_ignore(irn))
 		return 0;
 
-	reg = arch_get_irn_register(irn);
-	if (arch_register_type_is(reg, ignore))
-		return 0;
-
 	if (is_Reg_Phi(irn) || is_Perm_Proj(irn))
 		return 1;
 
-	req = arch_get_register_req_out(irn);
+	req = arch_get_irn_register_req(irn);
 	if (is_2addr_code(req))
 		return 1;
 
 	return 0;
 }
 
-int co_get_costs_loop_depth(const copy_opt_t *co, ir_node *root, ir_node* arg, int pos) {
-	int cost = 0;
+/**
+ * Computes the costs of a copy according to loop depth
+ * @param pos  the argument position of arg in the root arguments
+ * @return     Must be >= 0 in all cases.
+ */
+static int co_get_costs_loop_depth(const ir_node *root, int pos)
+{
+	ir_node *block = get_nodes_block(root);
 	ir_loop *loop;
-	ir_node *root_block = get_nodes_block(root);
-	(void) co;
-	(void) arg;
+	int      cost;
 
 	if (is_Phi(root)) {
-		/* for phis the copies are placed in the corresponding pred-block */
-		loop = get_irn_loop(get_Block_cfgpred_block(root_block, pos));
-	} else {
-		/* a perm places the copy in the same block as it resides */
-		loop = get_irn_loop(root_block);
+		block = get_Block_cfgpred_block(block, pos);
 	}
+	loop = get_irn_loop(block);
 	if (loop) {
 		int d = get_loop_depth(loop);
 		cost = d*d;
+	} else {
+		cost = 0;
 	}
 	return 1+cost;
 }
 
-int co_get_costs_exec_freq(const copy_opt_t *co, ir_node *root, ir_node* arg, int pos) {
-	int res;
+static ir_execfreq_int_factors factors;
+
+/**
+ * Computes the costs of a copy according to execution frequency
+ * @param pos  the argument position of arg in the root arguments
+ * @return Must be >= 0 in all cases.
+ */
+static int co_get_costs_exec_freq(const ir_node *root, int pos)
+{
 	ir_node *root_bl = get_nodes_block(root);
-	ir_node *copy_bl = is_Phi(root) ? get_Block_cfgpred_block(root_bl, pos) : root_bl;
-	(void) arg;
-	res = get_block_execfreq_ulong(co->cenv->birg->exec_freq, copy_bl);
+	ir_node *copy_bl
+		= is_Phi(root) ? get_Block_cfgpred_block(root_bl, pos) : root_bl;
+	int      res     = get_block_execfreq_int(&factors, copy_bl);
 
 	/* don't allow values smaller than one. */
 	return res < 1 ? 1 : res;
 }
 
-
-int co_get_costs_all_one(const copy_opt_t *co, ir_node *root, ir_node *arg, int pos) {
-	(void) co;
+/**
+ * All costs equal 1. Using this will reduce the _number_ of copies.
+ * @param co   The copy opt object.
+ * @return Must be >= 0 in all cases.
+ */
+static int co_get_costs_all_one(const ir_node *root, int pos)
+{
 	(void) root;
-	(void) arg;
 	(void) pos;
 	return 1;
 }
@@ -295,13 +325,13 @@ int co_get_costs_all_one(const copy_opt_t *co, ir_node *root, ir_node *arg, int
  * Determines a maximum weighted independent set with respect to
  * the interference and conflict edges of all nodes in a qnode.
  */
-static int ou_max_ind_set_costs(unit_t *ou) {
+static int ou_max_ind_set_costs(unit_t *ou)
+{
 	be_chordal_env_t *chordal_env = ou->co->cenv;
 	ir_node **safe, **unsafe;
 	int i, o, safe_count, safe_costs, unsafe_count, *unsafe_costs;
 	bitset_t *curr;
-	bitset_pos_t pos;
-	int max, curr_weight, best_weight = 0;
+	int curr_weight, best_weight = 0;
 
 	/* assign the nodes into two groups.
 	 * safe: node has no interference, hence it is in every max stable set.
@@ -313,9 +343,9 @@ static int ou_max_ind_set_costs(unit_t *ou) {
 	unsafe       = ALLOCAN(ir_node*, ou->node_count - 1);
 	unsafe_costs = ALLOCAN(int,      ou->node_count - 1);
 	unsafe_count = 0;
-	for(i=1; i<ou->node_count; ++i) {
+	for (i=1; i<ou->node_count; ++i) {
 		int is_safe = 1;
-		for(o=1; o<ou->node_count; ++o) {
+		for (o=1; o<ou->node_count; ++o) {
 			if (i==o)
 				continue;
 			if (nodes_interfere(chordal_env, ou->nodes[i], ou->nodes[o])) {
@@ -336,7 +366,7 @@ static int ou_max_ind_set_costs(unit_t *ou) {
 	/* now compute the best set out of the unsafe nodes*/
 	if (unsafe_count > MIS_HEUR_TRIGGER) {
 		bitset_t *best = bitset_alloca(unsafe_count);
-		/* Heuristik: Greedy trial and error form index 0 to unsafe_count-1 */
+		/* Heuristic: Greedy trial and error form index 0 to unsafe_count-1 */
 		for (i=0; i<unsafe_count; ++i) {
 			bitset_set(best, i);
 			/* check if it is a stable set */
@@ -353,7 +383,7 @@ static int ou_max_ind_set_costs(unit_t *ou) {
 		/* Exact Algorithm: Brute force */
 		curr = bitset_alloca(unsafe_count);
 		bitset_set_all(curr);
-		while ((max = bitset_popcnt(curr)) != 0) {
+		while (bitset_popcount(curr) != 0) {
 			/* check if curr is a stable set */
 			for (i=bitset_next_set(curr, 0); i!=-1; i=bitset_next_set(curr, i+1))
 				for (o=bitset_next_set(curr, i+1); o!=-1; o=bitset_next_set(curr, o+1)) /* !!!!! difference to qnode_max_ind_set(): NOT (curr, i) */
@@ -361,7 +391,7 @@ static int ou_max_ind_set_costs(unit_t *ou) {
 							goto no_stable_set;
 
 			/* if we arrive here, we have a stable set */
-			/* compute the weigth of the stable set*/
+			/* compute the weight of the stable set*/
 			curr_weight = 0;
 			bitset_foreach(curr, pos)
 				curr_weight += unsafe_costs[pos];
@@ -381,10 +411,13 @@ static int ou_max_ind_set_costs(unit_t *ou) {
 
 static void co_collect_units(ir_node *irn, void *env)
 {
-	const arch_register_req_t *req = arch_get_register_req_out(irn);
-	copy_opt_t                *co  = env;
+	const arch_register_req_t *req;
+	copy_opt_t                *co  = (copy_opt_t*)env;
 	unit_t *unit;
 
+	if (get_irn_mode(irn) == mode_T)
+		return;
+	req = arch_get_irn_register_req(irn);
 	if (req->cls != co->cls)
 		return;
 	if (!co_is_optimizable_root(irn))
@@ -411,11 +444,11 @@ static void co_collect_units(ir_node *irn, void *env)
 			int o, arg_pos;
 			ir_node *arg = get_irn_n(irn, i);
 
-			assert(arch_get_irn_reg_class_out(arg) == co->cls && "Argument not in same register class.");
+			assert(arch_get_irn_reg_class(arg) == co->cls && "Argument not in same register class.");
 			if (arg == irn)
 				continue;
 			if (nodes_interfere(co->cenv, irn, arg)) {
-				unit->inevitable_costs += co->get_costs(co, irn, arg, i);
+				unit->inevitable_costs += co->get_costs(irn, i);
 				continue;
 			}
 
@@ -437,11 +470,11 @@ static void co_collect_units(ir_node *irn, void *env)
 			if (!arg_pos) { /* a new argument */
 				/* insert node, set costs */
 				unit->nodes[unit->node_count] = arg;
-				unit->costs[unit->node_count] = co->get_costs(co, irn, arg, i);
+				unit->costs[unit->node_count] = co->get_costs(irn, i);
 				unit->node_count++;
 			} else { /* arg has occurred before in same phi */
 				/* increase costs for existing arg */
-				unit->costs[arg_pos] += co->get_costs(co, irn, arg, i);
+				unit->costs[arg_pos] += co->get_costs(irn, i);
 			}
 		}
 		unit->nodes = XREALLOC(unit->nodes, ir_node*, unit->node_count);
@@ -454,7 +487,7 @@ static void co_collect_units(ir_node *irn, void *env)
 		unit->node_count = 2;
 		unit->nodes[0] = irn;
 		unit->nodes[1] = get_Perm_src(irn);
-		unit->costs[1] = co->get_costs(co, irn, unit->nodes[1], -1);
+		unit->costs[1] = co->get_costs(irn, -1);
 	} else {
 		/* Src == Tgt of a 2-addr-code instruction */
 		if (is_2addr_code(req)) {
@@ -487,7 +520,7 @@ static void co_collect_units(ir_node *irn, void *env)
 						if (!arch_irn_is_ignore(o) &&
 								!nodes_interfere(co->cenv, irn, o)) {
 							unit->nodes[k] = o;
-							unit->costs[k] = co->get_costs(co, irn, o, -1);
+							unit->costs[k] = co->get_costs(irn, -1);
 							++k;
 						}
 					}
@@ -504,7 +537,7 @@ static void co_collect_units(ir_node *irn, void *env)
 		struct list_head *tmp;
 
 		/* Determine the maximum costs this unit can cause: all_nodes_cost */
-		for(i=1; i<unit->node_count; ++i) {
+		for (i=1; i<unit->node_count; ++i) {
 			unit->sort_key = MAX(unit->sort_key, unit->costs[i]);
 			unit->all_nodes_costs += unit->costs[i];
 		}
@@ -523,7 +556,8 @@ static void co_collect_units(ir_node *irn, void *env)
 
 #ifdef QUICK_AND_DIRTY_HACK
 
-static int compare_ous(const void *k1, const void *k2) {
+static int compare_ous(const void *k1, const void *k2)
+{
 	const unit_t *u1 = *((const unit_t **) k1);
 	const unit_t *u2 = *((const unit_t **) k2);
 	int i, o, u1_has_constr, u2_has_constr;
@@ -532,7 +566,7 @@ static int compare_ous(const void *k1, const void *k2) {
 	/* Units with constraints come first */
 	u1_has_constr = 0;
 	for (i=0; i<u1->node_count; ++i) {
-		arch_get_register_req_out(&req, u1->nodes[i]);
+		arch_get_irn_register_req(&req, u1->nodes[i]);
 		if (arch_register_req_is(&req, limited)) {
 			u1_has_constr = 1;
 			break;
@@ -541,7 +575,7 @@ static int compare_ous(const void *k1, const void *k2) {
 
 	u2_has_constr = 0;
 	for (i=0; i<u2->node_count; ++i) {
-		arch_get_register_req_out(&req, u2->nodes[i]);
+		arch_get_irn_register_req(&req, u2->nodes[i]);
 		if (arch_register_req_is(&req, limited)) {
 			u2_has_constr = 1;
 			break;
@@ -567,9 +601,10 @@ static int compare_ous(const void *k1, const void *k2) {
 /**
  * Sort the ou's according to constraints and their sort_key
  */
-static void co_sort_units(copy_opt_t *co) {
+static void co_sort_units(copy_opt_t *co)
+{
 	int i, count = 0, costs;
-	unit_t *ou, **ous;
+	unit_t **ous;
 
 	/* get the number of ous, remove them form the list and fill the array */
 	list_for_each_entry(unit_t, ou, &co->units, units)
@@ -603,7 +638,8 @@ static void co_sort_units(copy_opt_t *co) {
 }
 #endif
 
-void co_build_ou_structure(copy_opt_t *co) {
+void co_build_ou_structure(copy_opt_t *co)
+{
 	DBG((dbg, LEVEL_1, "\tCollecting optimization units\n"));
 	INIT_LIST_HEAD(&co->units);
 	irg_walk_graph(co->irg, co_collect_units, NULL, co);
@@ -612,8 +648,8 @@ void co_build_ou_structure(copy_opt_t *co) {
 #endif
 }
 
-void co_free_ou_structure(copy_opt_t *co) {
-	unit_t *curr, *tmp;
+void co_free_ou_structure(copy_opt_t *co)
+{
 	ASSERT_OU_AVAIL(co);
 	list_for_each_entry_safe(unit_t, curr, tmp, &co->units, units) {
 		xfree(curr->nodes);
@@ -625,9 +661,9 @@ void co_free_ou_structure(copy_opt_t *co) {
 
 /* co_solve_heuristic() is implemented in becopyheur.c */
 
-int co_get_max_copy_costs(const copy_opt_t *co) {
+int co_get_max_copy_costs(const copy_opt_t *co)
+{
 	int i, res = 0;
-	unit_t *curr;
 
 	ASSERT_OU_AVAIL(co);
 
@@ -639,9 +675,9 @@ int co_get_max_copy_costs(const copy_opt_t *co) {
 	return res;
 }
 
-int co_get_inevit_copy_costs(const copy_opt_t *co) {
+int co_get_inevit_copy_costs(const copy_opt_t *co)
+{
 	int res = 0;
-	unit_t *curr;
 
 	ASSERT_OU_AVAIL(co);
 
@@ -650,9 +686,9 @@ int co_get_inevit_copy_costs(const copy_opt_t *co) {
 	return res;
 }
 
-int co_get_copy_costs(const copy_opt_t *co) {
+int co_get_copy_costs(const copy_opt_t *co)
+{
 	int i, res = 0;
-	unit_t *curr;
 
 	ASSERT_OU_AVAIL(co);
 
@@ -671,9 +707,9 @@ int co_get_copy_costs(const copy_opt_t *co) {
 	return res;
 }
 
-int co_get_lower_bound(const copy_opt_t *co) {
+int co_get_lower_bound(const copy_opt_t *co)
+{
 	int res = 0;
-	unit_t *curr;
 
 	ASSERT_OU_AVAIL(co);
 
@@ -684,18 +720,16 @@ int co_get_lower_bound(const copy_opt_t *co) {
 
 void co_complete_stats(const copy_opt_t *co, co_complete_stats_t *stat)
 {
-	bitset_t *seen = bitset_irg_malloc(co->irg);
-	affinity_node_t *an;
+	bitset_t *seen = bitset_malloc(get_irg_last_idx(co->irg));
 
 	memset(stat, 0, sizeof(stat[0]));
 
 	/* count affinity edges. */
 	co_gs_foreach_aff_node(co, an) {
-		neighb_t *neigh;
 		stat->aff_nodes += 1;
-		bitset_add_irn(seen, an->irn);
+		bitset_set(seen, get_irn_idx(an->irn));
 		co_gs_foreach_neighb(an, neigh) {
-			if(!bitset_contains_irn(seen, neigh->irn)) {
+			if (!bitset_is_set(seen, get_irn_idx(neigh->irn))) {
 				stat->aff_edges += 1;
 				stat->max_costs += neigh->costs;
 
@@ -704,7 +738,7 @@ void co_complete_stats(const copy_opt_t *co, co_complete_stats_t *stat)
 					stat->unsatisfied_edges += 1;
 				}
 
-				if(nodes_interfere(co->cenv, an->irn, neigh->irn)) {
+				if (nodes_interfere(co->cenv, an->irn, neigh->irn)) {
 					stat->aff_int += 1;
 					stat->inevit_costs += neigh->costs;
 				}
@@ -727,15 +761,17 @@ void co_complete_stats(const copy_opt_t *co, co_complete_stats_t *stat)
                   |_|                                      |___/
  ******************************************************************************/
 
-static int compare_affinity_node_t(const void *k1, const void *k2, size_t size) {
-	const affinity_node_t *n1 = k1;
-	const affinity_node_t *n2 = k2;
+static int compare_affinity_node_t(const void *k1, const void *k2, size_t size)
+{
+	const affinity_node_t *n1 = (const affinity_node_t*)k1;
+	const affinity_node_t *n2 = (const affinity_node_t*)k2;
 	(void) size;
 
 	return (n1->irn != n2->irn);
 }
 
-static void add_edge(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs) {
+static void add_edge(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs)
+{
 	affinity_node_t new_node, *node;
 	neighb_t        *nbr;
 	int             allocnew = 1;
@@ -743,7 +779,7 @@ static void add_edge(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs) {
 	new_node.irn        = n1;
 	new_node.degree     = 0;
 	new_node.neighbours = NULL;
-	node = set_insert(co->nodes, &new_node, sizeof(new_node), hash_irn(new_node.irn));
+	node = set_insert(affinity_node_t, co->nodes, &new_node, sizeof(new_node), hash_irn(new_node.irn));
 
 	for (nbr = node->neighbours; nbr; nbr = nbr->next)
 		if (nbr->irn == n2) {
@@ -753,7 +789,7 @@ static void add_edge(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs) {
 
 	/* if we did not find n2 in n1's neighbourhood insert it */
 	if (allocnew) {
-		nbr        = obstack_alloc(&co->obst, sizeof(*nbr));
+		nbr        = OALLOC(&co->obst, neighb_t);
 		nbr->irn   = n2;
 		nbr->costs = 0;
 		nbr->next  = node->neighbours;
@@ -766,34 +802,34 @@ static void add_edge(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs) {
 	nbr->costs += costs;
 }
 
-static inline void add_edges(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs) {
+static inline void add_edges(copy_opt_t *co, ir_node *n1, ir_node *n2, int costs)
+{
 	if (! be_ifg_connected(co->cenv->ifg, n1, n2)) {
 		add_edge(co, n1, n2, costs);
 		add_edge(co, n2, n1, costs);
 	}
 }
 
-static void build_graph_walker(ir_node *irn, void *env) {
-	const arch_register_req_t *req = arch_get_register_req_out(irn);
-	copy_opt_t                *co  = env;
+static void build_graph_walker(ir_node *irn, void *env)
+{
+	const arch_register_req_t *req;
+	copy_opt_t                *co  = (copy_opt_t*)env;
 	int pos, max;
-	const arch_register_t *reg;
 
-	if (req->cls != co->cls || arch_irn_is_ignore(irn))
+	if (get_irn_mode(irn) == mode_T)
 		return;
-
-	reg = arch_get_irn_register(irn);
-	if (arch_register_type_is(reg, ignore))
+	req = arch_get_irn_register_req(irn);
+	if (req->cls != co->cls || arch_irn_is_ignore(irn))
 		return;
 
 	if (is_Reg_Phi(irn)) { /* Phis */
 		for (pos=0, max=get_irn_arity(irn); pos<max; ++pos) {
 			ir_node *arg = get_irn_n(irn, pos);
-			add_edges(co, irn, arg, co->get_costs(co, irn, arg, pos));
+			add_edges(co, irn, arg, co->get_costs(irn, pos));
 		}
 	} else if (is_Perm_Proj(irn)) { /* Perms */
 		ir_node *arg = get_Perm_src(irn);
-		add_edges(co, irn, arg, co->get_costs(co, irn, arg, 0));
+		add_edges(co, irn, arg, co->get_costs(irn, -1));
 	} else { /* 2-address code */
 		if (is_2addr_code(req)) {
 			const unsigned other = req->other_same;
@@ -803,21 +839,23 @@ static void build_graph_walker(ir_node *irn, void *env) {
 				if (other & (1U << i)) {
 					ir_node *other = get_irn_n(skip_Proj(irn), i);
 					if (!arch_irn_is_ignore(other))
-						add_edges(co, irn, other, co->get_costs(co, irn, other, 0));
+						add_edges(co, irn, other, co->get_costs(irn, -1));
 				}
 			}
 		}
 	}
 }
 
-void co_build_graph_structure(copy_opt_t *co) {
+void co_build_graph_structure(copy_opt_t *co)
+{
 	obstack_init(&co->obst);
 	co->nodes = new_set(compare_affinity_node_t, 32);
 
 	irg_walk_graph(co->irg, build_graph_walker, NULL, co);
 }
 
-void co_free_graph_structure(copy_opt_t *co) {
+void co_free_graph_structure(copy_opt_t *co)
+{
 	ASSERT_GS_AVAIL(co);
 
 	del_set(co->nodes);
@@ -825,15 +863,14 @@ void co_free_graph_structure(copy_opt_t *co) {
 	co->nodes = NULL;
 }
 
-/* co_solve_ilp1() co_solve_ilp2() are implemented in becopyilpX.c */
-
-int co_gs_is_optimizable(copy_opt_t *co, ir_node *irn) {
+int co_gs_is_optimizable(copy_opt_t *co, ir_node *irn)
+{
 	affinity_node_t new_node, *n;
 
 	ASSERT_GS_AVAIL(co);
 
 	new_node.irn = irn;
-	n = set_find(co->nodes, &new_node, sizeof(new_node), hash_irn(new_node.irn));
+	n = set_find(affinity_node_t, co->nodes, &new_node, sizeof(new_node), hash_irn(new_node.irn));
 	if (n) {
 		return (n->degree > 0);
 	} else
@@ -845,37 +882,51 @@ static int co_dump_appel_disjoint_constraints(const copy_opt_t *co, ir_node *a,
 	ir_node *nodes[]  = { a, b };
 	bitset_t *constr[] = { NULL, NULL };
 	int j;
-	
+
 	constr[0] = bitset_alloca(co->cls->n_regs);
 	constr[1] = bitset_alloca(co->cls->n_regs);
 
 	for (j = 0; j < 2; ++j) {
-		const arch_register_req_t *req = arch_get_register_req_out(nodes[j]);
-		if(arch_register_req_is(req, limited)) 
+		const arch_register_req_t *req = arch_get_irn_register_req(nodes[j]);
+		if (arch_register_req_is(req, limited))
 			rbitset_copy_to_bitset(req->limited, constr[j]);
 		else
 			bitset_set_all(constr[j]);
-		
+
 	}
 
 	return !bitset_intersect(constr[0], constr[1]);
 }
 
-void co_dump_appel_graph(const copy_opt_t *co, FILE *f)
+/**
+ * Dump the interference graph according to the Appel/George coalescing contest file format.
+ * See: http://www.cs.princeton.edu/~appel/coalesce/format.html
+ * @note Requires graph structure.
+ * @param co The copy opt object.
+ * @param f  A file to dump to.
+ */
+static void co_dump_appel_graph(const copy_opt_t *co, FILE *f)
 {
 	be_ifg_t *ifg       = co->cenv->ifg;
 	int      *color_map = ALLOCAN(int, co->cls->n_regs);
 	int      *node_map  = XMALLOCN(int, get_irg_last_idx(co->irg) + 1);
+	ir_graph *irg       = co->irg;
+	be_irg_t *birg      = be_birg_from_irg(irg);
 
 	ir_node *irn;
-	void *it, *nit;
+	nodes_iter_t it;
+	neighbours_iter_t nit;
 	int n, n_regs;
 	unsigned i;
 
 	n_regs = 0;
-	for(i = 0; i < co->cls->n_regs; ++i) {
+	for (i = 0; i < co->cls->n_regs; ++i) {
 		const arch_register_t *reg = &co->cls->regs[i];
-		color_map[i] = arch_register_type_is(reg, ignore) ? -1 : n_regs++;
+		if (rbitset_is_set(birg->allocatable_regs, reg->global_index)) {
+			color_map[i] = n_regs++;
+		} else {
+			color_map[i] = -1;
+		}
 	}
 
 	/*
@@ -883,11 +934,8 @@ void co_dump_appel_graph(const copy_opt_t *co, FILE *f)
 	 * the values below n are the pre-colored register nodes
 	 */
 
-	it  = be_ifg_nodes_iter_alloca(ifg);
-	nit = be_ifg_neighbours_iter_alloca(ifg);
-
 	n = n_regs;
-	be_ifg_foreach_node(ifg, it, irn) {
+	be_ifg_foreach_node(ifg, &it, irn) {
 		if (arch_irn_is_ignore(irn))
 			continue;
 		node_map[get_irn_idx(irn)] = n++;
@@ -895,36 +943,34 @@ void co_dump_appel_graph(const copy_opt_t *co, FILE *f)
 
 	fprintf(f, "%d %d\n", n, n_regs);
 
-	be_ifg_foreach_node(ifg, it, irn) {
+	be_ifg_foreach_node(ifg, &it, irn) {
 		if (!arch_irn_is_ignore(irn)) {
 			int idx                        = node_map[get_irn_idx(irn)];
 			affinity_node_t           *a   = get_affinity_info(co, irn);
-			const arch_register_req_t *req = arch_get_register_req_out(irn);
+			const arch_register_req_t *req = arch_get_irn_register_req(irn);
 			ir_node                   *adj;
 
-			if(arch_register_req_is(req, limited)) {
-				for(i = 0; i < co->cls->n_regs; ++i) {
-					if(!rbitset_is_set(req->limited, i) && color_map[i] >= 0)
+			if (arch_register_req_is(req, limited)) {
+				for (i = 0; i < co->cls->n_regs; ++i) {
+					if (!rbitset_is_set(req->limited, i) && color_map[i] >= 0)
 						fprintf(f, "%d %d -1\n", color_map[i], idx);
 				}
 			}
 
-			be_ifg_foreach_neighbour(ifg, nit, irn, adj) {
+			be_ifg_foreach_neighbour(ifg, &nit, irn, adj) {
 				if (!arch_irn_is_ignore(adj) &&
 						!co_dump_appel_disjoint_constraints(co, irn, adj)) {
 					int adj_idx = node_map[get_irn_idx(adj)];
-					if(idx < adj_idx)
+					if (idx < adj_idx)
 						fprintf(f, "%d %d -1\n", idx, adj_idx);
 				}
 			}
 
-			if(a) {
-				neighb_t *n;
-
+			if (a) {
 				co_gs_foreach_neighb(a, n) {
 					if (!arch_irn_is_ignore(n->irn)) {
 						int n_idx = node_map[get_irn_idx(n->irn)];
-						if(idx < n_idx)
+						if (idx < n_idx)
 							fprintf(f, "%d %d %d\n", idx, n_idx, (int) n->costs);
 					}
 				}
@@ -935,212 +981,18 @@ void co_dump_appel_graph(const copy_opt_t *co, FILE *f)
 	xfree(node_map);
 }
 
-/*
-	 ___ _____ ____   ____   ___ _____   ____                        _
-	|_ _|  ___/ ___| |  _ \ / _ \_   _| |  _ \ _   _ _ __ ___  _ __ (_)_ __   __ _
-	 | || |_ | |  _  | | | | | | || |   | | | | | | | '_ ` _ \| '_ \| | '_ \ / _` |
-	 | ||  _|| |_| | | |_| | |_| || |   | |_| | |_| | | | | | | |_) | | | | | (_| |
-	|___|_|   \____| |____/ \___/ |_|   |____/ \__,_|_| |_| |_| .__/|_|_| |_|\__, |
-	                                                          |_|            |___/
-*/
-
-static const char *get_dot_color_name(size_t col)
-{
-	static const char *names[] = {
-		"blue",
-		"red",
-		"green",
-		"yellow",
-		"cyan",
-		"magenta",
-		"orange",
-		"chocolate",
-		"beige",
-		"navy",
-		"darkgreen",
-		"darkred",
-		"lightPink",
-		"chartreuse",
-		"lightskyblue",
-		"linen",
-		"pink",
-		"lightslateblue",
-		"mintcream",
-		"red",
-		"darkolivegreen",
-		"mediumblue",
-		"mistyrose",
-		"salmon",
-		"darkseagreen",
-		"mediumslateblue"
-		"moccasin",
-		"tomato",
-		"forestgreen",
-		"darkturquoise",
-		"palevioletred"
-	};
-
-	return col < sizeof(names)/sizeof(names[0]) ? names[col] : "white";
-}
-
-typedef struct _co_ifg_dump_t {
-	const copy_opt_t *co;
-	unsigned flags;
-} co_ifg_dump_t;
-
-static void ifg_dump_graph_attr(FILE *f, void *self)
-{
-	(void) self;
-	fprintf(f, "overlap=scale");
-}
-
-static int ifg_is_dump_node(void *self, ir_node *irn)
-{
-	(void)self;
-	return !arch_irn_is_ignore(irn);
-}
-
-static void ifg_dump_node_attr(FILE *f, void *self, ir_node *irn)
-{
-	co_ifg_dump_t             *env     = self;
-	const arch_register_t     *reg     = arch_get_irn_register(irn);
-	const arch_register_req_t *req     = arch_get_register_req_out(irn);
-	int                        limited = arch_register_req_is(req, limited);
-
-	if(env->flags & CO_IFG_DUMP_LABELS) {
-		ir_fprintf(f, "label=\"%+F", irn);
-
-		if((env->flags & CO_IFG_DUMP_CONSTR) && limited) {
-			bitset_t *bs = bitset_alloca(env->co->cls->n_regs);
-			rbitset_copy_to_bitset(req->limited, bs);
-			ir_fprintf(f, "\\n%B", bs);
-		}
-		ir_fprintf(f, "\" ");
-	} else {
-		fprintf(f, "label=\"\" shape=point " );
-	}
-
-	if(env->flags & CO_IFG_DUMP_SHAPE)
-		fprintf(f, "shape=%s ", limited ? "diamond" : "ellipse");
-
-	if(env->flags & CO_IFG_DUMP_COLORS)
-		fprintf(f, "style=filled color=%s ", get_dot_color_name(reg->index));
-}
-
-static void ifg_dump_at_end(FILE *file, void *self)
-{
-	co_ifg_dump_t *env = self;
-	affinity_node_t *a;
-
-	co_gs_foreach_aff_node(env->co, a) {
-		const arch_register_t *ar = arch_get_irn_register(a->irn);
-		unsigned aidx = get_irn_idx(a->irn);
-		neighb_t *n;
-
-		co_gs_foreach_neighb(a, n) {
-			const arch_register_t *nr = arch_get_irn_register(n->irn);
-			unsigned nidx = get_irn_idx(n->irn);
-
-			if(aidx < nidx) {
-				const char *color = nr == ar ? "blue" : "red";
-				fprintf(file, "\tn%d -- n%d [weight=0.01 ", aidx, nidx);
-				if(env->flags & CO_IFG_DUMP_LABELS)
-					fprintf(file, "label=\"%d\" ", n->costs);
-				if(env->flags & CO_IFG_DUMP_COLORS)
-					fprintf(file, "color=%s ", color);
-				else
-					fprintf(file, "style=dotted");
-				fprintf(file, "];\n");
-			}
-		}
-	}
-}
-
-
-static be_ifg_dump_dot_cb_t ifg_dot_cb = {
-	ifg_is_dump_node,
-	ifg_dump_graph_attr,
-	ifg_dump_node_attr,
-	NULL,
-	NULL,
-	ifg_dump_at_end
-};
-
-
-
-void co_dump_ifg_dot(const copy_opt_t *co, FILE *f, unsigned flags)
-{
-	co_ifg_dump_t cod;
-
-	cod.co    = co;
-	cod.flags = flags;
-	be_ifg_dump_dot(co->cenv->ifg, co->irg, f, &ifg_dot_cb, &cod);
-}
-
-
-void co_solve_park_moon(copy_opt_t *opt)
-{
-	(void) opt;
-}
-
-static int void_algo(copy_opt_t *co)
-{
-	(void) co;
-	return 0;
-}
-
-/*
-		_    _                  _ _   _
-	   / \  | | __ _  ___  _ __(_) |_| |__  _ __ ___  ___
-	  / _ \ | |/ _` |/ _ \| '__| | __| '_ \| '_ ` _ \/ __|
-	 / ___ \| | (_| | (_) | |  | | |_| | | | | | | | \__ \
-	/_/   \_\_|\__, |\___/|_|  |_|\__|_| |_|_| |_| |_|___/
-			   |___/
-*/
-
-typedef struct {
-	co_algo_t  *algo;
-	const char *name;
-	int        can_improve_existing;
-} co_algo_info_t;
-
-static co_algo_info_t algos[] = {
-	{ void_algo,               "none",  0 },
-	{ co_solve_heuristic,      "heur1", 0 },
-	{ co_solve_heuristic_new,  "heur2", 0 },
-#ifdef WITH_JVM
-	{ co_solve_heuristic_java, "heur3", 0 },
-#else
-	{ NULL,                    "heur3", 0 },
-#endif
-	{ co_solve_heuristic_mst,  "heur4", 0 },
-#ifdef WITH_ILP
-	{ co_solve_ilp2,           "ilp",   1 },
-#else
-	{ NULL,                    "ilp",   1 },
-#endif
-	{ NULL,                    "",      0 }
-};
-
-/*
-    __  __       _         ____       _
-   |  \/  | __ _(_)_ __   |  _ \ _ __(_)_   _____ _ __
-   | |\/| |/ _` | | '_ \  | | | | '__| \ \ / / _ \ '__|
-   | |  | | (_| | | | | | | |_| | |  | |\ V /  __/ |
-   |_|  |_|\__,_|_|_| |_| |____/|_|  |_| \_/ \___|_|
-
-*/
-
-static FILE *my_open(const be_chordal_env_t *env, const char *prefix, const char *suffix)
+static FILE *my_open(const be_chordal_env_t *env, const char *prefix,
+                     const char *suffix)
 {
 	FILE *result;
 	char buf[1024];
 	size_t i, n;
 	char *tu_name;
+	const char *cup_name = be_get_irg_main_env(env->irg)->cup_name;
 
-	n = strlen(env->birg->main_env->cup_name);
+	n = strlen(cup_name);
 	tu_name = XMALLOCN(char, n + 1);
-	strcpy(tu_name, env->birg->main_env->cup_name);
+	strcpy(tu_name, cup_name);
 	for (i = 0; i < n; ++i)
 		if (tu_name[i] == '.')
 			tu_name[i] = '_';
@@ -1149,7 +1001,7 @@ static FILE *my_open(const be_chordal_env_t *env, const char *prefix, const char
 	ir_snprintf(buf, sizeof(buf), "%s%s_%F_%s%s", prefix, tu_name, env->irg, env->cls->name, suffix);
 	xfree(tu_name);
 	result = fopen(buf, "wt");
-	if(result == NULL) {
+	if (result == NULL) {
 		panic("Couldn't open '%s' for writing.", buf);
 	}
 
@@ -1158,16 +1010,18 @@ static FILE *my_open(const be_chordal_env_t *env, const char *prefix, const char
 
 void co_driver(be_chordal_env_t *cenv)
 {
-	ir_timer_t          *timer = ir_timer_register("firm.be.copyopt", "runtime");
+	ir_timer_t          *timer = ir_timer_new();
 	co_complete_stats_t before, after;
 	copy_opt_t          *co;
-	co_algo_t           *algo_func;
 	int                 was_optimal = 0;
 
-	if (algo >= CO_ALGO_LAST)
+	assert(selected_copyopt);
+
+	/* skip copymin if algo is 'none' */
+	if (selected_copyopt->copyopt == void_algo)
 		return;
 
-	be_liveness_assure_chk(be_get_birg_liveness(cenv->birg));
+	be_assure_live_chk(cenv->irg);
 
 	co = new_copy_opt(cenv, cost_func);
 	co_build_ou_structure(co);
@@ -1175,84 +1029,73 @@ void co_driver(be_chordal_env_t *cenv)
 
 	co_complete_stats(co, &before);
 
-	be_stat_ev_ull("co_aff_nodes",    before.aff_nodes);
-	be_stat_ev_ull("co_aff_edges",    before.aff_edges);
-	be_stat_ev_ull("co_max_costs",    before.max_costs);
-	be_stat_ev_ull("co_inevit_costs", before.inevit_costs);
-	be_stat_ev_ull("co_aff_int",      before.aff_int);
+	stat_ev_ull("co_aff_nodes",    before.aff_nodes);
+	stat_ev_ull("co_aff_edges",    before.aff_edges);
+	stat_ev_ull("co_max_costs",    before.max_costs);
+	stat_ev_ull("co_inevit_costs", before.inevit_costs);
+	stat_ev_ull("co_aff_int",      before.aff_int);
 
-	be_stat_ev_ull("co_init_costs",   before.costs);
-	be_stat_ev_ull("co_init_unsat",   before.unsatisfied_edges);
+	stat_ev_ull("co_init_costs",   before.costs);
+	stat_ev_ull("co_init_unsat",   before.unsatisfied_edges);
 
 	if (dump_flags & DUMP_BEFORE) {
-		FILE *f = my_open(cenv, "", "-before.dot");
-		co_dump_ifg_dot(co, f, style_flags);
+		FILE *f = my_open(cenv, "", "-before.vcg");
+		be_dump_ifg_co(f, co, style_flags & CO_IFG_DUMP_LABELS, style_flags & CO_IFG_DUMP_COLORS);
 		fclose(f);
 	}
 
-	/* if the algo can improve results, provide an initial solution with heur3 */
-	if (improve && algos[algo].can_improve_existing) {
+	/* if the algo can improve results, provide an initial solution with heur1 */
+	if (improve && selected_copyopt->can_improve_existing) {
 		co_complete_stats_t stats;
 
 		/* produce a heuristic solution */
-#ifdef WITH_JVM
-		co_solve_heuristic_java(co);
-#else
 		co_solve_heuristic(co);
-#endif /* WITH_JVM */
 
 		/* do the stats and provide the current costs */
 		co_complete_stats(co, &stats);
-		be_stat_ev_ull("co_prepare_costs", stats.costs);
+		stat_ev_ull("co_prepare_costs", stats.costs);
 	}
 
-#ifdef WITH_JVM
-	/* start the JVM here so that it does not tamper the timing. */
-	if (algo == CO_ALGO_HEUR3)
-		be_java_coal_start_jvm();
-#endif /* WITH_JVM */
-
-	algo_func = algos[algo].algo;
-
 	/* perform actual copy minimization */
 	ir_timer_reset_and_start(timer);
-	was_optimal = algo_func(co);
+	was_optimal = selected_copyopt->copyopt(co);
 	ir_timer_stop(timer);
 
-	be_stat_ev("co_time", ir_timer_elapsed_msec(timer));
-	be_stat_ev_ull("co_optimal", was_optimal);
+	stat_ev_dbl("co_time", ir_timer_elapsed_msec(timer));
+	stat_ev_ull("co_optimal", was_optimal);
+	ir_timer_free(timer);
 
 	if (dump_flags & DUMP_AFTER) {
-		FILE *f = my_open(cenv, "", "-after.dot");
-		co_dump_ifg_dot(co, f, style_flags);
+		FILE *f = my_open(cenv, "", "-after.vcg");
+		be_dump_ifg_co(f, co, style_flags & CO_IFG_DUMP_LABELS, style_flags & CO_IFG_DUMP_COLORS);
 		fclose(f);
 	}
 
 	co_complete_stats(co, &after);
 
 	if (do_stats) {
-		ulong64 optimizable_costs = after.max_costs - after.inevit_costs;
-		ulong64 evitable          = after.costs     - after.inevit_costs;
+		unsigned long long optimizable_costs = after.max_costs - after.inevit_costs;
+		unsigned long long evitable          = after.costs     - after.inevit_costs;
 
 		ir_printf("%30F ", cenv->irg);
-		printf("%10s %10" ULL_FMT "%10" ULL_FMT "%10" ULL_FMT, cenv->cls->name, after.max_costs, before.costs, after.inevit_costs);
+		printf("%10s %10llu%10llu%10llu", cenv->cls->name, after.max_costs, before.costs, after.inevit_costs);
 
-		if(optimizable_costs > 0)
-			printf("%10" ULL_FMT " %5.2f\n", after.costs, (evitable * 100.0) / optimizable_costs);
+		if (optimizable_costs > 0)
+			printf("%10llu %5.2f\n", after.costs, (evitable * 100.0) / optimizable_costs);
 		else
-			printf("%10" ULL_FMT " %5s\n", after.costs, "-");
+			printf("%10llu %5s\n", after.costs, "-");
 	}
 
 	/* Dump the interference graph in Appel's format. */
 	if (dump_flags & DUMP_APPEL) {
 		FILE *f = my_open(cenv, "", ".apl");
-		fprintf(f, "# %lld %lld\n", after.costs, after.unsatisfied_edges);
+		fprintf(f, "# %llu %llu\n", after.costs, after.unsatisfied_edges);
 		co_dump_appel_graph(co, f);
 		fclose(f);
 	}
 
-	be_stat_ev_ull("co_after_costs", after.costs);
-	be_stat_ev_ull("co_after_unsat", after.unsatisfied_edges);
+	stat_ev_ull("co_after_costs", after.costs);
+	stat_ev_ull("co_after_unsat", after.unsatisfied_edges);
 
 	co_free_graph_structure(co);
 	co_free_ou_structure(co);
