@@ -22,7 +22,6 @@
  * @brief       Base routines for register allocation.
  * @author      Sebastian Hack
  * @date        13.01.2005
- * @version     $Id: bera.h 17695 2008-02-13 18:00:20Z matze $
  */
 #ifndef FIRM_BE_BERA_H
 #define FIRM_BE_BERA_H
@@ -33,7 +32,7 @@
 #include "beirg.h"
 
 typedef struct be_ra_t {
-	void (*allocate)(be_irg_t *bi);   /**< allocate registers on a graph */
+	void (*allocate)(ir_graph *irg);   /**< allocate registers on a graph */
 } be_ra_t;
 
 void be_register_allocator(const char *name, be_ra_t *allocator);
@@ -41,8 +40,6 @@ void be_register_allocator(const char *name, be_ra_t *allocator);
 /**
  * Do register allocation with currently selected register allocator
  */
-void be_allocate_registers(be_irg_t *birg);
+void be_allocate_registers(ir_graph *irg);
 
-int (values_interfere)(const be_irg_t *birg, const ir_node *a, const ir_node *b);
-
-#endif /* FIRM_BE_BERA_H */
+#endif
