@@ -21,7 +21,6 @@
  * @file
  * @brief   This file implements functions to finalize the irg for emit.
  * @author  Christian Wuerdig
- * @version $Id: ia32_finish.h 23458 2008-11-06 00:36:17Z beck $
  */
 #ifndef FIRM_BE_IA32_IA32_FINISH_H
 #define FIRM_BE_IA32_IA32_FINISH_H
@@ -32,11 +31,10 @@
 /**
  * Check 2-Addresscode constraints and call peephole optimizations.
  * @param irg  The irg to finish
- * @param cg   The codegenerator object for the irg
  */
-void ia32_finish_irg(ir_graph *irg, ia32_code_gen_t *cg);
+void ia32_finish_irg(ir_graph *irg);
 
 /** Initialize the finisher. */
 void ia32_init_finish(void);
 
-#endif /* FIRM_BE_IA32_IA32_FINISH_H */
+#endif
