@@ -21,7 +21,6 @@
  * @file
  * @brief   Handles fpu rounding modes
  * @author  Matthias Braun
- * @version $Id: ia32_fpu.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_BE_IA32_IA32_FPU_H
 #define FIRM_BE_IA32_IA32_FPU_H
@@ -31,6 +30,6 @@
 /**
  * Handle switching of fpu mode
  */
-void ia32_setup_fpu_mode(ia32_code_gen_t *cg);
+void ia32_setup_fpu_mode(ir_graph *irg);
 
-#endif /* FIRM_BE_IA32_IA32_FPU_H */
+#endif
