@@ -22,28 +22,27 @@
  * @brief       This file implements the x87 support and virtual to stack
  *              register translation for the ia32 backend.
  * @author      Michael Beck
- * @version     $Id: ia32_x87.h 23458 2008-11-06 00:36:17Z beck $
  */
 #ifndef FIRM_BE_IA32_IA32_X87_H
 #define FIRM_BE_IA32_IA32_X87_H
 
-#include "../bearch.h"
-#include "../beirg.h"
+#include "bearch.h"
+#include "beirg.h"
 
 /**
  * Run a simulation and fix all virtual instructions for a graph.
  * Replaces all virtual floating point instructions and registers
  * by real ones.
  *
- * @param birg      the graph to simulate and patch
+ * @param irg      the graph to simulate and patch
  *
  * Registers must be allocated.
  */
-void x87_simulate_graph(be_irg_t *birg);
+void ia32_x87_simulate_graph(ir_graph *irg);
 
 /**
  * Initializes the x87 simulator.
  */
 void ia32_init_x87(void);
 
-#endif /* FIRM_BE_IA32_IA32_X87_H */
+#endif
