@@ -23,7 +23,6 @@
 # their coresponding node constructors for all operations in a given spec
 # so they can be used as normal firm nodes.
 # Creation: 2005/10/19
-# $Id: generate_new_opcodes.pl 23535 2008-11-08 10:42:30Z mallon $
 
 use strict;
 use Data::Dumper;
@@ -48,6 +47,7 @@ our $custom_init_attr_func;
 our %compare_attr;
 our %copy_attr;
 our %reg_classes;
+our %custom_irn_flags;
 
 # include spec file
 
@@ -67,15 +67,15 @@ my $target_h = $target_dir."/gen_".$arch."_new_nodes.h";
 if(!defined($default_attr_type)) {
 	$default_attr_type = "${arch}_attr_t";
 }
-if(!defined(%init_attr)) {
+if(! %init_attr) {
 	%init_attr = (
-		"$default_attr_type" => "\tinit_${arch}_attributes(res, flags, in_reqs, out_reqs, exec_units, n_res);",
+		"$default_attr_type" => "\tinit_${arch}_attributes(res, irn_flags_, in_reqs, n_res);",
 	);
 }
 if(!defined($default_cmp_attr)) {
 	$default_cmp_attr = "${arch}_compare_attr";
 }
-if(!defined(%compare_attr)) {
+if(! %compare_attr) {
 	%compare_attr = (
 		"${default_attr_type}" => "${default_cmp_attr}",
 	);
@@ -101,7 +101,7 @@ foreach my $op (keys(%operands)) {
 	$op_node{op_flags} = "O";
 	$op_node{cmp_attr} = 'return 1;';
 	$op_node{mode}     = 'mode_ANY';
-		
+
 	$nodes{$op} = \%op_node;
 }
 
@@ -114,13 +114,14 @@ my @obst_limit_func;
 my @obst_reg_reqs;
 my @obst_opvar;       # stack for the "ir_op *op_<arch>_<op-name> = NULL;" statements
 my @obst_get_opvar;   # stack for the get_op_<arch>_<op-name>() functions
-my @obst_constructor; # stack for node constructor functions
+my $obst_constructor; # stack for node constructor functions
 my @obst_new_irop;    # stack for the new_ir_op calls
+my @obst_free_irop;   # stack for free_ir_op calls
 my @obst_enum_op;     # stack for creating the <arch>_opcode enum
-my @obst_header;      # stack for function prototypes
+my $obst_header;      # stack for function prototypes
 my @obst_is_archirn;  # stack for the is_$arch_irn() function
 my @obst_cmp_attr;    # stack for the compare attribute functions
-my @obst_proj;        # stack for the pn_ numbers
+my $obst_proj = "";   # stack for the pn_ numbers
 my $orig_op;
 my $arity;
 my $cmp_attr_func;
@@ -158,16 +159,329 @@ foreach my $class_name (keys(%reg_classes)) {
 # for registering additional opcodes
 $n_opcodes += $additional_opcodes if (defined($additional_opcodes));
 
-push(@obst_header, "void ".$arch."_create_opcodes(const arch_irn_ops_t *be_ops);\n");
+$obst_header .= "void ${arch}_create_opcodes(const arch_irn_ops_t *be_ops);\n";
+$obst_header .= "void ${arch}_free_opcodes(void);\n";
 
-push(@obst_enum_op, "typedef enum _$arch\_opcodes {\n");
+sub create_constructor {
+	my $op   = shift;
+	my $name = shift;
+	my $n    = shift;
+	my $on   = shift;
+	my $known_mode;
+
+	my $suffix = "";
+	if ($name ne "") {
+		$suffix = "_${name}";
+	}
+
+	# determine mode
+	if (exists($n->{mode})) {
+		$known_mode = $n->{mode};
+	}
+
+	# determine arity
+	my $arity = 0;
+	if(exists($n->{"arity"})) {
+		$arity = $n->{"arity"};
+	} elsif (exists($n->{"reg_req"}) && exists($n->{"reg_req"}{"in"})) {
+		$arity = scalar(@{ $n->{"reg_req"}{"in"} });
+	} elsif (exists($n->{"ins"})) {
+		$arity = scalar(@{ $n->{"ins"} });
+	}
+	if($arity eq "variable") {
+		$arity = $ARITY_VARIABLE;
+	} elsif($arity eq "dynamic") {
+		$arity = $ARITY_DYNAMIC;
+	}
+
+	# determine out arity
+	my $out_arity = 0;
+	if(exists($n->{"out_arity"})) {
+		$out_arity = $n->{"out_arity"};
+	} elsif (exists($n->{"reg_req"}) && exists($n->{"reg_req"}{"out"})) {
+		$out_arity = scalar(@{ $n->{"reg_req"}{"out"} });
+	} elsif (exists($n->{"outs"})) {
+		$out_arity = scalar(@{ $n->{"outs"} });
+	}
+	if($out_arity eq "variable") {
+		$out_arity = $ARITY_VARIABLE;
+	} elsif($out_arity eq "dynamic") {
+		$out_arity = $ARITY_DYNAMIC;
+	}
+	if ($out_arity != 0 && $out_arity != 1 && !defined($known_mode)) {
+		$known_mode = "mode_T";
+	}
+
+	my $comment = $n->{"comment"};
+	if(!exists($n->{"comment"})) {
+		$comment = "construct ${orig_op} node";
+	}
+	$comment =
+		"/**\n".
+		" * ${comment}\n".
+		" */\n";
+
+	$obst_constructor .= $comment;
+
+	# create constructor head
+	my $complete_args = "";
+	$temp             = "";
+
+	$temp = "ir_node *new_bd_${arch}_${op}${suffix}(dbg_info *dbgi, ir_node *block";
+	if (!exists($n->{"args"})) { # default args
+		if ($arity == $ARITY_VARIABLE) {
+			$complete_args = ", int arity, ir_node *in[]";
+		} elsif ($arity == $ARITY_DYNAMIC) {
+			$complete_args = "";
+		} else {
+			for (my $i = 0; $i < $arity; $i++) {
+				my $opname = "op${i}";
+				if (exists($n->{"ins"})) {
+					my @ins = @{ $n->{"ins"} };
+					$opname = $ins[$i];
+				}
+
+				$complete_args .= ", ir_node *${opname}";
+			}
+		}
+		if ($out_arity == $ARITY_VARIABLE) {
+			$complete_args .= ", int n_res";
+		}
+
+		if (!defined($known_mode)) {
+			$complete_args .= ", ir_mode *mode";
+		}
+	} else { # user defined args
+		for my $href (@{ $n->{"args"} }) {
+			$href->{"type"} .= " " if ($href->{"type"} !~ / [*]?$/); # put a space between name and type if there is none at the end
+			$complete_args  .= ", ".$href->{"type"}.$href->{"name"};
+		}
+	}
+
+	# we have additional attribute arguements
+	if (exists($n->{"attr"})) {
+		$complete_args .= ", ".$n->{"attr"};
+	}
+
+	$temp .= "$complete_args)";
+	$obst_constructor .= "${temp}\n{\n";
+
+	$obst_header .= $comment;
+	$obst_header .= "${temp};\n";
+
+	# emit constructor code
+	$temp = <<EOF;
+	ir_graph         *irg        = get_irn_irg(block);
+	ir_op            *op         = op_${arch}_${op};
+	arch_irn_flags_t  irn_flags_ = arch_irn_flags_none;
+	ir_node          *res;
+	backend_info_t   *info;
+EOF
+
+	if($arity == $ARITY_DYNAMIC) {
+		$temp .= <<EOF;
+	int             arity   = -1;
+	ir_node       **in      = NULL;
+EOF
+	} elsif($arity == $ARITY_VARIABLE) {
+	} else {
+		$temp .= <<EOF;
+	int             arity   = $arity;
+EOF
+		if($arity > 0) {
+			$temp .= <<EOF;
+	ir_node        *in[$arity];
+EOF
+		} else {
+			$temp .= <<EOF;
+	ir_node       **in      = NULL;
+EOF
+		}
+	}
+	if($out_arity == $ARITY_DYNAMIC) {
+		$temp .= <<EOF;
+	int             n_res   = -1;
+EOF
+	} elsif($out_arity == $ARITY_VARIABLE) {
+	} else {
+		$temp .= <<EOF;
+	int             n_res   = ${out_arity};
+EOF
+	}
+
+	if (defined($known_mode)) {
+		$temp .= <<EOF;
+	ir_mode        *mode    = ${known_mode};
+EOF
+	}
+
+	undef my $in_req_var;
+	undef my $out_req_var;
+
+	my $set_out_reqs = "";
+
+	# set up static variables for requirements and registers
+	if (exists($n->{"reg_req"})) {
+		my %req = %{ $n->{"reg_req"} };
+		my $idx;
+
+		undef my @in;
+		@in = @{ $req{"in"} } if (exists($req{"in"}));
+		undef my @out;
+		@out = @{ $req{"out"} } if exists(($req{"out"}));
+
+		for(my $idx = 0; $idx < $#in; $idx++) {
+			my $req = $in[$idx];
+			generate_requirements($req, $n, "${arch}_${op}", $idx, 1);
+		}
+		for(my $idx = 0; $idx < $#out; $idx++) {
+			my $req = $out[$idx];
+			generate_requirements($req, $n, "${arch}_${op}", $idx, 0);
+		}
+
+		if (@in) {
+			if($arity >= 0 && scalar(@in) != $arity) {
+				die "Fatal error: Arity and number of in requirements don't match for ${op}\n";
+			}
+
+			$temp .= "\tstatic const arch_register_req_t *in_reqs[] =\n";
+			$temp .= "\t{\n";
+			for ($idx = 0; $idx <= $#in; $idx++) {
+				my $req = $in[$idx];
+				my $reqstruct = generate_requirements($req, $n, "${arch}_${op}", $idx, 1);
+				$temp .= "\t\t& ${reqstruct},\n";
+			}
+			$temp .= "\t};\n";
+		} else {
+			if($arity > 0) {
+				die "Fatal error: need in requirements for ${op}\n";
+			}
+			$temp .= "\tstatic const arch_register_req_t **in_reqs = NULL;\n";
+		}
+
+		if (@out) {
+			if($out_arity >= 0 && scalar(@out) != $out_arity) {
+				die "Fatal error: Out-Arity and number of out requirements don't match for ${op}\n";
+			}
+
+			for ($idx = 0; $idx <= $#out; $idx++) {
+				my $req = $out[$idx];
+				my $reqstruct = generate_requirements($req, $n, "${arch}_${op}", $idx, 0);
+				$set_out_reqs .= <<EOF;
+	info->out_infos[${idx}].req = &${reqstruct};
+EOF
+			}
+		} else {
+			if($out_arity > 0) {
+				die "Fatal error: need out requirements for ${op}\n";
+			}
+		}
+	} else {
+		$temp .= "\tstatic const arch_register_req_t **in_reqs = NULL;\n";
+	}
+	my $attr_type = $on->{attr_type};
+	if(exists($n->{"init_attr"})) {
+		$temp .= "\t${attr_type} *attr;\n";
+	}
+
+	$temp .= "\n";
+
+	if($arity > 0) {
+		$temp .= "\t/* construct in array */\n";
+		for (my $i = 0; $i < $arity; $i++) {
+			my $opname = "op${i}";
+			if (exists($n->{"ins"})) {
+				my @ins = @{ $n->{"ins"} };
+				$opname = $ins[$i];
+			}
+
+			$temp .= "\tin[${i}] = ${opname};\n";
+		}
+		$temp .= "\n";
+	}
+
+	# set flags
+	if (exists($n->{"irn_flags"})) {
+		$temp .= "\t/* flags */\n";
+		my %known_irn_flags = (
+			"none"             => "arch_irn_flags_none",
+			"dont_spill"       => "arch_irn_flags_dont_spill",
+			"rematerializable" => "arch_irn_flags_rematerializable",
+			"modify_flags"     => "arch_irn_flags_modify_flags",
+			"simple_jump"      => "arch_irn_flags_simple_jump",
+			"not_scheduled"    => "arch_irn_flags_not_scheduled",
+		);
+		if (%custom_irn_flags) {
+			%known_irn_flags = (%known_irn_flags, %custom_irn_flags);
+		}
+		foreach my $flag (@{$n->{"irn_flags"}}) {
+			if (not defined($known_irn_flags{$flag})) {
+				print STDERR "WARNING: irn_flag '$flag' in opcode $op is unknown\n";
+			} else {
+				$temp .= "\tirn_flags_ |= " . $known_irn_flags{$flag} . ";\n";
+			}
+		}
+		$temp .= "\n";
+	}
+
+	# lookup init function
+	my $attr_init_code = "(void)in;(void)irn_flags_;(void)in_reqs;(void)n_res;";
+	if ($attr_type ne "") {
+		$attr_init_code = $init_attr{$attr_type};
+		if(!defined($attr_init_code)) {
+			die "Fatal error: Couldn't find attribute initialisation code for type '${attr_type}'";
+		}
+	}
+	my $custominit = "";
+	if(defined($custom_init_attr_func)) {
+		$custominit .= &$custom_init_attr_func($n, $on, "${arch}_${op}");
+	}
+	if(defined($n->{custominit})) {
+		$custominit .= $n->{custominit};
+	}
+
+	$temp .= <<EOF;
+	/* create node */
+	assert(op != NULL);
+	res = new_ir_node(dbgi, irg, block, op, mode, arity, in);
+
+	/* init node attributes */
+	${attr_init_code}
+	${custominit}
+	info = be_get_info(res);
+	(void) info; /* avoid potential warning */
+${set_out_reqs}
+
+EOF
+
+	if (exists($n->{"init_attr"})) {
+		$temp .= "\tattr = (${attr_type}*)get_irn_generic_attr(res);\n";
+		$temp .= "\t(void) attr; /* avoid potential warning */\n";
+		$temp .= "\t".$n->{"init_attr"}."\n";
+	}
+
+	$temp .= <<EOF;
+	/* optimize node */
+	res = optimize_node(res);
+	irn_verify_irg(res, irg);
+
+	return res;
+EOF
+
+	$obst_constructor .= $temp;
+
+	# close constructor function
+	$obst_constructor .= "}\n\n";
+}
+
+push(@obst_enum_op, "typedef enum ${arch}_opcodes {\n");
 foreach my $op (keys(%nodes)) {
 	my %n        = %{ $nodes{"$op"} };
 	my $known_mode;
 	my $num_outs = 0;
 	my $out_arity;
 	my @out_flags;
-	
+
 	# determine arity
 	$arity = 0;
 	if(exists($n{"arity"})) {
@@ -205,28 +519,32 @@ foreach my $op (keys(%nodes)) {
 	# define proj numbers and in numbers
 	if (exists($n{"outs"})) {
 		undef my @outs;
-		
-		@outs     = @{ $n{"outs"} };
+
+		@outs = @{ $n{"outs"} };
 		if($out_arity >= 0 && scalar(@outs) != $out_arity) {
 			die "Fatal error: Op ${op} has different number of outs and out_arity\n";
 		}
 
 		$num_outs = $#outs + 1;
-		
-		push(@obst_proj, "\nenum pn_$op {\n");
-		
-		for (my $idx = 0; $idx <= $#outs; $idx++) {
-			# check, if we have additional flags annotated to out
-			if ($outs[$idx] =~ /:((S|I)(\|(S|I))*)/) {
-				push(@out_flags, $1);
-				$outs[$idx] =~ s/:((S|I)(\|(S|I))*)//;
+
+		if ($num_outs > 0) {
+			$obst_proj .= "\nenum pn_${op} {\n";
+
+			for (my $idx = 0; $idx <= $#outs; $idx++) {
+				# check, if we have additional flags annotated to out
+				if ($outs[$idx] =~ /:((S|I)(\|(S|I))*)/) {
+					push(@out_flags, $1);
+					$outs[$idx] =~ s/:((S|I)(\|(S|I))*)//;
+				}
+				$obst_proj .= "\tpn_${op}_".$outs[$idx]." = ${idx},\n";
 			}
-			push(@obst_proj, "\tpn_$op\_".$outs[$idx]." = $idx,\n");
+
+			$obst_proj .= "};\n";
 		}
-		
-		push(@obst_proj, "};\n");
 		# outs have names, it must be a mode_T node
-		$known_mode = "mode_T";
+		if (!defined($n{mode})) {
+			$n{mode} = "mode_T";
+		}
 	}
 	if (exists($n{"ins"})) {
 		undef my @ins;
@@ -236,33 +554,32 @@ foreach my $op (keys(%nodes)) {
 			die "Fatal error: Op ${op} has different number of ins and arity\n";
 		}
 
-		push(@obst_proj, "\nenum n_$op {\n");
-
-		for (my $idx = 0; $idx <= $#ins; $idx++) {
-			push(@obst_proj, "\tn_${op}_".$ins[$idx]." = $idx,\n");
+		if ($#ins >= 0) {
+			$obst_proj .= "\nenum n_$op {\n";
+			for (my $idx = 0; $idx <= $#ins; $idx++) {
+				$obst_proj .= "\tn_${op}_".$ins[$idx]." = ${idx},\n";
+			}
+			$obst_proj .= "};\n";
 		}
-
-		push(@obst_proj, "};\n");
-	}
-
-	# determine mode
-	if (exists($n{"mode"})) {
-		$known_mode = $n{"mode"};
 	}
 
+	# Create opcode
 	push(@obst_opvar, "ir_op *op_$op = NULL;\n");
 	push(@obst_get_opvar, "ir_op *get_op_$op(void)         { return op_$op; }\n");
 	push(@obst_get_opvar, "int    is_$op(const ir_node *n) { return get_$arch\_irn_opcode(n) == iro_$op; }\n\n");
 
 	push(@obst_is_archirn, "is_$op(node)");
 
-	push(@obst_header, "extern ir_op *op_$op;\n");
-	push(@obst_header, "ir_op *get_op_$op(void);\n");
-	push(@obst_header, "int is_$op(const ir_node *n);\n");
+	$obst_header .= <<EOF;
+extern ir_op *op_${op};
+ir_op *get_op_${op}(void);
+int is_${op}(const ir_node *n);
+EOF
 
-	my $attr_type= $n{"attr_type"};
+	my $attr_type= $n{attr_type};
 	if(!defined($attr_type)) {
 		$attr_type = $default_attr_type;
+		$n{attr_type} = $attr_type;
 	}
 
 	# determine hash function
@@ -276,14 +593,14 @@ foreach my $op (keys(%nodes)) {
 	if (exists($n{"cmp_attr"})) {
 		my $cmpcode = $n{"cmp_attr"};
 
-		push(@obst_cmp_attr, "static int cmp_attr_$op(ir_node *a, ir_node *b) {\n");
+		push(@obst_cmp_attr, "static int cmp_attr_$op(const ir_node *a, const ir_node *b) {\n");
 		if($cmpcode =~ m/attr_a/) {
-			push(@obst_cmp_attr, "\t${attr_type} *attr_a = get_irn_generic_attr(a);\n");
+			push(@obst_cmp_attr, "\tconst ${attr_type} *attr_a = get_irn_generic_attr_const(a);\n");
 		} else {
 			push(@obst_cmp_attr, "\t(void) a;\n");
 		}
 		if($cmpcode =~ m/attr_b/) {
-			push(@obst_cmp_attr, "\t${attr_type} *attr_b = get_irn_generic_attr(b);\n");
+			push(@obst_cmp_attr, "\tconst ${attr_type} *attr_b = get_irn_generic_attr_const(b);\n");
 		} else {
 			push(@obst_cmp_attr, "\t(void) b;\n");
 		}
@@ -291,6 +608,8 @@ foreach my $op (keys(%nodes)) {
 		push(@obst_cmp_attr, "}\n\n");
 
 		$cmp_attr_func = "cmp_attr_${op}";
+	} elsif ($attr_type eq "") {
+		$cmp_attr_func = "NULL";
 	} else {
 		if(defined($compare_attr{${attr_type}})) {
 			$cmp_attr_func = $compare_attr{${attr_type}};
@@ -299,283 +618,100 @@ foreach my $op (keys(%nodes)) {
 		}
 	}
 
-	if (exists($n{"rd_constructor"}) && $n{"rd_constructor"} =~ /^NONE$/i) {
-		# we explicitly skip the constructor if the specification entry says NONE
+	my %constructors;
+	if (exists($n{constructors})) {
+		%constructors = %{ $n{constructors} };
 	} else {
-		my $comment = $n{"comment"};
-		if(!exists($n{"comment"})) {
-			$comment = "construct ${orig_op} node";
-		}
-		$comment = 
-			"/**\n".
-			" * ${comment}\n".
-			" */\n";
-
-		push(@obst_constructor, $comment);
-		
-		# create constructor head
-		my $complete_args = "";
-		$temp             = "";
-	
-		$temp = "ir_node *new_bd_$op(dbg_info *db, ir_node *block";
-		if (!exists($n{"args"})) { # default args
-			if ($arity == $ARITY_VARIABLE) {
-				$complete_args = ", int arity, ir_node *in[]";
-			} elsif ($arity == $ARITY_DYNAMIC) {
-				$complete_args = "";
-			} else {
-				for (my $i = 0; $i < $arity; $i++) {
-					my $opname = "op${i}";
-					if (exists($n{"ins"})) {
-						my @ins = @{ $n{"ins"} };
-						$opname = $ins[$i];
-					}
-
-					$complete_args .= ", ir_node *${opname}";
-				}
-			}
-			if ($out_arity == $ARITY_VARIABLE) {
-				$complete_args .= ", int n_res";
-			}
-
-			if (!defined($known_mode)) {
-				$complete_args .= ", ir_mode *mode";
-			}
-		} else { # user defined args
-			for my $href (@{ $n{"args"} }) {
-				$href->{"type"} .= " " if ($href->{"type"} !~ / [*]?$/); # put a space between name and type if there is none at the end
-				$complete_args  .= ", ".$href->{"type"}.$href->{"name"};
+		# Create 1 default constructor
+		my %constructor = ();
+		foreach my $a ("comment", "ins", "outs", "args", "attr", "units",
+				"reg_req", "init_attr", "irn_flags", "mode", "arity",
+				"out_arity", "custominit") {
+			if (defined($n{$a})) {
+				$constructor{$a} = $n{$a};
 			}
 		}
+		%constructors = ( "" => \%constructor );
+	}
 
-		# we have additional attribute arguements
-		if (exists($n{"attr"})) {
-			$complete_args .= ", ".$n{"attr"};
+	foreach my $constr (keys(%constructors)) {
+		my %cstr = %{ $constructors{$constr} };
+		# Copy some values from outer node if they don't exists in the constr
+		foreach my $a ("ins", "outs", "irn_flags", "mode", "args", "attr",
+		               "custominit") {
+			if (!defined($cstr{$a}) && defined($n{$a})) {
+				$cstr{$a} = $n{$a};
+			}
 		}
-		
-		$temp .= "$complete_args)";
-		push(@obst_constructor, $temp."\n{\n");
-		push(@obst_header, $comment);
-		push(@obst_header, $temp.";\n");
-
-		# emit constructor code
-		if (!exists($n{"rd_constructor"})) { # default constructor
-			$temp  = "\tir_node  *res;\n";
-			$temp .= "\tir_op    *op      = op_${op};\n";
-			$temp .= "\tint       flags   = 0;\n";
-
-			if($arity == $ARITY_DYNAMIC) {
-				$temp .= "\tint        arity   = -1;\n";
-				$temp .= "\tir_node  **in      = NULL;\n";
-			} elsif($arity == $ARITY_VARIABLE) {
-			} else {
-				$temp .= "\tint       arity   = $arity;\n";
-				if($arity > 0) {
-					$temp .= "\tir_node  *in[$arity];\n";
-				} else {
-					$temp .= "\tir_node **in    = NULL;\n";
-				}
-			}
-			if($out_arity == $ARITY_DYNAMIC) {
-				$temp .= "\tint       n_res   = -1;\n";
-			} elsif($out_arity == $ARITY_VARIABLE) {
-			} else {
-				$temp .= "\tint       n_res   = ${out_arity};\n";
-			}
-
-			if (defined($known_mode)) {
-				$temp .= "\tir_mode  *mode    = ${known_mode};\n";
-			}
-
-			# set up static variables for cpu execution unit assigments
-			if (exists($n{"units"})) {
-				$temp .= gen_execunit_list_initializer($n{"units"});
-			} else {
-				$temp .= "\tstatic const be_execution_unit_t ***exec_units = NULL;\n";
-			}
-
-			undef my $in_req_var;
-			undef my $out_req_var;
-			
-			# set up static variables for requirements and registers
-			if (exists($n{"reg_req"})) {
-				my %req = %{ $n{"reg_req"} };
-				my $idx;
-
-				undef my @in;
-				@in = @{ $req{"in"} } if (exists($req{"in"}));
-				undef my @out;
-				@out = @{ $req{"out"} } if exists(($req{"out"}));
-
-				for(my $idx = 0; $idx < $#in; $idx++) {
-					my $req = $in[$idx];
-					generate_requirements($req, \%n, $op, $idx, 1);
-				}
-				for(my $idx = 0; $idx < $#out; $idx++) {
-					my $req = $out[$idx];
-					generate_requirements($req, \%n, $op, $idx, 0);
-				}
-
-				if (@in) {
-					if($arity >= 0 && scalar(@in) != $arity) {
-						die "Fatal error: Arity and number of in requirements don't match for ${op}\n";
-					}
-
-					$temp .= "\tstatic const arch_register_req_t *in_reqs[] =\n";
-					$temp .= "\t{\n";
-					for ($idx = 0; $idx <= $#in; $idx++) {
-						my $req = $in[$idx];
-						my $reqstruct = generate_requirements($req, \%n, $op, $idx, 1);
-						$temp .= "\t\t& ${reqstruct},\n";
-					}
-					$temp .= "\t};\n";
-				} else {
-					if($arity > 0) {
-						die "Fatal error: need in requirements for ${op}\n";
-					}
-					$temp .= "\tstatic const arch_register_req_t **in_reqs = NULL;\n";
-				}
-
-				if (@out) {
-					if($out_arity >= 0 && scalar(@out) != $out_arity) {
-						die "Fatal error: Out-Arity and number of out requirements don't match for ${op}\n";
-					}
-
-					$temp .= "\tstatic const arch_register_req_t *out_reqs[] =\n";
-					$temp .= "\t{\n";
-					for ($idx = 0; $idx <= $#out; $idx++) {
-						my $req = $out[$idx];
-						my $reqstruct = generate_requirements($req, \%n, $op, $idx, 0);
-						$temp .= "\t\t& ${reqstruct},\n";
-					}
-					$temp .= "\t};\n";
-				} else {
-					if($out_arity > 0) {
-						die "Fatal error: need out requirements for ${op}\n";
-					}
-					$temp .= "\tstatic const arch_register_req_t **out_reqs = NULL;\n";
-				}
-			} else {
-				$temp .= "\tstatic const arch_register_req_t **in_reqs = NULL;\n";
-				$temp .= "\tstatic const arch_register_req_t **out_reqs = NULL;\n";
-			}
-			if(exists($n{"init_attr"})) {
-				$temp .= "\t${attr_type} *attr;\n";
-			}
-			
-			$temp .= "\n";
-
-			if($arity > 0) {
-				$temp .= "\t/* construct in array */\n";
-				for (my $i = 0; $i < $arity; $i++) {
-					my $opname = "op${i}";
-					if (exists($n{"ins"})) {
-						my @ins = @{ $n{"ins"} };
-						$opname = $ins[$i];
-					}
-
-					$temp .= "\tin[${i}] = ${opname};\n";
-				}
-				$temp .= "\n";
-			}
-
-			# set flags
-			if (exists($n{"irn_flags"})) {
-				$temp .= "\t/* flags */\n";
-				foreach my $flag (split(/\|/, $n{"irn_flags"})) {
-					if ($flag eq "R") {
-						$temp .= "\tflags |= arch_irn_flags_rematerializable;\n";
-					} elsif ($flag eq "N") {
-						$temp .= "\tflags |= arch_irn_flags_dont_spill;\n";
-					} else {
-						die "Fatal error: unknown flag $flag for ${op}\n";
-					}
-				}
-				$temp .= "\n";
-			}
-			
-			$temp .= "\t/* create node */\n";
-			$temp .= "\tassert(op != NULL);\n";
-			$temp .= "\tres = new_ir_node(db, current_ir_graph, block, op, mode, arity, in);\n";
-			$temp .= "\n";
-			
-			$temp .= "\t/* init node attributes */\n";
-			# lookup init function
-			my $attr_init_code = $init_attr{$attr_type};
-			if(!defined($attr_init_code)) {
-				die "Fatal error: Couldn't find attribute initialisation code for type '${attr_type}'";
-			}
-			$temp .= "${attr_init_code}\n";
-			if(defined($custom_init_attr_func)) {
-				$temp .= &$custom_init_attr_func(\%n, $op);
-			}
-			$temp .= "\n";
-			
-			if (exists($n{"init_attr"})) {
-				$temp .= "\tattr = get_irn_generic_attr(res);\n";
-				$temp .= "\t".$n{"init_attr"}."\n";
-			}
-			
-			$temp .= "\t/* optimize node */\n";
-			$temp .= "\tres = optimize_node(res);\n";
-			$temp .= "\tirn_vrfy_irg(res, current_ir_graph);\n";
-			$temp .= "\n";
-			
-			$temp .= "\treturn res;\n";
-			
-			push(@obst_constructor, $temp);
-		}
-		else { # user defined constructor
-			push(@obst_constructor, $n{"rd_constructor"});
-		}
-
-		# close constructor function
-		push(@obst_constructor, "}\n\n");
-	} # constructor creation
+		create_constructor($orig_op, $constr, \%cstr, \%n);
+	}
 
 	# set default values for state and flags if not given
-	$n{"state"}    = "floats" if (! exists($n{"state"}));
-	$n{"op_flags"} = "N"      if (! exists($n{"op_flags"}));
- 
-    
-	push(@obst_new_irop, "\n\tmemset(&ops, 0, sizeof(ops));\n");
-	push(@obst_new_irop, "\tops.be_ops        = be_ops;\n");
-	push(@obst_new_irop, "\tops.dump_node     = $arch\_dump_node;\n");
+	$n{"state"}     = "floats" if (! exists($n{"state"}));
+	$n{"op_flags"}  = ["none"] if (! exists($n{"op_flags"}));
+	$n{"dump_func"} = "${arch}_dump_node" if (!exists($n{"dump_func"}));
+	my $dump_func = $n{"dump_func"};
 
-	if (defined($cmp_attr_func)) {
-		push(@obst_new_irop, "\tops.node_cmp_attr = ${cmp_attr_func};\n");
+	my %known_flags = map { $_ => 1 } (
+		"none", "commutative", "cfopcode", "unknown_jump", "fragile",
+		"forking", "highlevel", "constlike", "keep", "start_block",
+		"uses_memory", "dump_noblock", "cse_neutral"
+	);
+	my $is_fragile = 0;
+	foreach my $flag (@{$n{"op_flags"}}) {
+		if (not defined($known_flags{$flag})) {
+			print STDERR "WARNING: Flag '$flag' in opcode $op is unknown\n";
+		}
+		if ($flag eq "fragile") {
+			$is_fragile = 1;
+		}
 	}
-	my $copy_attr_func = $copy_attr{$attr_type};
-	if (!defined($copy_attr_func)) {
-		$copy_attr_func = $default_copy_attr;
-	}
-	if (defined($copy_attr_func)) {
-		push(@obst_new_irop, "\tops.copy_attr = ${copy_attr_func};\n");
-	}
-	if (defined($hash_func)) {
-		push(@obst_new_irop, "\tops.hash = ${hash_func};\n");
+	my @mapped = map { "irop_flag_$_" } @{$n{"op_flags"}};
+	my $op_flags = join('|', @mapped);
+
+	my $attr_size = "0";
+	if ($attr_type ne "") {
+		$attr_size = "sizeof(${attr_type})"
 	}
 
 	$n_opcodes++;
-	my $n_res = $out_arity;
-	if($n_res < 0) {
-		$n_res = "20"; # hacky....
-	}
-	$temp  = "\top_$op = new_ir_op(cur_opcode + iro_$op, \"$op\", op_pin_state_".$n{"state"}.", ".$n{"op_flags"};
-	$temp .= "|M, ".translate_arity($arity).", 0, sizeof(${attr_type}), &ops);\n";
+	$temp  = "\top = new_ir_op(cur_opcode + iro_$op, \"$op\", op_pin_state_".$n{"state"}.", $op_flags";
+	$temp .= ", ".translate_arity($arity).", 0, ${attr_size});\n";
 	push(@obst_new_irop, $temp);
-	push(@obst_new_irop, "\tset_op_tag(op_$op, $arch\_op_tag);\n");
-	if(defined($default_op_attr_type)) {
-		push(@obst_new_irop, "\tattr = &attrs[iro_$op];\n");
-		if(defined($n{op_attr_init})) {
-			push(@obst_new_irop, "\t".$n{op_attr_init}."\n");
-		}
-		push(@obst_new_irop, "\tset_op_attr(op_$op, attr);\n");
+	push(@obst_new_irop, "\top->ops.be_ops        = be_ops;\n");
+	push(@obst_new_irop, "\top->ops.dump_node     = ${dump_func};\n");
+	if (defined($cmp_attr_func)) {
+		push(@obst_new_irop, "\top->ops.node_cmp_attr = ${cmp_attr_func};\n");
 	}
+	my $copy_attr_func = $copy_attr{$attr_type};
+	if (!defined($copy_attr_func)) {
+		# don't set a copy_attr function if the node has no additional attributes.
+		if ($attr_type ne "") {
+			$copy_attr_func = $default_copy_attr;
+		}
+	}
+	if (defined($copy_attr_func)) {
+		push(@obst_new_irop, "\top->ops.copy_attr     = ${copy_attr_func};\n");
+	}
+	if (defined($hash_func)) {
+		push(@obst_new_irop, "\top->ops.hash          = ${hash_func};\n");
+	}
+
+	if ($is_fragile) {
+		push(@obst_new_irop, "\tir_op_set_memory_index(op, n_${op}_mem);\n");
+		push(@obst_new_irop, "\tir_op_set_fragile_indices(op, pn_${op}_X_regular, pn_${op}_X_except);\n");
+	}
+	push(@obst_new_irop, "\tset_op_tag(op, $arch\_op_tag);\n");
+	if(defined($n{op_attr_init})) {
+		push(@obst_new_irop, "\t".$n{op_attr_init}."\n");
+	}
+	push(@obst_new_irop, "\top_${op} = op;\n");
+
+	push(@obst_free_irop, "\tfree_ir_op(op_$op); op_$op = NULL;\n");
 
 	push(@obst_enum_op, "\tiro_$op,\n");
 
-	push(@obst_header, "\n");
+	$obst_header .= "\n";
 }
 push(@obst_enum_op, "\tiro_$arch\_last_generated,\n");
 push(@obst_enum_op, "\tiro_$arch\_last = iro_$arch\_last_generated");
@@ -586,7 +722,10 @@ push(@obst_enum_op, "\n} $arch\_opcodes;\n\n");
 
 open(OUT, ">$target_c") || die("Fatal error: Could not open $target_c, reason: $!\n");
 
-print OUT "#include \"gen_$arch\_regalloc_if.h\"\n\n";
+print OUT "#include \"gen_$arch\_regalloc_if.h\"\n";
+print OUT "#include \"irverify_t.h\"\n";
+print OUT "#include \"fourcc.h\"\n";
+print OUT "\n";
 print OUT @obst_cmp_attr;
 print OUT "\n";
 print OUT @obst_opvar;
@@ -658,7 +797,7 @@ print OUT <<END;
 #ifdef BIT
 #undef BIT
 #endif
-#define BIT(x)  (1 << (x % 32))
+#define BIT(x)  (1 << (x))
 
 END
 
@@ -668,70 +807,39 @@ print OUT "\n";
 print OUT @obst_reg_reqs;
 print OUT "\n";
 
-print OUT @obst_constructor;
-
 print OUT<<ENDOFMAIN;
+$obst_constructor
+
 /**
  * Creates the $arch specific Firm machine operations
  * needed for the assembler irgs.
  */
-void $arch\_create_opcodes(const arch_irn_ops_t *be_ops) {
-#define N   irop_flag_none
-#define L   irop_flag_labeled
-#define C   irop_flag_commutative
-#define X   irop_flag_cfopcode
-#define I   irop_flag_ip_cfopcode
-#define F   irop_flag_fragile
-#define Y   irop_flag_forking
-#define H   irop_flag_highlevel
-#define c   irop_flag_constlike
-#define K   irop_flag_keep
-#define M   irop_flag_machine
-#define O   irop_flag_machine_op
-#define NB  irop_flag_dump_noblock
-#define NI  irop_flag_dump_noinput
-#define R   (irop_flag_user << 0)
+void $arch\_create_opcodes(const arch_irn_ops_t *be_ops)
+{
+	ir_op *op;
+	int    cur_opcode = get_next_ir_opcodes(iro_$arch\_last);
 
-	ir_op_ops  ops;
-	int        cur_opcode;
-	static int run_once = 0;
-	int        i;
-ENDOFMAIN
-
-	if (defined($default_op_attr_type)) {
-		print OUT "\t$default_op_attr_type *attr, *attrs;\n";
-	}
-
-print OUT<<ENDOFMAIN;
-
-	if (run_once)
-		return;
-	run_once = 1;
-	
-	/* we handle all middleend nodes as well that have no other handler */
-	for (i = 0; i <= iro_Last; ++i) {
-		ir_op *op      = get_irp_opcode(i);
-		if (op->ops.be_ops == NULL)
-			op->ops.be_ops = be_ops;
-	}
-
-	cur_opcode = get_next_ir_opcodes(iro_$arch\_last);
-  
 	$arch\_opcode_start = cur_opcode;
 ENDOFMAIN
 
-	if (defined($default_op_attr_type)) {
-		print OUT "\tattrs = xmalloc(sizeof(attr[0]) * iro_$arch\_last);\n";
-		print OUT "\tmemset(attrs, 0, sizeof(attr[0]) * iro_$arch\_last);\n";
-	}
-	
 print OUT @obst_new_irop;
 print OUT "\n";
 print OUT "\t$arch\_register_additional_opcodes(cur_opcode);\n" if (defined($additional_opcodes));
 print OUT "\t$arch\_opcode_end = cur_opcode + iro_$arch\_last";
 print OUT " + $additional_opcodes" if (defined($additional_opcodes));
 print OUT ";\n";
-print OUT "}\n";
+print OUT <<ENDOFMAIN;
+}
+
+void $arch\_free_opcodes(void)
+{
+ENDOFMAIN
+
+print OUT @obst_free_irop;
+
+print OUT <<ENDOFMAIN;
+}
+ENDOFMAIN
 
 close(OUT);
 
@@ -755,14 +863,16 @@ print OUT<<EOF;
 EOF
 
 print OUT @obst_enum_op;
-print OUT "int is_$arch\_irn(const ir_node *node);\n\n";
-print OUT "int get_$arch\_opcode_first(void);\n";
-print OUT "int get_$arch\_opcode_last(void);\n";
-print OUT "int get_$arch\_irn_opcode(const ir_node *node);\n";
-print OUT @obst_header;
-print OUT @obst_proj;
-
 print OUT <<EOF;
+int is_${arch}_irn(const ir_node *node);
+int is_${arch}_op(const ir_op *op);
+
+int get_${arch}_opcode_first(void);
+int get_${arch}_opcode_last(void);
+int get_${arch}_irn_opcode(const ir_node *node);
+${obst_header}
+${obst_proj}
+
 #endif
 EOF
 
@@ -773,7 +883,7 @@ close(OUT);
 ###
 sub translate_arity {
 	my $arity = shift;
-	
+
 	if ($arity =~ /^\d+$/) {
 		if    ($arity == 0) {
 			return "oparity_zero";
@@ -799,74 +909,6 @@ sub translate_arity {
 	}
 }
 
-###
-# Return the list of pointers for the given execution units.
-###
-sub gen_execunit_list_initializer {
-	my $units   = shift;
-	my $uc_arch = uc($arch);
-	my $ret     = "";
-	my $ret2    = "";
-	my %init;
-
-	foreach my $unit (@{ $units }) {
-		if ($unit eq "DUMMY") {
-			push(@{ $init{"DUMMY"} }, "\t\t&be_machine_execution_units_DUMMY[0]");
-		}
-		elsif (exists($cpu{"$unit"})) {
-			# operation can be executed on all units of this type
-			# -> add them all
-			my $tp_name = "$arch\_execution_units_$unit";
-			my $idx     = 0;
-			foreach (@{ $cpu{"$unit"} }) {
-				next if ($idx++ == 0);  # skip first element (it's not a unit)
-				my $unit_name = "$uc_arch\_EXECUNIT_TP_$unit\_$_";
-				push(@{ $init{"$unit"} }, "\t\t&".$tp_name."[".$unit_name."]");
-			}
-		}
-		else {
-			# operation can be executed only a certain unit
-			# -> find corresponding unit type
-			my $found = 0;
-TP_SEARCH:	foreach my $cur_type (keys(%cpu)) {
-				foreach my $cur_unit (@{ $cpu{"$cur_type"} }) {
-					if ($unit eq $cur_unit) {
-						my $tp_name   = "$arch\_execution_units_$cur_type";
-						my $unit_name = "$uc_arch\_EXECUNIT_TP_$cur_type\_$unit";
-						push(@{ $init{"$unit"} }, "\t\t&".$tp_name."[".$unit_name."]");
-						$found = 1;
-						last TP_SEARCH;
-					}
-				}
-			}
-
-			if (! $found) {
-				print STDERR "Invalid execution unit $unit specified!\n";
-			}
-		}
-	}
-	
-	# prepare the 2-dim array init
-	foreach my $key (keys(%init)) {
-		$ret .= "\tstatic const be_execution_unit_t *allowed_units_".$key."[] =\n";
-		$ret .= "\t{\n";
-		foreach (@{ $init{"$key"} }) {
-			$ret .= "$_,\n";
-		}
-		$ret .= "\t\tNULL\n";
-		$ret .= "\t};\n";
-		$ret2 .= "\t\tallowed_units_$key,\n";
-	}
-	$ret2 .= "\t\tNULL\n";
-
-	$ret .= "\tstatic const be_execution_unit_t **exec_units[] =\n";
-	$ret .= "\t{\n";
-	$ret .= $ret2;
-	$ret .= "\t};\n";
-
-	return $ret;
-}
-
 sub mangle_requirements {
 	my $reqs  = shift;
 	my $class = shift;
@@ -910,7 +952,7 @@ sub get_reg_class {
 }
 
 ###
-# Returns the index of a given register within it's register class.
+# Returns the index of a given register within its register class.
 # @return index or undef
 ###
 sub get_reg_index {
@@ -931,7 +973,7 @@ sub build_inout_idx_class {
 	my @idx_class;
 
 	my $inout = ($is_in ? "in" : "out");
-	
+
 	if (exists($n->{"reg_req"}{"$inout"})) {
 		my @reqs = @{ $n->{"reg_req"}{"$inout"} };
 
@@ -958,7 +1000,7 @@ GET_CLASS:		foreach my $reg (@regs) {
 					} # if (reg =~ ...
 				} # foreach
 			} # if
-			
+
 			push(@idx_class, $class);
 		} # for
 	} # if
@@ -995,7 +1037,7 @@ sub build_subset_class_func {
 	my $flags = shift;
 
 	my @idx_class = build_inout_idx_class($node, $op, !$is_in);
-	
+
 	# set/unset registers
 CHECK_REQS: foreach (@regs) {
 		if (!$is_in && /(!)?in_r(\d+)/) {
@@ -1017,7 +1059,7 @@ CHECK_REQS: foreach (@regs) {
 				}
 				return (undef, undef, undef, undef);
 			}
-			
+
 			if ($1) {
 				$different_pos |= $bit_pos;
 			} else {
@@ -1036,7 +1078,7 @@ CHECK_REQS: foreach (@regs) {
 				print STDERR "Mixed positive and negative constraints for the same slot are not allowed.\n";
 				return (undef, undef, undef, undef);
 			}
-	
+
 			if (!defined($neg)) {
 				$has_limit = 1;
 			}
@@ -1061,7 +1103,7 @@ CHECK_REQS: foreach (@regs) {
 			print STDERR "Unknown register '$_'!\n";
 			return (undef, undef, undef, undef);
 		}
-		
+
 		# set class
 		if (!defined($class)) {
 			$class = $temp;
@@ -1070,15 +1112,15 @@ CHECK_REQS: foreach (@regs) {
 			print STDERR "Registerclass mismatch. '$_' is not member of class '$class'.\n";
 			return (undef, undef, undef, undef);
 		}
-	
+
 		# calculate position inside the initializer bitfield (only 32 bits per
 		# element)
 		my $regidx = get_reg_index($_);
 		my $arrayp = $regidx / 32;
-		push(@{$limit_array[$arrayp]}, $_); 
+		push(@{$limit_array[$arrayp]}, $_);
 		$limit_reqs .= "$_ ";
 	}
-	
+
 	# don't allow ignore regs in negative constraints
 	if($neg) {
 		my @cur_class = @{ $reg_classes{"$class"} };
@@ -1106,7 +1148,7 @@ CHECK_REQS: foreach (@regs) {
 		push(@obst_limit_func, "static const unsigned " . $limit_name . "[] = { ");
 		my $first = 1;
 		my $limitbitsetlen = $regclass2len{$class};
-		my $limitarraylen = $limitbitsetlen / 32 + ($limitbitsetlen % 32 > 0 ? 1 : 0);
+		my $limitarraylen = ($limitbitsetlen+31) / 32;
 		for(my $i = 0; $i < $limitarraylen; $i++) {
 
 			my $limitarraypart = $limit_array[$i];
@@ -1125,7 +1167,10 @@ CHECK_REQS: foreach (@regs) {
 				} elsif(defined($temp)) {
 					$temp .= " | ";
 				}
-				$temp .= "BIT(REG_".uc(${reg}).")";
+				my $firstreg = uc($reg_classes{$class}[0]->{"name"});
+				my $classuc = uc($class);
+				my $reguc = uc($reg);
+				$temp .= "BIT(REG_${classuc}_${reguc})";
 			}
 			if(defined($temp)) {
 				push(@obst_limit_func, "${temp}");
@@ -1149,6 +1194,7 @@ sub generate_requirements {
 	my $idx   = shift;
 	my $is_in = shift;
 	my $class = "";
+	my $width = 1;
 	my $result;
 
 	my @req_type_mask;
@@ -1158,19 +1204,24 @@ sub generate_requirements {
 				push(@req_type_mask, "arch_register_req_type_ignore");
 			} elsif ($f eq "S") {
 				push(@req_type_mask, "arch_register_req_type_produces_sp");
+			} elsif ($f eq "a") {
+				push(@req_type_mask, "arch_register_req_type_aligned");
+			} elsif ($f eq "2" or $f eq "4" or $f eq "8") {
+				$width = int($f);
 			}
 		}
 	}
 
 	if ($reqs eq "none") {
-		
+
 		$result = <<EOF;
 {
 	arch_register_req_type_none,
 	NULL,                         /* regclass */
 	NULL,                         /* limit bitset */
 	0,                            /* same pos */
-	0                             /* different pos */
+	0,                            /* different pos */
+	0                             /* width */
 };
 
 EOF
@@ -1184,7 +1235,8 @@ EOF
 	& ${arch}_reg_classes[CLASS_${arch}_${class}],
 	NULL,        /* limit bitset */
 	0,           /* same pos */
-	0            /* different pos */
+	0,           /* different pos */
+	$width            /* width */
 };
 
 EOF
@@ -1192,11 +1244,11 @@ EOF
 	} else {
 		my ($regclass, $limit_bitset, $same_pos, $different_pos)
 			= build_subset_class_func($node, $op, $idx, $is_in, $reqs, $flags);
-			
+
 		if (!defined($regclass)) {
 			die("Fatal error: Could not build subset for requirements '$reqs' of '$op' pos $idx ... exiting.\n");
 		}
-			
+
 		if (defined($limit_bitset) && $limit_bitset ne "NULL") {
 			push(@req_type_mask, "arch_register_req_type_limited");
 		}
@@ -1219,7 +1271,8 @@ EOF
 	& ${arch}_reg_classes[CLASS_${arch}_${class}],
 	${limit_bitset},
 	${same_pos},        /* same pos */
-	${different_pos}        /* different pos */
+	${different_pos},       /* different pos */
+	$width             /* width */
 };
 
 EOF
@@ -1236,5 +1289,3 @@ EOF
 
 	return $name;
 }
-
-
