@@ -22,11 +22,12 @@
  * @brief     Helper function for integrated debug support
  * @author    Michael Beck
  * @date      2005
- * @version   $Id: debugger.h 17143 2008-01-02 20:56:33Z beck $
  */
 #ifndef FIRM_DEBUG_DEBUGGER_H
 #define FIRM_DEBUG_DEBUGGER_H
 
+#include "firm_types.h"
+
 /** Break into the debugger. */
 void firm_debug_break(void);
 
@@ -42,4 +43,45 @@ void firm_break(const char *cmd);
 /** Creates the debugger tables. */
 void firm_init_debugger(void);
 
+void firm_finish_debugger(void);
+
+/**
+ * @defgroup external_debug    helper functions for debuggers
+ *
+ * @{
+ */
+
+/**
+ * Returns non-zero, if the debug extension is active
+ */
+int firm_debug_active(void);
+
+/**
+ * Return the content of the debug text buffer.
+ *
+ * To be called from the debugger.
+ */
+const char *firm_debug_text(void);
+
+/**
+ * A gdb helper function to print tarvals.
+ */
+const char *gdb_tarval_helper(void *tv_object);
+
+/**
+ * A gdb helper to print all (new-style-) out edges of a node
+ */
+const char *gdb_out_edge_helper(const ir_node *node);
+
+/**
+ * High level function to use from debugger interface
+ *
+ * See show_commands() for supported commands.
+ */
+void firm_debug(const char *cmd);
+
+/**
+ * @}
+ */
+
 #endif
