@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1995-2008 University of Karlsruhe.  All right reserved.
+ * Copyright (C) 1995-2011 University of Karlsruhe.  All right reserved.
  *
  * This file is part of libFirm.
  *
@@ -21,14 +21,14 @@
  * @file
  * @brief   Combining congruent blocks
  * @author  Michael Beck
- * @version $Id: opt_blocks.c 25900 2009-05-08 10:39:58Z beck $
  *
  * This phase find congruent blocks.
  * Two block are congruent, if they contains only equal calculations.
  */
 #include "config.h"
-#include "ircons.h"
+
 #include "iroptimize.h"
+#include "ircons.h"
 #include "irgmod.h"
 #include "irgraph_t.h"
 #include "irnode_t.h"
@@ -37,7 +37,9 @@
 #include "trouts.h"
 #include "irgwalk.h"
 #include "set.h"
+#include "irpass.h"
 #include "debug.h"
+#include "util.h"
 
 /* define this for general block shaping: congruent blocks
    are found not only before the end block but anywhere in the graph */
@@ -55,13 +57,13 @@ typedef struct pred_t          pred_t;
 
 /** An opcode map key. */
 struct opcode_key_t {
-	ir_opcode   code;   /**< The Firm opcode. */
+	unsigned    code;   /**< The Firm opcode. */
 	ir_mode     *mode;  /**< The mode of all nodes in the partition. */
 	int         arity;  /**< The arity of this opcode (needed for Phi etc. */
 	union {
 		long            proj;   /**< For Proj nodes, its proj number */
 		ir_entity       *ent;   /**< For Sel nodes, its entity */
-		tarval          *tv;    /**< For Const nodes, its tarval */
+		ir_tarval       *tv;    /**< For Const nodes, its tarval */
 		symconst_symbol sym;    /**< For SymConst nodes, its symbol .*/
 		void            *addr;  /**< Alias all addresses. */
 		int             intVal; /**< For Conv/Div nodes: strict/remainderless. */
@@ -152,15 +154,15 @@ typedef struct listmap_t {
 DEBUG_ONLY(static firm_dbg_module_t *dbg;)
 
 /** Next partition number. */
-DEBUG_ONLY(static unsigned part_nr = 0);
+DEBUG_ONLY(static unsigned part_nr = 0;)
 
 #ifdef DEBUG_libfirm
 /**
  * Dump partition to output.
  */
-static void dump_partition(const char *msg, const partition_t *part) {
-	const block_t *block;
-	int           first = 1;
+static void dump_partition(const char *msg, const partition_t *part)
+{
+	int first = 1;
 
 	DB((dbg, LEVEL_2, " %s part%u (%u blocks) {\n  ", msg, part->nr, part->n_blocks));
 	list_for_each_entry(block_t, block, &part->blocks, block_list) {
@@ -173,7 +175,8 @@ static void dump_partition(const char *msg, const partition_t *part) {
 /**
  * Dumps a list.
  */
-static void dump_list(const char *msg, const block_t *block) {
+static void dump_list(const char *msg, const block_t *block)
+{
 	const block_t *p;
 	int           first = 1;
 
@@ -192,9 +195,10 @@ static void dump_list(const char *msg, const block_t *block) {
 /**
  * Compare two pointer values of a listmap.
  */
-static int listmap_cmp_ptr(const void *elt, const void *key, size_t size) {
-	const listmap_entry_t *e1 = elt;
-	const listmap_entry_t *e2 = key;
+static int listmap_cmp_ptr(const void *elt, const void *key, size_t size)
+{
+	const listmap_entry_t *e1 = (const listmap_entry_t*)elt;
+	const listmap_entry_t *e2 = (const listmap_entry_t*)key;
 
 	(void) size;
 	return e1->id != e2->id;
@@ -205,7 +209,8 @@ static int listmap_cmp_ptr(const void *elt, const void *key, size_t size) {
  *
  * @param map  the listmap
  */
-static void listmap_init(listmap_t *map) {
+static void listmap_init(listmap_t *map)
+{
 	map->map    = new_set(listmap_cmp_ptr, 16);
 	map->values = NULL;
 }  /* listmap_init */
@@ -215,7 +220,8 @@ static void listmap_init(listmap_t *map) {
  *
  * @param map  the listmap
  */
-static void listmap_term(listmap_t *map) {
+static void listmap_term(listmap_t *map)
+{
 	del_set(map->map);
 }  /* listmap_term */
 
@@ -227,13 +233,14 @@ static void listmap_term(listmap_t *map) {
  *
  * @return the associated listmap entry for the given id
  */
-static listmap_entry_t *listmap_find(listmap_t *map, void *id) {
+static listmap_entry_t *listmap_find(listmap_t *map, void *id)
+{
 	listmap_entry_t key, *entry;
 
 	key.id   = id;
 	key.list = NULL;
 	key.next = NULL;
-	entry = set_insert(map->map, &key, sizeof(key), HASH_PTR(id));
+	entry    = set_insert(listmap_entry_t, map->map, &key, sizeof(key), hash_ptr(id));
 
 	if (entry->list == NULL) {
 		/* a new entry, put into the list */
@@ -250,17 +257,19 @@ static listmap_entry_t *listmap_find(listmap_t *map, void *id) {
  *
  * @return a hash value for the given opcode map entry
  */
-static unsigned opcode_hash(const opcode_key_t *entry) {
+static unsigned opcode_hash(const opcode_key_t *entry)
+{
 	/* assume long >= int */
-	return (entry->mode - (ir_mode *)0) * 9 + entry->code + entry->u.proj * 3 + HASH_PTR(entry->u.addr) + entry->arity;
+	return (unsigned)(PTR_TO_INT(entry->mode) * 9 + entry->code + entry->u.proj * 3 + hash_ptr(entry->u.addr) + entry->arity);
 }  /* opcode_hash */
 
 /**
  * Compare two entries in the opcode map.
  */
-static int cmp_opcode(const void *elt, const void *key, size_t size) {
-	const opcode_key_t *o1 = elt;
-	const opcode_key_t *o2 = key;
+static int cmp_opcode(const void *elt, const void *key, size_t size)
+{
+	const opcode_key_t *o1 = (opcode_key_t*)elt;
+	const opcode_key_t *o2 = (opcode_key_t*)key;
 
 	(void) size;
 	return o1->code != o2->code || o1->mode != o2->mode ||
@@ -275,13 +284,14 @@ static int cmp_opcode(const void *elt, const void *key, size_t size) {
  * @param meet_block  the control flow meet block of this partition
  * @param env         the environment
  */
-static partition_t *create_partition(ir_node *meet_block, environment_t *env) {
-	partition_t *part = obstack_alloc(&env->obst, sizeof(*part));
+static partition_t *create_partition(ir_node *meet_block, environment_t *env)
+{
+	partition_t *part = OALLOC(&env->obst, partition_t);
 
 	INIT_LIST_HEAD(&part->blocks);
 	part->meet_block = meet_block;
 	part->n_blocks   = 0;
-	DEBUG_ONLY(part->nr = part_nr++);
+	DEBUG_ONLY(part->nr = part_nr++;)
 	list_add_tail(&part->part_list, &env->partitions);
 	return part;
 }  /* create_partition */
@@ -294,8 +304,9 @@ static partition_t *create_partition(ir_node *meet_block, environment_t *env) {
  * @param partition  the partition to add to
  * @param env        the environment
  */
-static block_t *create_block(ir_node *block, int meet_input, partition_t *partition, environment_t *env) {
-	block_t *bl = obstack_alloc(&env->obst, sizeof(*bl));
+static block_t *create_block(ir_node *block, int meet_input, partition_t *partition, environment_t *env)
+{
+	block_t *bl = OALLOC(&env->obst, block_t);
 
 	set_irn_link(block, bl);
 
@@ -326,8 +337,9 @@ static block_t *create_block(ir_node *block, int meet_input, partition_t *partit
  * @param block  the block to add to
  * @param env    the environment
  */
-static node_t *create_node(ir_node *irn, block_t *block, environment_t *env) {
-	node_t *node = obstack_alloc(&env->obst, sizeof(*node));
+static node_t *create_node(ir_node *irn, block_t *block, environment_t *env)
+{
+	node_t *node = OALLOC(&env->obst, node_t);
 
 	node->node     = irn;
 	node->is_input = 0;
@@ -345,8 +357,9 @@ static node_t *create_node(ir_node *irn, block_t *block, environment_t *env) {
  * @param idx    the index of the block input in node's predecessors
  * @param env    the environment
  */
-static void add_pair(block_t *block, ir_node *irn, int idx, environment_t *env) {
-	pair_t *pair = obstack_alloc(&env->obst, sizeof(*pair));
+static void add_pair(block_t *block, ir_node *irn, int idx, environment_t *env)
+{
+	pair_t *pair = OALLOC(&env->obst, pair_t);
 
 	pair->next  = block->input_pairs;
 	pair->irn   = irn;
@@ -363,8 +376,9 @@ static void add_pair(block_t *block, ir_node *irn, int idx, environment_t *env)
  * @param phi    the Phi node
  * @param env    the environment
  */
-static void add_phi(block_t *block, ir_node *phi, environment_t *env) {
-	phi_t *node = obstack_alloc(&env->obst, sizeof(*node));
+static void add_phi(block_t *block, ir_node *phi, environment_t *env)
+{
+	phi_t *node = OALLOC(&env->obst, phi_t);
 
 	node->next = block->phis;
 	node->phi  = phi;
@@ -376,7 +390,8 @@ static void add_phi(block_t *block, ir_node *phi, environment_t *env) {
 /**
  * Creates an opcode from a node.
  */
-static opcode_key_t *opcode(const node_t *node, environment_t *env) {
+static opcode_key_t *opcode(const node_t *node, environment_t *env)
+{
 	opcode_key_t key, *entry;
 	ir_node      *irn = node->node;
 
@@ -406,9 +421,6 @@ static opcode_key_t *opcode(const node_t *node, environment_t *env) {
 	case iro_Const:
 		key.u.tv  = get_Const_tarval(irn);
 		break;
-	case iro_Conv:
-		key.u.intVal = get_Conv_strict(irn);
-		break;
 	case iro_Load:
 		key.mode = get_Load_mode(irn);
 		break;
@@ -422,7 +434,7 @@ static opcode_key_t *opcode(const node_t *node, environment_t *env) {
 		break;
 	}
 
-	entry = set_insert(env->opcode2id_map, &key, sizeof(key), opcode_hash(&key));
+	entry = set_insert(opcode_key_t, env->opcode2id_map, &key, sizeof(key), opcode_hash(&key));
 	return entry;
 }  /* opcode */
 
@@ -435,7 +447,8 @@ static opcode_key_t *opcode(const node_t *node, environment_t *env) {
  *
  * @return  a new partition containing the nodes of g
  */
-static partition_t *split(partition_t *Z, block_t *g, environment_t *env) {
+static partition_t *split(partition_t *Z, block_t *g, environment_t *env)
+{
 	partition_t *Z_prime;
 	block_t     *block;
 	unsigned    n = 0;
@@ -468,7 +481,8 @@ static partition_t *split(partition_t *Z, block_t *g, environment_t *env) {
 /**
  * Return non-zero if pred should be tread as a input node.
  */
-static int is_input_node(ir_node *pred, ir_node *irn, int index) {
+static int is_input_node(ir_node *pred, ir_node *irn, int index)
+{
 	/* for now, do NOT turn direct calls into indirect one */
 	if (index != 1)
 		return 1;
@@ -485,10 +499,10 @@ static int is_input_node(ir_node *pred, ir_node *irn, int index) {
  * @param part  the partition
  * @param env   the environment
  */
-static void propagate_blocks(partition_t *part, environment_t *env) {
+static void propagate_blocks(partition_t *part, environment_t *env)
+{
 	block_t         *ready_blocks = NULL;
 	unsigned        n_ready       = 0;
-	block_t         *bl, *next;
 	listmap_t       map;
 	listmap_entry_t *iter;
 
@@ -523,10 +537,9 @@ static void propagate_blocks(partition_t *part, environment_t *env) {
 			for (i = get_irn_arity(irn) - 1; i >= 0; --i) {
 				ir_node *pred  = get_irn_n(irn, i);
 				ir_node *block = get_nodes_block(skip_Proj(pred));
-				node_t *p_node;
 
 				if (block != bl->block) {
-					p_node = create_node(pred, bl, env);
+					node_t *p_node = create_node(pred, bl, env);
 					if (is_input_node(pred, irn, i)) {
 						/* is a block live input */
 						p_node->is_input = 1;
@@ -538,7 +551,7 @@ static void propagate_blocks(partition_t *part, environment_t *env) {
 					}
 				} else if (! irn_visited_else_mark(pred)) {
 					/* not yet visited, ok */
-					p_node = create_node(pred, bl, env);
+					create_node(pred, bl, env);
 
 					if (is_Phi(pred)) {
 						/* update the Phi list */
@@ -596,9 +609,8 @@ static void propagate_blocks(partition_t *part, environment_t *env) {
  *
  * @param env    the environment
  */
-static void propagate(environment_t *env) {
-	partition_t *part, *next;
-
+static void propagate(environment_t *env)
+{
 	list_for_each_entry_safe(partition_t, part, next, &env->partitions, part_list) {
 		if (part->n_blocks < 2) {
 			/* zero or one block left, kill this partition */
@@ -612,7 +624,8 @@ static void propagate(environment_t *env) {
 /**
  * Map a block to the phi[block->input] live-trough.
  */
-static void *live_throughs(const block_t *bl, const ir_node *phi) {
+static void *live_throughs(const block_t *bl, const ir_node *phi)
+{
 	ir_node *input = get_Phi_pred(phi, bl->meet_input);
 
 	/* If this input is inside our block, this
@@ -630,9 +643,9 @@ static void *live_throughs(const block_t *bl, const ir_node *phi) {
  * @param part  the partition
  * @param env   the environment
  */
-void propagate_blocks_live_troughs(partition_t *part, environment_t *env) {
+static void propagate_blocks_live_troughs(partition_t *part, environment_t *env)
+{
 	const ir_node   *meet_block = part->meet_block;
-	block_t         *bl, *next;
 	listmap_t       map;
 	listmap_entry_t *iter;
 	const ir_node   *phi;
@@ -656,7 +669,7 @@ void propagate_blocks_live_troughs(partition_t *part, environment_t *env) {
 			listmap_entry_t *entry;
 
 			/* Add bl to map[live_trough(bl)]. */
-			id          = live_throughs(bl, phi);
+			id          = (opcode_key_t*)live_throughs(bl, phi);
 			entry       = listmap_find(&map, id);
 			bl->next    = entry->list;
 			entry->list = bl;
@@ -684,9 +697,8 @@ void propagate_blocks_live_troughs(partition_t *part, environment_t *env) {
  *
  * @param env    the environment
  */
-void propagate_live_troughs(environment_t *env) {
-	partition_t *part, *next;
-
+static void propagate_live_troughs(environment_t *env)
+{
 	list_for_each_entry_safe(partition_t, part, next, &env->partitions, part_list) {
 		propagate_blocks_live_troughs(part, env);
 	}
@@ -703,14 +715,14 @@ void propagate_live_troughs(environment_t *env) {
  *
  * @param part  the partition to process
  */
-static void apply(ir_graph *irg, partition_t *part) {
+static void apply(ir_graph *irg, partition_t *part)
+{
 	block_t *repr = list_entry(part->blocks.next, block_t, block_list);
-	block_t *bl;
 	ir_node *block, *end, *meet_block, *p, *next;
 	ir_node **ins, **phi_ins;
 	phi_t   *repr_phi, *phi;
 	pair_t  *repr_pair, *pair;
-	int     i, j, k, n, block_nr, n_phis;
+	int     i, j, k, n, n_phis;
 
 	list_del(&repr->block_list);
 
@@ -744,10 +756,8 @@ static void apply(ir_graph *irg, partition_t *part) {
 
 	/* collect new in arrays */
 	end = get_irg_end(irg);
-	block_nr = 0;
 	list_for_each_entry(block_t, bl, &part->blocks, block_list) {
 		block = bl->block;
-		++block_nr;
 
 		DB((dbg, LEVEL_1, "%+F, ", block));
 
@@ -772,7 +782,7 @@ static void apply(ir_graph *irg, partition_t *part) {
 		}
 
 		/* third step: update Phis */
-		for (repr_phi = repr->phis, phi = bl->phis; 
+		for (repr_phi = repr->phis, phi = bl->phis;
 		     repr_phi != NULL;
 		     repr_phi = repr_phi->next, phi = phi->next) {
 			for (i = 0; i < n; ++i) {
@@ -796,8 +806,8 @@ static void apply(ir_graph *irg, partition_t *part) {
 
 	/* rewire block input ... */
 	n = ARR_LEN(ins);
-	
-	/* 
+
+	/*
 	 * Some problem here. For:
 	 * if (x) y = 1; else y = 2;
 	 *
@@ -826,7 +836,7 @@ static void apply(ir_graph *irg, partition_t *part) {
 		if (is_op_forking(cfop)) {
 			/* a critical edge */
 			ir_node *block = new_r_Block(irg, 1, &ins[i]);
-			ir_node *jmp   = new_r_Jmp(irg, block);
+			ir_node *jmp   = new_r_Jmp(block);
 			ins[i] = jmp;
 		}
 	}
@@ -845,7 +855,7 @@ static void apply(ir_graph *irg, partition_t *part) {
 	for (repr_pair = repr->input_pairs; repr_pair != NULL; repr_pair = repr_pair->next) {
 		ir_node *input = get_irn_n(repr_pair->irn, repr_pair->index);
 		ir_mode *mode  = get_irn_mode(input);
-		ir_node *phi   = new_r_Phi(current_ir_graph, block, n, repr_pair->ins, mode);
+		ir_node *phi   = new_r_Phi(block, n, repr_pair->ins, mode);
 
 		set_irn_n(repr_pair->irn, repr_pair->index, phi);
 		DEL_ARR_F(repr_pair->ins);
@@ -858,14 +868,14 @@ static void apply(ir_graph *irg, partition_t *part) {
 	/* ... finally rewire the meet block and fix its Phi-nodes */
 	meet_block = part->meet_block;
 	n         = get_Block_n_cfgpreds(meet_block);
-	
+
 	ins = NEW_ARR_F(ir_node *, n);
 
 	n_phis = 0;
 	for (p = get_Block_phis(meet_block); p != NULL; p = get_Phi_next(p)) {
 		++n_phis;
 	}
-	
+
 	phi_ins = NEW_ARR_F(ir_node *, n_phis * n);
 
 	for (i = j = 0; i < n; ++i) {
@@ -915,44 +925,45 @@ continue_outer:
  * @param end_block  the end block
  * @param env        the environment
  */
-static void partition_for_end_block(ir_node *end_block, environment_t *env) {
+static void partition_for_end_block(ir_node *end_block, environment_t *env)
+{
 	partition_t *part = create_partition(end_block, env);
 	ir_node     *end;
 	int         i;
-	
+
 	/* collect normal blocks */
 	for (i = get_Block_n_cfgpreds(end_block) - 1; i >= 0; --i) {
 		ir_node *pred = get_Block_cfgpred(end_block, i);
 		ir_node *block;
 		block_t *bl;
 		node_t  *node;
-		
+
 		mark_irn_visited(pred);
-		
+
 		block = get_nodes_block(pred);
 		bl    = create_block(block, i, part, env);
 		node  = create_node(pred, bl, env);
-		
+
 		bl->cf_root = node;
 	}
-	
+
 	/* collect all no-return blocks */
-	end = get_irg_end(current_ir_graph);
+	end = get_irg_end(get_irn_irg(end_block));
 	for (i = get_End_n_keepalives(end) - 1; i >= 0; --i) {
 		ir_node *ka    = get_End_keepalive(end, i);
 		ir_node *block;
 		block_t *bl;
 		node_t  *node;
-		
+
 		if (! is_Call(ka))
 			continue;
 		mark_irn_visited(ka);
-		
+
 		/* found one */
 		block = get_nodes_block(ka);
 		bl    = create_block(block, -1, part, env);
 		node  = create_node(ka, bl, env);
-		
+
 		bl->cf_root = node;
 	}
 
@@ -968,25 +979,26 @@ static void partition_for_end_block(ir_node *end_block, environment_t *env) {
  * @param n_preds  number of elements in preds
  * @param env      the environment
  */
-static void partition_for_block(ir_node *block, pred_t preds[], int n_preds, environment_t *env) {
+static void partition_for_block(ir_node *block, pred_t preds[], int n_preds, environment_t *env)
+{
 	partition_t *part = create_partition(block, env);
 	int         i;
-	
+
 	for (i = n_preds - 1; i >= 0; --i) {
 		ir_node *pred = preds[i].pred;
 		ir_node *block;
 		block_t *bl;
 		node_t  *node;
-		
+
 		mark_irn_visited(pred);
-		
+
 		block = get_nodes_block(pred);
 		bl    = create_block(block, preds[i].index, part, env);
 		node  = create_node(pred, bl, env);
-		
+
 		bl->cf_root = node;
 	}
-		
+
 	dump_partition("Created", part);
 }  /* partition_for_block */
 
@@ -994,7 +1006,8 @@ static void partition_for_block(ir_node *block, pred_t preds[], int n_preds, env
  * Walker: clear the links of all block phi lists and normal
  * links.
  */
-static void clear_phi_links(ir_node *irn, void *env) {
+static void clear_phi_links(ir_node *irn, void *env)
+{
 	(void) env;
 	if (is_Block(irn)) {
 		set_Block_phis(irn, NULL);
@@ -1005,8 +1018,9 @@ static void clear_phi_links(ir_node *irn, void *env) {
 /**
  * Walker, detect live-out nodes.
  */
-static void find_liveouts(ir_node *irn, void *ctx) {
-	environment_t *env        = ctx;
+static void find_liveouts(ir_node *irn, void *ctx)
+{
+	environment_t *env        = (environment_t*)ctx;
 	ir_node       **live_outs = env->live_outs;
 	ir_node       *this_block;
 	int           i;
@@ -1045,14 +1059,15 @@ static void find_liveouts(ir_node *irn, void *ctx) {
 }  /* find_liveouts */
 
 /**
- * Check if the current block is the meet block of a its predecessors.
+ * Check if the current block is the meet block of its predecessors.
  */
-static void check_for_cf_meet(ir_node *block, void *ctx) {
-	environment_t *env = ctx;
+static void check_for_cf_meet(ir_node *block, void *ctx)
+{
+	environment_t *env = (environment_t*)ctx;
 	int           i, k, n;
 	pred_t        *preds;
 
-	if (block == get_irg_end_block(current_ir_graph)) {
+	if (block == get_irg_end_block(get_irn_irg(block))) {
 		/* always create a partition for the end block */
 		partition_for_end_block(block, env);
 		return;
@@ -1068,16 +1083,14 @@ static void check_for_cf_meet(ir_node *block, void *ctx) {
 	k = 0;
 	for (i = n - 1; i >= 0; --i) {
 		ir_node *pred = get_Block_cfgpred(block, i);
-		ir_node *pred_block;
 
 		/* pred must be a direct jump to us */
 		if (! is_Jmp(pred) && ! is_Raise(pred) && !is_Bad(pred))
 			continue;
 
-		pred_block = get_nodes_block(skip_Proj(pred));
-
 		preds[k].pred  = pred;
 		preds[k].index = i;
+		++k;
 	}
 
 	if (k > 1)
@@ -1087,13 +1100,14 @@ static void check_for_cf_meet(ir_node *block, void *ctx) {
 /**
  * Compare two nodes for root ordering.
  */
-static int cmp_nodes(const void *a, const void *b) {
-	ir_node *const *pa     = a;
-	ir_node *const *pb     = b;
+static int cmp_nodes(const void *a, const void *b)
+{
+	const ir_node *const *pa = (const ir_node*const*)a;
+	const ir_node *const *pb = (const ir_node*const*)b;
 	const ir_node  *irn_a  = *pa;
 	const ir_node  *irn_b  = *pb;
-	ir_opcode      code_a  = get_irn_opcode(irn_a);
-	ir_opcode      code_b  = get_irn_opcode(irn_b);
+	unsigned       code_a  = get_irn_opcode(irn_a);
+	unsigned       code_b  = get_irn_opcode(irn_b);
 	ir_mode        *mode_a, *mode_b;
 	unsigned       idx_a, idx_b;
 
@@ -1118,7 +1132,8 @@ static int cmp_nodes(const void *a, const void *b) {
 /**
  * Add the roots to all blocks.
  */
-static void add_roots(ir_graph *irg, environment_t *env) {
+static void add_roots(ir_graph *irg, environment_t *env)
+{
 	unsigned idx, n      = get_irg_last_idx(irg);
 	ir_node  **live_outs = env->live_outs;
 	block_t  *bl;
@@ -1143,7 +1158,7 @@ static void add_roots(ir_graph *irg, environment_t *env) {
      * Else, we will split identical blocks if we start which different roots.
 	 */
 	for (bl = env->all_blocks; bl != NULL; bl = bl->all_next) {
-		int i, n = ARR_LEN(bl->roots);
+		size_t i, n = ARR_LEN(bl->roots);
 
 #if 1
 		/* TODO: is this really needed? The roots are already in
@@ -1165,20 +1180,16 @@ static void add_roots(ir_graph *irg, environment_t *env) {
 #endif /* GENERAL_SHAPE */
 
 /* Combines congruent end blocks into one. */
-int shape_blocks(ir_graph *irg) {
-	ir_graph      *rem;
+void shape_blocks(ir_graph *irg)
+{
 	environment_t env;
-	partition_t   *part;
 	block_t       *bl;
 	int           res, n;
 
-	rem = current_ir_graph;
-	current_ir_graph = irg;
-
 	/* register a debug mask */
 	FIRM_DBG_REGISTER(dbg, "firm.opt.blocks");
 
-	DEBUG_ONLY(part_nr = 0);
+	DEBUG_ONLY(part_nr = 0;)
 	DB((dbg, LEVEL_1, "Shaping blocks for %+F\n", irg));
 
 	/* works better, when returns are placed at the end of the blocks */
@@ -1236,15 +1247,7 @@ int shape_blocks(ir_graph *irg) {
 
 	if (res) {
 		/* control flow changed */
-		set_irg_outs_inconsistent(irg);
-		set_irg_extblk_inconsistent(irg);
-		set_irg_doms_inconsistent(irg);
-		set_irg_loopinfo_inconsistent(irg);
-
-		/* Calls might be removed. */
-		set_trouts_inconsistent();
-
-	//	dump_ir_block_graph(irg, "-after");
+		clear_irg_properties(irg, IR_GRAPH_PROPERTY_CONSISTENT_DOMINANCE);
 	}
 
 	for (bl = env.all_blocks; bl != NULL; bl = bl->all_next) {
@@ -1254,7 +1257,9 @@ int shape_blocks(ir_graph *irg) {
 	DEL_ARR_F(env.live_outs);
 	del_set(env.opcode2id_map);
 	obstack_free(&env.obst, NULL);
-	current_ir_graph = rem;
-
-	return res;
 }  /* shape_blocks */
+
+ir_graph_pass_t *shape_blocks_pass(const char *name)
+{
+	return def_graph_pass(name ? name : "shape_blocks", shape_blocks);
+}  /* shape_blocks_pass */
