@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1995-2008 University of Karlsruhe.  All right reserved.
+ * Copyright (C) 1995-2011 University of Karlsruhe.  All right reserved.
  *
  * This file is part of libFirm.
  *
@@ -21,14 +21,15 @@
  * @file
  * @brief   Scalar replacement of compounds.
  * @author  Beyhan Veliev, Michael Beck
- * @version $Id: scalar_replace.c 25646 2009-03-09 14:54:35Z beck $
  */
 #include "config.h"
 
 #include <string.h>
+#include <stdbool.h>
 
 #include "iroptimize.h"
 #include "scalar_replace.h"
+#include "opt_init.h"
 #include "irflag_t.h"
 #include "irouts.h"
 #include "set.h"
@@ -40,13 +41,21 @@
 #include "irgwalk.h"
 #include "irgmod.h"
 #include "irnode_t.h"
-#include "irtools.h"
+#include "irpass.h"
+#include "util.h"
 #include "xmalloc.h"
 #include "debug.h"
 #include "error.h"
 
-#define SET_VNUM(node, vnum) set_irn_link(node, INT_TO_PTR(vnum))
-#define GET_VNUM(node)       (unsigned)PTR_TO_INT(get_irn_link(node))
+static unsigned get_vnum(const ir_node *node)
+{
+	return (unsigned)PTR_TO_INT(get_irn_link(node));
+}
+
+static void set_vnum(ir_node *node, unsigned vnum)
+{
+	set_irn_link(node, INT_TO_PTR(vnum));
+}
 
 /**
  * A path element entry: it is either an entity
@@ -55,25 +64,27 @@
  */
 typedef union {
 	ir_entity *ent;
-	tarval *tv;
+	ir_tarval *tv;
 } path_elem_t;
 
 /**
  * An access path, used to assign value numbers
  * to variables that will be scalar replaced.
  */
-typedef struct _path_t {
+typedef struct path_t {
 	unsigned    vnum;      /**< The value number. */
-	unsigned    path_len;  /**< The length of the access path. */
+	size_t      path_len;  /**< The length of the access path. */
 	path_elem_t path[1];   /**< The path. */
 } path_t;
 
 /** The size of a path in bytes. */
-#define PATH_SIZE(p)  (sizeof(*(p)) + sizeof((p)->path[0]) * ((p)->path_len - 1))
+static size_t path_size(path_t *p)
+{
+	return sizeof(*p) + sizeof(p->path[0]) * (p->path_len-1);
+}
 
-typedef struct _scalars_t {
+typedef struct scalars_t {
 	ir_entity *ent;              /**< A entity for scalar replacement. */
-	ir_type *ent_owner;          /**< The owner of this entity. */
 } scalars_t;
 
 DEBUG_ONLY(static firm_dbg_module_t *dbg;)
@@ -83,9 +94,10 @@ DEBUG_ONLY(static firm_dbg_module_t *dbg;)
  *
  * @return 0 if they are identically
  */
-static int path_cmp(const void *elt, const void *key, size_t size) {
-	const path_t *p1 = elt;
-	const path_t *p2 = key;
+static int path_cmp(const void *elt, const void *key, size_t size)
+{
+	const path_t *p1 = (const path_t*)elt;
+	const path_t *p2 = (const path_t*)key;
 	(void) size;
 
 	/* we can use memcmp here, because identical tarvals should have identical addresses */
@@ -97,9 +109,10 @@ static int path_cmp(const void *elt, const void *key, size_t size) {
  *
  * @return 0 if they are identically
  */
-static int ent_cmp(const void *elt, const void *key, size_t size) {
-	const scalars_t *c1 = elt;
-	const scalars_t *c2 = key;
+static int ent_cmp(const void *elt, const void *key, size_t size)
+{
+	const scalars_t *c1 = (const scalars_t*)elt;
+	const scalars_t *c2 = (const scalars_t*)key;
 	(void) size;
 
 	return c1->ent != c2->ent;
@@ -108,7 +121,8 @@ static int ent_cmp(const void *elt, const void *key, size_t size) {
 /**
  * Calculate a hash value for a path.
  */
-static unsigned path_hash(const path_t *path) {
+static unsigned path_hash(const path_t *path)
+{
 	unsigned hash = 0;
 	unsigned i;
 
@@ -123,16 +137,17 @@ static unsigned path_hash(const path_t *path) {
  *
  * @param sel  the Sel node that will be checked
  */
-static int is_const_sel(ir_node *sel) {
+static bool is_const_sel(ir_node *sel)
+{
 	int i, n = get_Sel_n_indexs(sel);
 
 	for (i = 0; i < n; ++i) {
 		ir_node *idx = get_Sel_index(sel, i);
 
 		if (!is_Const(idx))
-			return 0;
+			return false;
 	}
-	return 1;
+	return true;
 }
 
 /**
@@ -152,62 +167,62 @@ static int is_const_sel(ir_node *sel) {
  * @param mode     the mode of the Load/Store
  * @param ent_mode the mode of the accessed entity
  */
-static int check_load_store_mode(ir_mode *mode, ir_mode *ent_mode) {
+static bool check_load_store_mode(ir_mode *mode, ir_mode *ent_mode)
+{
 	if (ent_mode != mode) {
 		if (ent_mode == NULL ||
 		    get_mode_size_bits(ent_mode) != get_mode_size_bits(mode) ||
-		    get_mode_sort(ent_mode) != get_mode_sort(mode) ||
 		    get_mode_arithmetic(ent_mode) != irma_twos_complement ||
 		    get_mode_arithmetic(mode) != irma_twos_complement)
-			return 0;
+			return false;
 	}
-	return 1;
+	return true;
 }
 
 /*
  * Returns non-zero, if the address of an entity
- * represented by a Sel node (or it's successor Sels) is taken.
+ * represented by a Sel node (or its successor Sels) is taken.
  */
-int is_address_taken(ir_node *sel)
+bool is_address_taken(ir_node *sel)
 {
-	int       i, input_nr, k;
+	int       input_nr;
 	ir_mode   *emode, *mode;
 	ir_node   *value;
 	ir_entity *ent;
 
 	if (! is_const_sel(sel))
-		return 1;
+		return true;
 
-	for (i = get_irn_n_outs(sel) - 1; i >= 0; --i) {
+	for (unsigned i = get_irn_n_outs(sel); i-- > 0; ) {
 		ir_node *succ = get_irn_out(sel, i);
 
 		switch (get_irn_opcode(succ)) {
 		case iro_Load:
 			/* do not remove volatile variables */
 			if (get_Load_volatility(succ) == volatility_is_volatile)
-				return 1;
+				return true;
 			/* check if this load is not a hidden conversion */
 			mode = get_Load_mode(succ);
 			ent = get_Sel_entity(sel);
 			emode = get_type_mode(get_entity_type(ent));
 			if (! check_load_store_mode(mode, emode))
-				return 1;
+				return true;
 			break;
 
 		case iro_Store:
 			/* check that Sel is not the Store's value */
 			value = get_Store_value(succ);
 			if (value == sel)
-				return 1;
+				return true;
 			/* do not remove volatile variables */
 			if (get_Store_volatility(succ) == volatility_is_volatile)
-				return 1;
+				return true;
 			/* check if this Store is not a hidden conversion */
 			mode = get_irn_mode(value);
 			ent = get_Sel_entity(sel);
 			emode = get_type_mode(get_entity_type(ent));
 			if (! check_load_store_mode(mode, emode))
-				return 1;
+				return true;
 			break;
 
 		case iro_Sel: {
@@ -215,12 +230,12 @@ int is_address_taken(ir_node *sel)
 			ir_entity *entity = get_Sel_entity(succ);
 			/* we can't handle unions correctly yet -> address taken */
 			if (is_Union_type(get_entity_owner(entity)))
-				return 1;
+				return true;
 
 			/* Check the Sel successor of Sel */
 			res = is_address_taken(succ);
 			if (res)
-				return 1;
+				return true;
 			break;
 		}
 
@@ -231,12 +246,12 @@ int is_address_taken(ir_node *sel)
 			 * One special case: If the Call type tells that it's a
 			 * value parameter, the address is NOT taken.
 			 */
-			return 1;
+			return true;
 
 		case iro_Id: {
 			int res = is_address_taken(succ);
 			if (res)
-				return 1;
+				return true;
 			break;
 		}
 
@@ -247,13 +262,13 @@ int is_address_taken(ir_node *sel)
 
 				if (pred == sel) {
 					/* we found one input */
-					for (k = get_irn_n_outs(succ) - 1; k >= 0; --k) {
+					for (unsigned k = get_irn_n_outs(succ); k-- > 0; ) {
 						ir_node *proj = get_irn_out(succ, k);
 
 						if (is_Proj(proj) && get_Proj_proj(proj) == input_nr) {
 							int res = is_address_taken(proj);
 							if (res)
-								return 1;
+								return true;
 						}
 					}
 				}
@@ -262,10 +277,10 @@ int is_address_taken(ir_node *sel)
 
 		default:
 			/* another op, the address is taken */
-			return 1;
+			return true;
 		}
 	}
-	return 0;
+	return false;
 }
 
 /**
@@ -274,15 +289,16 @@ int is_address_taken(ir_node *sel)
  * @param ent  the entity that will be scalar replaced
  * @param sel  a Sel node that selects some fields of this entity
  */
-static int link_all_leave_sels(ir_entity *ent, ir_node *sel) {
-	int i, is_leave = 1;
+static bool link_all_leave_sels(ir_entity *ent, ir_node *sel)
+{
+	bool is_leave = true;
 
-	for (i = get_irn_n_outs(sel) - 1; i >= 0; --i) {
+	for (unsigned i = get_irn_n_outs(sel); i-- > 0; ) {
 		ir_node *succ = get_irn_out(sel, i);
 
 		if (is_Sel(succ)) {
 			/* the current leave has further Sel's, no leave */
-			is_leave = 0;
+			is_leave = false;
 			link_all_leave_sels(ent, succ);
 		} else if (is_Id(succ)) {
 			is_leave &= link_all_leave_sels(ent, succ);
@@ -314,44 +330,46 @@ static void *ADDRESS_TAKEN = &_x;
  *
  * This function finds variables on the (members of the) frame type
  * that can be scalar replaced, because their address is never taken.
- * If such a variable is found, it's entity link will hold a list of all
+ * If such a variable is found, its entity link will hold a list of all
  * Sel nodes, that selects the atomic fields of this entity.
  * Otherwise, the link will be ADDRESS_TAKEN or NULL.
  *
  * @return  non-zero if at least one entity could be replaced
  *          potentially
  */
-static int find_possible_replacements(ir_graph *irg) {
+static int find_possible_replacements(ir_graph *irg)
+{
 	ir_node *irg_frame;
 	ir_type *frame_tp;
-	int     i, j, k, static_link_arg;
+	size_t  mem_idx;
+	long    static_link_arg;
 	int     res = 0;
 
 	/*
 	 * First, clear the link field of all interesting entities.
 	 */
 	frame_tp = get_irg_frame_type(irg);
-	for (i = get_class_n_members(frame_tp) - 1; i >= 0; --i) {
-		ir_entity *ent = get_class_member(frame_tp, i);
+	for (mem_idx = get_class_n_members(frame_tp); mem_idx > 0;) {
+		ir_entity *ent = get_class_member(frame_tp, --mem_idx);
 		set_entity_link(ent, NULL);
 	}
 
 	/* check for inner functions:
 	 * FIXME: need a way to get the argument position for the static link */
 	static_link_arg = 0;
-	for (i = get_class_n_members(frame_tp) - 1; i >= 0; --i) {
-		ir_entity *ent = get_class_member(frame_tp, i);
+	for (mem_idx = get_class_n_members(frame_tp); mem_idx > 0;) {
+		ir_entity *ent = get_class_member(frame_tp, --mem_idx);
 		if (is_method_entity(ent)) {
 			ir_graph *inner_irg = get_entity_irg(ent);
 			ir_node  *args;
 
-			assure_irg_outs(inner_irg);
+			assure_irg_properties(inner_irg, IR_GRAPH_PROPERTY_CONSISTENT_OUTS);
 			args = get_irg_args(inner_irg);
-			for (j = get_irn_n_outs(args) - 1; j >= 0; --j) {
+			for (unsigned j = get_irn_n_outs(args); j-- > 0; ) {
 				ir_node *arg = get_irn_out(args, j);
 
 				if (get_Proj_proj(arg) == static_link_arg) {
-					for (k = get_irn_n_outs(arg) - 1; k >= 0; --k) {
+					for (unsigned k = get_irn_n_outs(arg); k-- > 0; ) {
 						ir_node *succ = get_irn_out(arg, k);
 
 						if (is_Sel(succ)) {
@@ -374,7 +392,7 @@ static int find_possible_replacements(ir_graph *irg) {
 	 * equal ADDRESS_TAKEN.
 	 */
 	irg_frame = get_irg_frame(irg);
-	for (i = get_irn_n_outs(irg_frame) - 1; i >= 0; --i) {
+	for (unsigned i = get_irn_n_outs(irg_frame); i-- > 0; ) {
 		ir_node *succ = get_irn_out(irg_frame, i);
 
 		if (is_Sel(succ)) {
@@ -389,17 +407,6 @@ static int find_possible_replacements(ir_graph *irg) {
 			if (get_entity_link(ent) == ADDRESS_TAKEN)
 				continue;
 
-			/*
-			 * Beware: in rare cases even entities on the frame might be
-			 * volatile. This might happen if the entity serves as a store
-			 * to a value that must survive a exception. Do not optimize
-			 * such entities away.
-			 */
-			if (get_entity_volatility(ent) == volatility_is_volatile) {
-				set_entity_link(ent, ADDRESS_TAKEN);
-				continue;
-			}
-
 			ent_type = get_entity_type(ent);
 
 			/* we can handle arrays, structs and atomic types yet */
@@ -423,13 +430,15 @@ static int find_possible_replacements(ir_graph *irg) {
 }
 
 /**
- * Return a path from the Sel node sel to it's root.
+ * Return a path from the Sel node "sel" to its root.
  *
  * @param sel  the Sel node
  * @param len  the length of the path so far
  */
-static path_t *find_path(ir_node *sel, unsigned len) {
-	int pos, i, n;
+static path_t *find_path(ir_node *sel, size_t len)
+{
+	size_t pos;
+	int i, n;
 	path_t *res;
 	ir_node *pred = get_Sel_ptr(sel);
 
@@ -444,6 +453,7 @@ static path_t *find_path(ir_node *sel, unsigned len) {
 	} else
 		res = find_path(pred, len);
 
+	assert(len <= res->path_len);
 	pos = res->path_len - len;
 
 	res->path[pos++].ent = get_Sel_entity(sel);
@@ -476,27 +486,27 @@ static unsigned allocate_value_numbers(pset *sels, ir_entity *ent, unsigned vnum
 
 	DB((dbg, SET_LEVEL_3, "  Visiting Sel nodes of entity %+F\n", ent));
 	/* visit all Sel nodes in the chain of the entity */
-	for (sel = get_entity_link(ent); sel; sel = next) {
-		next = get_irn_link(sel);
+	for (sel = (ir_node*)get_entity_link(ent); sel != NULL; sel = next) {
+		next = (ir_node*)get_irn_link(sel);
 
 		/* we must mark this sel for later */
 		pset_insert_ptr(sels, sel);
 
 		key  = find_path(sel, 0);
-		path = set_find(pathes, key, PATH_SIZE(key), path_hash(key));
+		path = set_find(path_t, pathes, key, path_size(key), path_hash(key));
 
 		if (path) {
-			SET_VNUM(sel, path->vnum);
+			set_vnum(sel, path->vnum);
 			DB((dbg, SET_LEVEL_3, "  %+F represents value %u\n", sel, path->vnum));
 		} else {
 			key->vnum = vnum++;
 
-			set_insert(pathes, key, PATH_SIZE(key), path_hash(key));
+			(void)set_insert(path_t, pathes, key, path_size(key), path_hash(key));
 
-			SET_VNUM(sel, key->vnum);
+			set_vnum(sel, key->vnum);
 			DB((dbg, SET_LEVEL_3, "  %+F represents value %u\n", sel, key->vnum));
 
-			ARR_EXTO(ir_mode *, *modes, (int)((key->vnum + 15) & ~15));
+			ARR_EXTO(ir_mode *, *modes, (key->vnum + 15) & ~15);
 
 			(*modes)[key->vnum] = get_type_mode(get_entity_type(get_Sel_entity(sel)));
 
@@ -525,50 +535,44 @@ static unsigned allocate_value_numbers(pset *sels, ir_entity *ent, unsigned vnum
 	return vnum;
 }
 
-/**
- * A list entry for the fixing lists
- */
-typedef struct _list_entry_t {
-	ir_node  *node;   /**< the node that must be fixed */
-	unsigned vnum;    /**< the value number of this node */
-} list_entry_t;
-
 /**
  * environment for memory walker
  */
-typedef struct _env_t {
-	int          nvals;       /**< number of values */
-	ir_mode      **modes;     /**< the modes of the values */
-	pset         *sels;       /**< A set of all Sel nodes that have a value number */
+typedef struct env_t {
+	unsigned nvals;      /**< number of values */
+	ir_mode  **modes;    /**< the modes of the values */
+	pset     *sels;      /**< A set of all Sel nodes that have a value number */
 } env_t;
 
 /**
  * topological post-walker.
  */
-static void topologic_walker(ir_node *node, void *ctx) {
-	env_t        *env = ctx;
-	ir_node      *adr, *block, *mem, *val;
-	ir_mode      *mode;
-	unsigned     vnum;
+static void walker(ir_node *node, void *ctx)
+{
+	env_t    *env = (env_t*)ctx;
+	ir_graph *irg = get_irn_irg(node);
+	ir_node  *addr, *block, *mem, *val;
+	ir_mode  *mode;
+	unsigned vnum;
 
 	if (is_Load(node)) {
 		/* a load, check if we can resolve it */
-		adr = get_Load_ptr(node);
+		addr = get_Load_ptr(node);
 
 		DB((dbg, SET_LEVEL_3, "  checking %+F for replacement ", node));
-		if (! is_Sel(adr)) {
-			DB((dbg, SET_LEVEL_3, "no Sel input (%+F)\n", adr));
+		if (! is_Sel(addr)) {
+			DB((dbg, SET_LEVEL_3, "no Sel input (%+F)\n", addr));
 			return;
 		}
 
-		if (! pset_find_ptr(env->sels, adr)) {
-			DB((dbg, SET_LEVEL_3, "Sel %+F has no VNUM\n", adr));
+		if (! pset_find_ptr(env->sels, addr)) {
+			DB((dbg, SET_LEVEL_3, "Sel %+F has no VNUM\n", addr));
 			return;
 		}
 
 		/* ok, we have a Load that will be replaced */
-		vnum = GET_VNUM(adr);
-		assert(vnum < (unsigned)env->nvals);
+		vnum = get_vnum(addr);
+		assert(vnum < env->nvals);
 
 		DB((dbg, SET_LEVEL_3, "replacing by value %u\n", vnum));
 
@@ -587,32 +591,32 @@ static void topologic_walker(ir_node *node, void *ctx) {
 		Handle this here. */
 		mode = get_Load_mode(node);
 		if (mode != get_irn_mode(val))
-			val = new_d_Conv(get_irn_dbg_info(node), val, mode);
+			val = new_rd_Conv(get_irn_dbg_info(node), block, val, mode);
 
 		mem = get_Load_mem(node);
-		turn_into_tuple(node, pn_Load_max);
+		turn_into_tuple(node, pn_Load_max+1);
 		set_Tuple_pred(node, pn_Load_M,         mem);
 		set_Tuple_pred(node, pn_Load_res,       val);
-		set_Tuple_pred(node, pn_Load_X_regular, new_Jmp());
-		set_Tuple_pred(node, pn_Load_X_except,  new_Bad());
+		set_Tuple_pred(node, pn_Load_X_regular, new_r_Jmp(block));
+		set_Tuple_pred(node, pn_Load_X_except,  new_r_Bad(irg, mode_X));
 	} else if (is_Store(node)) {
 		DB((dbg, SET_LEVEL_3, "  checking %+F for replacement ", node));
 
 		/* a Store always can be replaced */
-		adr = get_Store_ptr(node);
+		addr = get_Store_ptr(node);
 
-		if (! is_Sel(adr)) {
-			DB((dbg, SET_LEVEL_3, "no Sel input (%+F)\n", adr));
+		if (! is_Sel(addr)) {
+			DB((dbg, SET_LEVEL_3, "no Sel input (%+F)\n", addr));
 			return;
 		}
 
-		if (! pset_find_ptr(env->sels, adr)) {
-			DB((dbg, SET_LEVEL_3, "Sel %+F has no VNUM\n", adr));
+		if (! pset_find_ptr(env->sels, addr)) {
+			DB((dbg, SET_LEVEL_3, "Sel %+F has no VNUM\n", addr));
 			return;
 		}
 
-		vnum = GET_VNUM(adr);
-		assert(vnum < (unsigned)env->nvals);
+		vnum = get_vnum(addr);
+		assert(vnum < env->nvals);
 
 		DB((dbg, SET_LEVEL_3, "replacing by value %u\n", vnum));
 
@@ -622,15 +626,15 @@ static void topologic_walker(ir_node *node, void *ctx) {
 		/* Beware: A Store can contain a hidden conversion in Firm. */
 		val = get_Store_value(node);
 		if (get_irn_mode(val) != env->modes[vnum])
-			val = new_d_Conv(get_irn_dbg_info(node), val, env->modes[vnum]);
+			val = new_rd_Conv(get_irn_dbg_info(node), block, val, env->modes[vnum]);
 
 		set_value(vnum, val);
 
 		mem = get_Store_mem(node);
-		turn_into_tuple(node, pn_Store_max);
+		turn_into_tuple(node, pn_Store_max+1);
 		set_Tuple_pred(node, pn_Store_M,         mem);
-		set_Tuple_pred(node, pn_Store_X_regular, new_Jmp());
-		set_Tuple_pred(node, pn_Store_X_except,  new_Bad());
+		set_Tuple_pred(node, pn_Store_X_regular, new_r_Jmp(block));
+		set_Tuple_pred(node, pn_Store_X_except,  new_r_Bad(irg, mode_X));
 	}
 }
 
@@ -642,23 +646,25 @@ static void topologic_walker(ir_node *node, void *ctx) {
  * @param modes   A flexible array, containing all the modes of
  *                the value numbers.
  */
-static void do_scalar_replacements(pset *sels, int nvals, ir_mode **modes) {
+static void do_scalar_replacements(ir_graph *irg, pset *sels, unsigned nvals,
+                                   ir_mode **modes)
+{
 	env_t env;
 
-	ssa_cons_start(current_ir_graph, nvals);
+	ssa_cons_start(irg, (int)nvals);
 
-	env.nvals     = nvals;
-	env.modes     = modes;
-	env.sels      = sels;
+	env.nvals = nvals;
+	env.modes = modes;
+	env.sels  = sels;
 
 	/*
 	 * second step: walk over the graph blockwise in topological order
 	 * and fill the array as much as possible.
 	 */
-	DB((dbg, SET_LEVEL_3, "Substituting Loads and Stores in %+F\n", current_ir_graph));
-	irg_walk_blkwise_graph(current_ir_graph, NULL, topologic_walker, &env);
+	DB((dbg, SET_LEVEL_3, "Substituting Loads and Stores in %+F\n", irg));
+	irg_walk_blkwise_graph(irg, NULL, walker, &env);
 
-	ssa_cons_finish(current_ir_graph);
+	ssa_cons_finish(irg);
 }
 
 /*
@@ -666,30 +672,23 @@ static void do_scalar_replacements(pset *sels, int nvals, ir_mode **modes) {
  *
  * @param irg  The current ir graph.
  */
-int scalar_replacement_opt(ir_graph *irg) {
+void scalar_replacement_opt(ir_graph *irg)
+{
 	unsigned  nvals;
-	int       i;
-	scalars_t key, *value;
+	scalars_t key;
 	ir_node   *irg_frame;
 	ir_mode   **modes;
 	set       *set_ent;
 	pset      *sels;
-	ir_type   *ent_type;
-	ir_graph  *rem;
-	int       res = 0;
+	ir_type   *ent_type, *frame_tp;
 
-	if (! get_opt_scalar_replacement())
-		return 0;
-
-	rem = current_ir_graph;
-	current_ir_graph = irg;
-
-	/* Call algorithm that computes the out edges */
-	assure_irg_outs(irg);
+	assure_irg_properties(irg,
+		IR_GRAPH_PROPERTY_NO_UNREACHABLE_CODE
+		| IR_GRAPH_PROPERTY_CONSISTENT_OUTS);
 
 	/* we use the link field to store the VNUM */
 	ir_reserve_resources(irg, IR_RESOURCE_IRN_LINK);
-	irp_reserve_resources(irp, IR_RESOURCE_ENTITY_LINK);
+	irp_reserve_resources(irp, IRP_RESOURCE_ENTITY_LINK);
 
 	/* Find possible scalar replacements */
 	if (find_possible_replacements(irg)) {
@@ -697,25 +696,31 @@ int scalar_replacement_opt(ir_graph *irg) {
 
 		/* Insert in set the scalar replacements. */
 		irg_frame = get_irg_frame(irg);
-		nvals = 0;
-		modes = NEW_ARR_F(ir_mode *, 16);
-		set_ent = new_set(ent_cmp, 8);
-		sels    = pset_new_ptr(8);
+		nvals     = 0;
+		modes     = NEW_ARR_F(ir_mode *, 16);
+		set_ent   = new_set(ent_cmp, 8);
+		sels      = pset_new_ptr(8);
+		frame_tp  = get_irg_frame_type(irg);
 
-		for (i = get_irn_n_outs(irg_frame) - 1; i >= 0; --i) {
+		for (unsigned i = get_irn_n_outs(irg_frame); i-- > 0; ) {
 			ir_node *succ = get_irn_out(irg_frame, i);
 
 			if (is_Sel(succ)) {
 				ir_entity *ent = get_Sel_entity(succ);
 
+				/* we are only interested in entities on the frame, NOT
+				   parameters */
+				if (get_entity_owner(ent) != frame_tp
+				    || is_parameter_entity(ent))
+					continue;
+
 				if (get_entity_link(ent) == NULL || get_entity_link(ent) == ADDRESS_TAKEN)
 					continue;
 
 				ent_type = get_entity_type(ent);
 
 				key.ent       = ent;
-				key.ent_owner = get_entity_owner(ent);
-				set_insert(set_ent, &key, sizeof(key), HASH_PTR(key.ent));
+				(void)set_insert(scalars_t, set_ent, &key, sizeof(key), hash_ptr(key.ent));
 
 #ifdef DEBUG_libfirm
 				if (is_Array_type(ent_type)) {
@@ -737,10 +742,10 @@ int scalar_replacement_opt(ir_graph *irg) {
 
 		/* If scalars were found. */
 		if (nvals > 0) {
-			do_scalar_replacements(sels, nvals, modes);
+			do_scalar_replacements(irg, sels, nvals, modes);
 
-			foreach_set(set_ent, value) {
-				remove_class_member(value->ent_owner, value->ent);
+			foreach_set(set_ent, scalars_t, value) {
+				free_entity(value->ent);
 			}
 
 			/*
@@ -748,23 +753,24 @@ int scalar_replacement_opt(ir_graph *irg) {
 			 * neither changed control flow, cf-backedges should be still
 			 * consistent.
 			 */
-			set_irg_outs_inconsistent(irg);
-			set_irg_loopinfo_inconsistent(irg);
-
-			res = 1;
 		}
 		del_pset(sels);
 		del_set(set_ent);
 		DEL_ARR_F(modes);
 	}
-	
-	ir_free_resources(irg, IR_RESOURCE_IRN_LINK);
-	irp_free_resources(irp, IR_RESOURCE_ENTITY_LINK);
 
-	current_ir_graph = rem;
-	return res;
+	ir_free_resources(irg, IR_RESOURCE_IRN_LINK);
+	irp_free_resources(irp, IRP_RESOURCE_ENTITY_LINK);
+
+	confirm_irg_properties(irg, IR_GRAPH_PROPERTIES_NONE);
 }
 
-void firm_init_scalar_replace(void) {
+ir_graph_pass_t *scalar_replacement_opt_pass(const char *name)
+{
+	return def_graph_pass(name ? name : "scalar_rep", scalar_replacement_opt);
+}
+
+void firm_init_scalar_replace(void)
+{
 	FIRM_DBG_REGISTER(dbg, "firm.opt.scalar_replace");
 }
