@@ -21,21 +21,19 @@
  * @file
  * @brief   Scalar replacement of compounds.
  * @author  Beyhan Veliev, Michael Beck
- * @version $Id: scalar_replace.h 19145 2008-04-05 18:38:22Z beck $
  */
 #ifndef FIRM_OPT_SCALAR_REPLACE_H
 #define FIRM_OPT_SCALAR_REPLACE_H
 
 #include "firm_types.h"
+#include <stdbool.h>
 
 /**
  * Returns non-zero, if the address of an entity
- * represented by a Sel node (or it's successor Sels) is taken.
+ * represented by a Sel node (or its successor Sels) is taken.
  *
  * @param sel  the Sel node
  */
-int is_address_taken(ir_node *sel);
-
-void firm_init_scalar_replace(void);
+bool is_address_taken(ir_node *sel);
 
 #endif /* FIRM_OPT_SCALAR_REPLACE_H */
